/*-
# MOTIF/X-BASED DINOSAUR CUBE
#
#  xdino.c
#
###
#
#  Copyright (c) 1995 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 5: 95/10/06 Xt/Motif
  Version 4: 94/05/30 Xt
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifdef VMS
#include <unixlib.h>
#define getlogin() cuserid(NULL)
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#if HAVE_FCNTL_H
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include "mouse-l.xbm"
#include "mouse-r.xbm"
#endif
#include "Dino.h"
#include "Dino2d.h"
#include "Dino3d.h"
#include "dino.xbm"

#ifndef SCOREFILE
#define SCOREFILE "/usr/games/lib/dino.scores"
#endif

/* The following are in DinoP.h also */
#define PERIOD2 2
#define PERIOD3 3
#define BOTH 4
#define MAXMODES 3
#define MAXFACES 6

#define NEVER -1
#define FILENAMELEN 1024
#define USERNAMELEN 128
#define NOACCESS "noaccess"
#define NOBODY "nobody"

typedef struct {
	int         score;
	char        name[USERNAMELEN];
} GameRecord;

#ifdef HAVE_MOTIF
static Widget moves, record, message, modes[MAXMODES],
	orientSwitch, practiceSwitch;
static char buff[16];
static const char *modeString[] =
{
	"Period 2", "Period 3", "Both"
};
#else
static char titleDsp[256] = "";
#endif

static Widget dino2d, dino3d;
static Arg  arg[5];
static GameRecord dinoRecord[MAXMODES][2];
static int  movesDsp = 0;
static char *progDsp;
static char messageDsp[128] = "Welcome";
static char recordDsp[16] = "NOT RECORDED";
static char usernameDsp[USERNAMELEN] = "";

static void
Usage(char * programName)
{
	(void) fprintf(stderr, "usage: %s\n", programName);
	(void) fprintf(stderr,
		"\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
	(void) fprintf(stderr,
		"\t[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n");
	(void) fprintf(stderr,
		"\t[-{foreground|fg} {color}] [-{background|bg} {color}]\n");
	(void) fprintf(stderr,
		"\t[-{border|bd} {color}] [-face{0|1|2|3|4|5} {color}]\n");
	(void) fprintf(stderr,
		"\t[-{mode {int} | both}] [-[no]orient] [-[no]practice]\n");
	(void) fprintf(stderr,
		"\t[-username {string}]\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*dino.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*dino.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*dino.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*dino.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*dino.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*dino.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*dino.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*dino.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*dino.pieceBorder", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*dino.pieceBorder", XrmoptionSepArg, NULL},
	{(char *) "-face0", (char *) "*dino.faceColor0", XrmoptionSepArg, NULL},
	{(char *) "-face1", (char *) "*dino.faceColor1", XrmoptionSepArg, NULL},
	{(char *) "-face2", (char *) "*dino.faceColor2", XrmoptionSepArg, NULL},
	{(char *) "-face3", (char *) "*dino.faceColor3", XrmoptionSepArg, NULL},
	{(char *) "-face4", (char *) "*dino.faceColor4", XrmoptionSepArg, NULL},
	{(char *) "-face5", (char *) "*dino.faceColor5", XrmoptionSepArg, NULL},
	{(char *) "-mode", (char *) "*dino.mode", XrmoptionSepArg, NULL},
	{(char *) "-both", (char *) "*dino.mode", XrmoptionNoArg, (char *) "4"},
	{(char *) "-orient", (char *) "*dino.orient", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noorient", (char *) "*dino.orient", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-practice", (char *) "*dino.practice", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nopractice", (char *) "*dino.practice", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-username", (char *) "*dino.userName", XrmoptionSepArg, NULL}
};

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *message)
{
	Arg         wargs[1];
	XmString    xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(wargs[0], XmNlabelString, xmstr);
	XtSetValues(w, wargs, 1);
}
#else
static void
PrintState(Widget w, int dim, int mode, int moves, char *message)
{
	char        mb[10];

	if (mode == BOTH)
		(void) strcpy(mb, "both");
	else
		(void) sprintf(mb, "%d", mode);
        (void) sprintf(titleDsp, "%s%dd.%s: (%d/%s) - %s",
        	progDsp, dim, mb, moves, recordDsp, message);
        XtSetArg(arg[0], XtNtitle, titleDsp);
        XtSetValues(w, arg, 1);
}
#endif

static void
InitRecords(void)
{
	int         mode, orient;

	for (mode = 0; mode < MAXMODES; mode++)
		for (orient = 0; orient < 2; orient++) {
			dinoRecord[mode][orient].score = NEVER;
			(void) strcpy(dinoRecord[mode][orient].name, NOACCESS);
		}
}

static void
ReadRecords(void)
{
	FILE       *fp;
	int         n, mode, orient;
	char        username[USERNAMELEN];

	if ((fp = fopen(SCOREFILE, "r")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
		for (mode = 0; mode < MAXMODES; mode++)
			for (orient = 0; orient < 2; orient++) {
				(void) fscanf(fp, "%d %s\n", &n, username);
				if (n <= dinoRecord[mode][orient].score ||
				    dinoRecord[mode][orient].score <= NEVER) {
					dinoRecord[mode][orient].score = n;
					(void) strcpy(dinoRecord[mode][orient].name, username);
				}
			}
		(void) fclose(fp);
	}
}

static void
WriteRecords(void)
{
	FILE       *fp;
	int         mode, orient;

	if ((fp = fopen(SCOREFILE, "w")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
#if HAVE_FCNTL_H
		int         lfd;
		char        lockfile[FILENAMELEN];

		(void) strcpy(lockfile, SCOREFILE);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
		       errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr, "Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr, "Lock file exists... score not recorded - sorry.\n"
				);
			return;
#endif
		}
#endif
		for (mode = 0; mode < MAXMODES; mode++) {
			for (orient = 0; orient < 2; orient++)
				(void) fprintf(fp, "%d %s\n",
					       dinoRecord[mode][orient].score, dinoRecord[mode][orient].name);
			(void) fprintf(fp, "\n");
		}
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
}

static void
PrintRecord(int mode, Boolean orient, Boolean practice)
{
	int         i = mode - PERIOD2;
	int         j = (orient) ? 1 : 0;

	if (practice)
		(void) strcpy(recordDsp, "practice");
	else if (dinoRecord[i][j].score <= NEVER) {
		(void) sprintf(recordDsp, "NEVER %s", NOACCESS);
	} else {
		(void) sprintf(recordDsp, "%d %s",
			dinoRecord[i][j].score, dinoRecord[i][j].name);
	}
#ifdef HAVE_MOTIF
	PrintState(record, recordDsp);
#endif
}

static      Boolean
HandleSolved(int counter, int mode, Boolean orient)
{
	int         i = mode - PERIOD2;
	int         j = (orient) ? 1 : 0;

	if (counter < dinoRecord[i][j].score ||
	    dinoRecord[i][j].score <= NEVER) {
		ReadRecords();	/* Maybe its been updated by another */
		dinoRecord[i][j].score = counter;
		(void) strcpy(dinoRecord[i][j].name, usernameDsp);
		if (orient && (counter < dinoRecord[!i][j].score ||
		    dinoRecord[!i][j].score <= NEVER)) {
			dinoRecord[!i][j].score = counter;
			(void) strcpy(dinoRecord[!i][j].name, usernameDsp);
		}
		WriteRecords();
		PrintRecord(mode, orient, False);
		return True;
	}
	return False;
}

/* There's probably a better way to assure that they are the same but I don't
   know it off hand. */
static void
MakeEquivalent(String * username, int *mode, Boolean * orient, Boolean * practice)
{
	Boolean     mono, reverse;
	Pixel       foreground, background, pieceBorder;
	String      faceColor[MAXFACES];

	XtVaGetValues(dino2d,
		      XtNuserName, username,
		      XtNmode, mode,
		      XtNorient, orient,
		      XtNpractice, practice,
		      XtNmono, &mono,
		      XtNreverse, &reverse,
		      XtNforeground, &foreground,
		      XtNbackground, &background,
		      XtNpieceBorder, &pieceBorder,
		      XtNfaceColor0, &(faceColor[0]),
		      XtNfaceColor1, &(faceColor[1]),
		      XtNfaceColor2, &(faceColor[2]),
		      XtNfaceColor3, &(faceColor[3]),
		      XtNfaceColor4, &(faceColor[4]),
		      XtNfaceColor5, &(faceColor[5]), NULL);
	XtVaSetValues(dino2d,
		      XtNdirection, DINO_IGNORE,
		      XtNstart, False, NULL);
	XtVaSetValues(dino3d,
		      XtNuserName, *username,
		      XtNmode, *mode,
		      XtNorient, *orient,
		      XtNpractice, *practice,
		      XtNmono, mono,
		      XtNreverse, reverse,
		      XtNdirection, DINO_IGNORE,
		      XtNstart, False,
		      XtNforeground, foreground,
		      XtNbackground, background,
		      XtNpieceBorder, pieceBorder,
		      XtNfaceColor0, faceColor[0],
		      XtNfaceColor1, faceColor[1],
		      XtNfaceColor2, faceColor[2],
		      XtNfaceColor3, faceColor[3],
		      XtNfaceColor4, faceColor[4],
		      XtNfaceColor5, faceColor[5], NULL);
}

static void
Initialize(void)
{
	int         mode;
	Boolean     orient, practice;
	String      username;

	MakeEquivalent(&username, &mode, &orient, &practice);
#ifdef HAVE_MOTIF
	XmToggleButtonSetState(modes[mode - PERIOD2], True, False);
	XmToggleButtonSetState(orientSwitch, orient, True);
	XmToggleButtonSetState(practiceSwitch, practice, True);
#endif
	InitRecords();
	ReadRecords();
	(void) strcpy(usernameDsp, username);
	if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
	    !strcmp(usernameDsp, NOACCESS) || !strcmp(usernameDsp, NOBODY)) {
		/* The NOACCESS is not necessary, but it stops people from being cute. */
		(void) sprintf(usernameDsp, "%s", getlogin());
		if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
		    !strcmp(usernameDsp, NOACCESS))
			(void) sprintf(usernameDsp, "%s", NOBODY);	/* It really IS nobody */
	}
	PrintRecord(mode, orient, practice);
#ifndef HAVE_MOTIF
	PrintState(XtParent(dino2d), 2, mode, movesDsp, messageDsp);
	PrintState(XtParent(dino3d), 3, mode, movesDsp, messageDsp);
#endif
}

static void
CallbackDino(Widget w, caddr_t clientData, dinoCallbackStruct * callData)
{
	int         mode;
	Boolean     orient, practice, start;
	Widget      otherw;
#ifndef HAVE_MOTIF
	int         dim, otherdim;
#endif

	if (w == dino2d) {
		otherw = dino3d;
#ifndef HAVE_MOTIF
		dim = 2, otherdim = 3;
#endif
	} else {	/* (w == dino3d) */
		otherw = dino2d;
#ifndef HAVE_MOTIF
		dim = 3, otherdim = 2;
#endif
	}
	XtVaGetValues(w,
		      XtNorient, &orient,
		      XtNmode, &mode,
		      XtNpractice, &practice,
		      XtNstart, &start, NULL);
	(void) strcpy(messageDsp, "");
	switch (callData->reason) {
		case DINO_RESTORE:
			XtSetArg(arg[0], XtNdirection, DINO_RESTORE);
			XtSetValues(otherw, arg, 1);
			XtSetValues(w, arg, 1);
			movesDsp = 0;
			break;
		case DINO_CLEAR:
			XtSetArg(arg[0], XtNdirection, DINO_CLEAR);
			XtSetValues(otherw, arg, 1);
			XtSetValues(w, arg, 1);
			movesDsp = 0;
			break;
		case DINO_RESET:
			movesDsp = 0;
			break;
		case DINO_ILLEGAL:
			if (practice || start)
				(void) strcpy(messageDsp, "Illegal move");
			else
				(void) strcpy(messageDsp, "Randomize to start");
			break;
		case DINO_MOVED:
			movesDsp++;
			XtSetArg(arg[0], XtNstart, True);
			XtSetArg(arg[1], XtNface, callData->face);
			XtSetArg(arg[2], XtNpos, callData->position);
			XtSetArg(arg[3], XtNdirection, callData->direction);
			XtSetArg(arg[4], XtNstyle, callData->style);
			XtSetValues(otherw, arg, 5);
			XtSetValues(w, arg, 1);
			break;
		case DINO_CONTROL:
			XtSetArg(arg[0], XtNface, callData->face);
			XtSetArg(arg[1], XtNpos, callData->position);
			XtSetArg(arg[2], XtNdirection, callData->direction);
			XtSetArg(arg[3], XtNstyle, callData->style);
			XtSetValues(otherw, arg, 4);
			return;
		case DINO_SOLVED:
			if (practice)
				movesDsp = 0;
			else {
				if (HandleSolved(movesDsp, mode, orient))
					(void) sprintf(messageDsp, "Congratulations %s!!", usernameDsp);
				else
					(void) strcpy(messageDsp, "Solved!");
			}
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			break;
		case DINO_PRACTICE:
			movesDsp = 0;
			practice = !practice;
			if (!practice)
				(void) strcpy(messageDsp, "Randomize to start");
			XtSetArg(arg[0], XtNpractice, practice);
			XtSetArg(arg[1], XtNstart, False);
			XtSetValues(w, arg, 2);
			XtSetValues(otherw, arg, 2);
			PrintRecord(mode, orient, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(practiceSwitch, practice, True);
#endif
			break;
		case DINO_RANDOMIZE:
			movesDsp = 0;
			XtSetArg(arg[0], XtNpractice, False);
			XtSetArg(arg[1], XtNstart, False);
			XtSetValues(w, arg, 2);
			XtSetValues(otherw, arg, 2);
			break;
		case DINO_ORIENT:
			movesDsp = 0;
			orient = !orient;
			XtSetArg(arg[0], XtNorient, orient);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(mode, orient, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(orientSwitch, orient, True);
#endif
			break;
		case DINO_PERIOD2:
		case DINO_PERIOD3:
		case DINO_BOTH:
			movesDsp = 0;
			mode = callData->reason - DINO_PERIOD2 + PERIOD2;
			XtSetArg(arg[0], XtNmode, mode);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(mode, orient, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(modes[mode - PERIOD2], True, True);
#endif
			break;
		case DINO_COMPUTED:
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			break;
		case DINO_UNDO:
			movesDsp--;
			XtSetArg(arg[0], XtNstart, True);
			XtSetArg(arg[1], XtNface, callData->face);
			XtSetArg(arg[2], XtNpos, callData->position);
			XtSetArg(arg[3], XtNdirection, callData->direction);
			XtSetArg(arg[4], XtNstyle, callData->style);
			XtSetValues(otherw, arg, 5);
			XtSetValues(w, arg, 1);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
#else
	PrintState(XtParent(w), dim, mode, movesDsp, messageDsp);
	PrintState(XtParent(otherw), otherdim, mode, movesDsp, messageDsp);
#endif
}

#ifdef HAVE_MOTIF
static void
ModeToggle(Widget w, int mode, XmToggleButtonCallbackStruct * cbs)
{
	Boolean     orient, practice;

	if (cbs->set) {
		XtVaGetValues(dino2d,
			      XtNorient, &orient,
			      XtNpractice, &practice, NULL);
		XtVaSetValues(dino2d,
			      XtNmode, mode + PERIOD2, NULL);
		XtVaSetValues(dino3d,
			      XtNmode, mode + PERIOD2, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(mode + PERIOD2, orient, practice);
	}
}

static void
OrientToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         mode;
	Boolean     orient = cbs->set, practice;

	XtVaGetValues(dino2d,
		      XtNmode, &mode,
		      XtNpractice, &practice, NULL);
	XtVaSetValues(dino2d,
		      XtNorient, orient, NULL);
	XtVaSetValues(dino3d,
		      XtNorient, orient, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	PrintRecord(mode, orient, practice);
}

static void
PracticeToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         mode;
	Boolean     orient, practice = cbs->set;

	XtVaSetValues(dino2d,
		      XtNpractice, practice,
		      XtNstart, False, NULL);
	XtVaSetValues(dino3d,
		      XtNpractice, practice,
		      XtNstart, False, NULL);
	XtVaGetValues(dino2d,
		      XtNmode, &mode,
		      XtNpractice, &orient, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	if (!practice)
		(void) strcpy(messageDsp, "Randomize to start");
	PrintRecord(mode, orient, practice);
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2)
		exit(0);
	XtSetArg(arg[0], XtNmenu, val);
	XtSetValues(dino2d, arg, 1);
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;
	XtSetArg(arg[0], XtNmenu, val + 3);  /* GWQ */
	XtSetValues(dino2d, arg, 1);
}

static const char aboutHelp[] = {
	"Version 5.6.1 by David Bagley <bagleyd@tux.org> http://www.tux.org/~bagleyd"
};

static const char helpHelp[] = {
	"Consult man page.  :)"
};

static void
helpCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		(void) printf("%s: %s\n", progDsp, aboutHelp);
	else
		(void) printf("%s: %s\n", progDsp, helpHelp);
}
#endif

int
main(int argc, char **argv)
{
	Widget      topLevel;
#ifdef HAVE_MOTIF
	Widget      menubar, panel, panel2, panel3;
	Widget      rowcol, rowcol2, rowcol3, rowcol4;
	Pixmap      mouseLeftCursor, mouseRightCursor;
	Pixel       fg, bg;
	XmString    file, play, help, about;
	XmString    get, write, quit;
	XmString    clear, undo, randomize, solve, orient, practice;
	unsigned int i;
#else
	Widget      shell;
#endif

	progDsp = argv[0];
	topLevel = XtInitialize(argv[0], "Dino",
				options, XtNumber(options), &argc, argv);
	if (argc != 1)
		Usage(argv[0]);

	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(topLevel),
				      RootWindowOfScreen(XtScreen(topLevel)),
			       (char *) dino_bits, dino_width, dino_height));
#ifdef HAVE_MOTIF
	XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER);	/* not XmEXPLICIT */
	XtSetValues(topLevel, arg, 2);
	panel = XtVaCreateManagedWidget("panel",
				xmPanedWindowWidgetClass, topLevel,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	file = XmStringCreateSimple((char *) "File");
	play = XmStringCreateSimple((char *) "Play");
	help = XmStringCreateSimple((char *) "Help");
	menubar = XmVaCreateSimpleMenuBar(panel, (char *) "menubar",
		XmVaCASCADEBUTTON, file, 'F',
		XmVaCASCADEBUTTON, play, 'P',
		XmVaCASCADEBUTTON, help, 'H',
		NULL);
	XmStringFree(file);
	XmStringFree(play);
	get = XmStringCreateSimple((char *) "Get");
	write = XmStringCreateSimple((char *) "Write");
	quit = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "file_menu", 0, fileCB,
		XmVaPUSHBUTTON, get, 'G', NULL, NULL,
		XmVaPUSHBUTTON, write, 'W', NULL, NULL,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quit, 'Q', NULL, NULL,
		NULL);
	XmStringFree(get);
	XmStringFree(write);
	XmStringFree(quit);
	clear = XmStringCreateSimple((char *) "Clear");
	undo = XmStringCreateSimple((char *) "Undo");
	randomize = XmStringCreateSimple((char *) "Randomize");
	solve = XmStringCreateSimple((char *) "(Solve)");
	orient = XmStringCreateSimple((char *) "Orientize");
	practice = XmStringCreateSimple((char *) "Practice");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, clear, 'C', NULL, NULL,
		XmVaPUSHBUTTON, undo, 'U', NULL, NULL,
		XmVaPUSHBUTTON, randomize, 'R', NULL, NULL,
		XmVaPUSHBUTTON, solve, 'S', NULL, NULL,
		XmVaPUSHBUTTON, orient, 'O', NULL, NULL,
		XmVaPUSHBUTTON, practice, 'P', NULL, NULL,
		NULL);
	XmStringFree(clear);
	XmStringFree(undo);
	XmStringFree(randomize);
	XmStringFree(solve);
	XmStringFree(orient);
	XmStringFree(practice);
	about = XmStringCreateSimple((char *) "About");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "help_menu", 2, helpCB,
		XmVaPUSHBUTTON, about, 'A', NULL, NULL,
		XmVaPUSHBUTTON, help, 'H', NULL, NULL,
		NULL);
	XmStringFree(about);
	XmStringFree(help);
	XtManageChild(menubar);
	panel2 = XtCreateManagedWidget("panel2",
				       xmPanedWindowWidgetClass, panel,
				       NULL, 0);
	panel3 = XtVaCreateManagedWidget("panel3",
					 xmPanedWindowWidgetClass, panel2,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	rowcol = XtVaCreateManagedWidget("Rowcol", xmRowColumnWidgetClass, panel3,
					 XmNnumColumns, 2,
					 XmNorientation, XmHORIZONTAL,
					 XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(rowcol,
		      XmNforeground, &fg,
		      XmNbackground, &bg, NULL);
	mouseLeftCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	      RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_left_bits,
				 mouse_left_width, mouse_left_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	mouseRightCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	     RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_right_bits,
			       mouse_right_width, mouse_right_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	(void) XtVaCreateManagedWidget("mouseLeftText",
				xmLabelGadgetClass, rowcol,
		   XtVaTypedArg, XmNlabelString, XmRString, "Move", 5, NULL);
	(void) XtVaCreateManagedWidget("mouseLeft",
				xmLabelGadgetClass, rowcol,
	      XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseLeftCursor, NULL);
	(void) XtVaCreateManagedWidget("mouseRightText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Randomize", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseRight",
				xmLabelGadgetClass, rowcol,
	     XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseRightCursor, NULL);
	(void) XtVaCreateManagedWidget("movesText",
				xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Moves", 6, NULL);
	moves = XtVaCreateManagedWidget("0",
					xmLabelWidgetClass, rowcol, NULL);
	(void) XtVaCreateManagedWidget("recordText",
				xmLabelGadgetClass, rowcol,
		 XtVaTypedArg, XmNlabelString, XmRString, "Record", 7, NULL);
	record = XtVaCreateManagedWidget("0",
					 xmLabelWidgetClass, rowcol, NULL);

	rowcol2 = XtVaCreateManagedWidget("Rowcol2",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN,
					  XmNradioBehavior, True, NULL);
	for (i = 0; i < XtNumber(modeString); i++) {
		modes[i] = XtVaCreateManagedWidget(modeString[i],
					  xmToggleButtonGadgetClass, rowcol2,
					       XmNradioBehavior, True, NULL);
		XtAddCallback(modes[i],
			      XmNvalueChangedCallback, (XtCallbackProc) ModeToggle, (XtPointer) i);
	}

	rowcol3 = XtVaCreateManagedWidget("Rowcol3",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	orientSwitch = XtVaCreateManagedWidget("Oriented",
				   xmToggleButtonWidgetClass, rowcol3, NULL);
	XtAddCallback(orientSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) OrientToggle, (XtPointer) NULL);
	practiceSwitch = XtVaCreateManagedWidget("Practice",
				   xmToggleButtonWidgetClass, rowcol3, NULL);
	XtAddCallback(practiceSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) PracticeToggle, (XtPointer) NULL);

	rowcol4 = XtVaCreateManagedWidget("Rowcol4",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	message = XtVaCreateManagedWidget("Play Dino! (use mouse and keypad)",
					  xmLabelWidgetClass, rowcol4, NULL);

	dino2d = XtCreateManagedWidget("dino",
				       dino2dWidgetClass, panel2, NULL, 0);
	XtVaSetValues(dino2d,
		      XtNheight, 200, NULL);
	XtAddCallback(dino2d,
	  XtNselectCallback, (XtCallbackProc) CallbackDino, (XtPointer) NULL);
	dino3d = XtCreateManagedWidget("dino",
				       dino3dWidgetClass, panel2, NULL, 0);
	XtVaSetValues(dino3d,
		      XtNheight, 200, NULL);
#else
	XtSetArg(arg[1], XtNinput, True);
	XtSetValues(topLevel, arg, 2);
	shell = XtCreateApplicationShell(argv[0],
                                         topLevelShellWidgetClass, NULL, 0);
        XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(shell),
					 RootWindowOfScreen(XtScreen(shell)),
			    (char *) dino_bits, dino_width, dino_height));
	XtSetValues(shell, arg, 2);
	dino2d = XtCreateManagedWidget("dino",
				       dino2dWidgetClass, topLevel, NULL, 0);
	XtAddCallback(dino2d,
	  XtNselectCallback, (XtCallbackProc) CallbackDino, (XtPointer) NULL);
	dino3d = XtCreateManagedWidget("dino",
				       dino3dWidgetClass, shell, NULL, 0);
#endif
	XtAddCallback(dino3d,
	  XtNselectCallback, (XtCallbackProc) CallbackDino, (XtPointer) NULL);
	Initialize();
	XtRealizeWidget(topLevel);
#ifndef HAVE_MOTIF
	XtRealizeWidget(shell);
#endif
	XGrabButton(XtDisplay(dino2d), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(dino2d), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(dino2d),
		    XCreateFontCursor(XtDisplay(dino2d), XC_crosshair));
	XGrabButton(XtDisplay(dino3d), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(dino3d), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(dino3d),
		    XCreateFontCursor(XtDisplay(dino3d), XC_crosshair));
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
