/*-
# MOTIF/X-BASED HEXAGONS
#
#  xhexagons.c
#
###
#
#  Copyright (c) 1993 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 5: 95/10/01 Xt/Motif
  Version 4: 94/06/07 Xt
  Version 3: 93/04/01 Motif
  Version 2: 92/01/07 XView
  Version 1: 91/09/05 SunView
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifdef VMS
#include <unixlib.h>
#define getlogin() cuserid(NULL)
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#if HAVE_FCNTL_H
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include "mouse-l.xbm"
#include "mouse-r.xbm"
#endif
#include "Hexagons.h"
#include "hexagons.xbm"

#ifndef SCOREFILE
#define SCOREFILE "/usr/games/lib/hexagons.scores"
#endif

/* The following are in HexagonsP.h also */
#define MINHEXAGONS 1
#define MAXORIENT 2
#define NOCORN 0
#define CORNERS 1

#define MAXHEXAGONS 10
#define NEVER -1
#define FILENAMELEN 1024
#define USERNAMELEN 128
#define NOACCESS "noaccess"
#define NOBODY "nobody"

typedef struct {
	int         score;
	char        name[USERNAMELEN];
} GameRecord;

#ifdef HAVE_MOTIF
static Widget moves, record, message, cornersSwitch, hexX, hexY;
static char buff[16];
#else
static char titleDsp[256] = "";
#endif

static Widget hexagons;
static Arg  arg[2];
static GameRecord hexagonsRecord[MAXORIENT][MAXHEXAGONS - MINHEXAGONS + 1][MAXHEXAGONS - MINHEXAGONS + 1];
static int  movesDsp = 0;
static char *progDsp;
static char messageDsp[128] = "Welcome";
static char recordDsp[16] = "NOT RECORDED";
static char usernameDsp[USERNAMELEN] = "";

static void
Usage(char * programName)
{
	(void) fprintf(stderr, "usage: %s\n", programName);
	(void) fprintf(stderr,
	     "\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
	(void) fprintf(stderr,
	   "\t[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n");
	(void) fprintf(stderr,
	   "\t[-{foreground|fg} {color}] [-{background|bg} {color}]\n");
	(void) fprintf(stderr,
	   "\t[-{border|bd} {color}] [-tile {color}] [-size{x|y} {int}]\n");
	(void) fprintf(stderr,
	   "\t[-[no]corners] [-base {int}] [-username {string}]\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*hexagons.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*hexagons.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*hexagons.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*hexagons.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*hexagons.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*hexagons.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*hexagons.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*hexagons.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*hexagons.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*hexagons.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-tile", (char *) "*hexagons.tileColor", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*hexagons.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-sizex", (char *) "*hexagons.sizeX", XrmoptionSepArg, NULL},
	{(char *) "-sizey", (char *) "*hexagons.sizeY", XrmoptionSepArg, NULL},
	{(char *) "-corners", (char *) "*hexagons.corners", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nocorners", (char *) "*hexagons.corners", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-base", (char *) "*hexagons.base", XrmoptionSepArg, NULL},
	{(char *) "-username", (char *) "*hexagons.userName", XrmoptionSepArg, NULL}
};

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *message)
{
	Arg         wargs[1];
	XmString    xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(wargs[0], XmNlabelString, xmstr);
	XtSetValues(w, wargs, 1);
}
#else
static void
PrintState(Widget w, int sizeX, int sizeY, int moves, char *message)
{
	(void) sprintf(titleDsp, "%s: %dx%d@ (%d/%s) - %s",
		progDsp, sizeX, sizeY, moves, recordDsp, message);
	XtSetArg(arg[0], XtNtitle, titleDsp);
	XtSetValues(w, arg, 1);
}
#endif

static void
InitRecords(void)
{
	int         i, j, k;

	for (i = 0; i < MAXORIENT; i++)
		for (j = 0; j < MAXHEXAGONS - MINHEXAGONS + 1; j++)
			for (k = 0; k < MAXHEXAGONS - MINHEXAGONS + 1; k++) {
				hexagonsRecord[i][j][k].score = NEVER;
				(void) strcpy(hexagonsRecord[i][j][k].name,
					NOACCESS);
			}
}

static void
ReadRecords(void)
{
	FILE       *fp;
	int         i, j, k, n;
	char        username[USERNAMELEN];

	if ((fp = fopen(SCOREFILE, "r")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
		for (i = 0; i < MAXORIENT; i++)
			for (j = 0; j < MAXHEXAGONS - MINHEXAGONS + 1; j++)
			  for (k = 0; k < MAXHEXAGONS - MINHEXAGONS + 1; k++) {
			    (void) fscanf(fp, "%d %s\n", &n, username);
			    if (n <= hexagonsRecord[i][j][k].score ||
			        hexagonsRecord[i][j][k].score <= NEVER) {
			      hexagonsRecord[i][j][k].score = n;
			      (void) strcpy(hexagonsRecord[i][j][k].name,
			        username);
			    }
			}
		(void) fclose(fp);
	}
}

static void
WriteRecords(void)
{
	FILE       *fp;
	int         i, j, k;

	if ((fp = fopen(SCOREFILE, "w")) == NULL) {
		(void) sprintf(messageDsp, "Can not write to %s.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
#if HAVE_FCNTL_H
		int         lfd;
		char        lockfile[FILENAMELEN];

		(void) strcpy(lockfile, SCOREFILE);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
		       errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr, "Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr, "Lock file exists... score not recorded - sorry.\n"
				);
			return;
#endif
		}
#endif
		for (i = 0; i < MAXORIENT; i++) {
			for (j = 0; j < MAXHEXAGONS - MINHEXAGONS + 1; j++)
			  for (k = 0; k < MAXHEXAGONS - MINHEXAGONS + 1; k++)
			    (void) fprintf(fp, "%d %s\n",
			    hexagonsRecord[i][j][k].score,
			    hexagonsRecord[i][j][k].name);
			(void) fprintf(fp, "\n");
		}
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
}

static void
PrintRecord(int sizeX, int sizeY, Boolean corners)
{
	int         i = (corners) ? 1 : 0;
	int         j = sizeX - MINHEXAGONS, k = sizeY - MINHEXAGONS;

	if (sizeX > MAXHEXAGONS || sizeY > MAXHEXAGONS)
		(void) strcpy(recordDsp, "NOT RECORDED");
	else if (hexagonsRecord[i][j][k].score <= NEVER) {
		(void) sprintf(recordDsp, "NEVER %s", NOACCESS);
	} else {
		(void) sprintf(recordDsp, "%d %s",
		      hexagonsRecord[i][j][k].score,
		      hexagonsRecord[i][j][k].name);
	}
#ifdef HAVE_MOTIF
	PrintState(record, recordDsp);
#endif
}

static Boolean
HandleSolved(int counter, int sizeX, int sizeY, int corners)
{
	int         i = (corners) ? 1 : 0;
	int         j = sizeX - MINHEXAGONS, k = sizeY - MINHEXAGONS;

	if (sizeX <= MAXHEXAGONS && sizeY <= MAXHEXAGONS &&
	    (counter < hexagonsRecord[i][j][k].score ||
	    hexagonsRecord[i][j][k].score <= NEVER)) {
		ReadRecords();	/* Maybe its been updated by another */
		hexagonsRecord[i][j][k].score = counter;
		(void) strcpy(hexagonsRecord[i][j][k].name, usernameDsp);
		WriteRecords();
		PrintRecord(sizeX, sizeY, corners);
		return True;
	}
	return False;
}

static void
Initialize(Widget w)
{
	int         sizeX, sizeY;
	Boolean     corners;
	String      username;

	XtVaSetValues(w,
		      XtNstart, False, NULL);
	XtVaGetValues(w,
		      XtNuserName, &username,
		      XtNsizeX, &sizeX,
		      XtNsizeY, &sizeY,
		      XtNcorners, &corners, NULL);
#ifdef HAVE_MOTIF
	if (sizeX > MAXHEXAGONS)
		XtVaSetValues(hexX, XmNmaximum, sizeX, NULL);
	XmScaleSetValue(hexX, sizeX);
	if (sizeY > MAXHEXAGONS)
		XtVaSetValues(hexY, XmNmaximum, sizeY, NULL);
	XmScaleSetValue(hexY, sizeY);
	XmToggleButtonSetState(cornersSwitch, corners, True);
#endif
	InitRecords();
	ReadRecords();
	(void) strcpy(usernameDsp, username);
	if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
	    !strcmp(usernameDsp, NOACCESS) || !strcmp(usernameDsp, NOBODY)) {
		/* The NOACCESS is not necessary, but it stops people from being cute. */
		(void) sprintf(usernameDsp, "%s", getlogin());
		if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
		    !strcmp(usernameDsp, NOACCESS))
			(void) sprintf(usernameDsp, "%s", NOBODY);	/* It really IS nobody */
	}
	PrintRecord(sizeX, sizeY, corners);
#ifndef HAVE_MOTIF
	PrintState(XtParent(w), sizeX, sizeY, movesDsp, messageDsp);
#endif

}

static void
CallbackHexagons(Widget w, caddr_t clientData,
		 hexagonsCallbackStruct * callData)
{
	int         sizeX, sizeY;
	Boolean     corners;

	XtVaGetValues(w,
		      XtNsizeX, &sizeX,
		      XtNsizeY, &sizeY,
		      XtNcorners, &corners, NULL);
	(void) strcpy(messageDsp, "");
	switch (callData->reason) {
		case HEXAGONS_RESTORE:
		case HEXAGONS_RESET:
			movesDsp = 0;
			break;
		case HEXAGONS_BLOCKED:
			(void) strcpy(messageDsp, "Blocked");
			break;
		case HEXAGONS_SPACE:
#if 0
			/* Too annoying */
			(void) strcpy(messageDsp, "Spaces can not slide");
#endif
			break;
		case HEXAGONS_IGNORE:
			(void) strcpy(messageDsp, "Randomize to start");
			break;
		case HEXAGONS_MOVED:
			movesDsp++;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
		case HEXAGONS_SOLVED:
			if (HandleSolved(movesDsp, sizeX, sizeY, corners))
				(void) sprintf(messageDsp, "Congratulations %s!!", usernameDsp);
			else
				(void) strcpy(messageDsp, "Solved!");
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case HEXAGONS_RANDOMIZE:
			movesDsp = 0;
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case HEXAGONS_DEC_X:
			movesDsp = 0;
			sizeX--;
			XtSetArg(arg[0], XtNsizeX, sizeX);
			XtSetValues(w, arg, 1);
			PrintRecord(sizeX, sizeY, corners);
#ifdef HAVE_MOTIF
			XmScaleSetValue(hexX, sizeX);
			if (sizeX >= MAXHEXAGONS)
				XtVaSetValues(hexX, XmNmaximum, sizeX, NULL);
#endif
			break;
		case HEXAGONS_INC_X:
			movesDsp = 0;
			sizeX++;
			XtSetArg(arg[0], XtNsizeX, sizeX);
			XtSetValues(w, arg, 1);
			PrintRecord(sizeX, sizeY, corners);
#ifdef HAVE_MOTIF
			if (sizeX > MAXHEXAGONS)
				XtVaSetValues(hexX, XmNmaximum, sizeX, NULL);
			XmScaleSetValue(hexX, sizeX);
#endif
			break;
		case HEXAGONS_DEC_Y:
			movesDsp = 0;
			sizeY--;
			XtSetArg(arg[0], XtNsizeY, sizeY);
			XtSetValues(w, arg, 1);
			PrintRecord(sizeX, sizeY, corners);
#ifdef HAVE_MOTIF
			XmScaleSetValue(hexY, sizeY);
			if (sizeY >= MAXHEXAGONS)
				XtVaSetValues(hexY, XmNmaximum, sizeY, NULL);
#endif
			break;
		case HEXAGONS_INC_Y:
			movesDsp = 0;
			sizeY++;
			XtSetArg(arg[0], XtNsizeY, sizeY);
			XtSetValues(w, arg, 1);
			PrintRecord(sizeX, sizeY, corners);
#ifdef HAVE_MOTIF
			if (sizeY > MAXHEXAGONS)
				XtVaSetValues(hexY, XmNmaximum, sizeY, NULL);
			XmScaleSetValue(hexY, sizeY);
#endif
			break;
		case HEXAGONS_CORNERS:
			movesDsp = 0;
			corners = !corners;
			XtSetArg(arg[0], XtNcorners, corners);
			XtSetValues(w, arg, 1);
			PrintRecord(sizeX, sizeY, corners);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(cornersSwitch, corners, True);
#endif
			break;
		case HEXAGONS_COMPUTED:
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case HEXAGONS_UNDO:
			movesDsp--;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
#else
	PrintState(XtParent(w), sizeX, sizeY, movesDsp, messageDsp);
#endif
}

#ifdef HAVE_MOTIF
static void
HexXSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         sizeX = cbs->value, sizeY, oldX;

	XtVaGetValues(hexagons,
		      XtNsizeX, &oldX,
		      XtNsizeY, &sizeY, NULL);
	if (oldX != sizeX) {
		XtVaSetValues(hexagons,
			      XtNsizeX, sizeX,
			      XtNsizeY, sizeY, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(sizeX, sizeY,
			XmToggleButtonGetState(cornersSwitch));
	}
}

static void
HexYSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         sizeX, sizeY = cbs->value, oldY;
	Boolean     corners;

	XtVaGetValues(hexagons,
		      XtNcorners, &corners,
		      XtNsizeX, &sizeX,
		      XtNsizeY, &oldY, NULL);
	if (oldY == sizeY)
		return;
	if (corners && sizeY == 1) {
		XmScaleSetValue(hexY, oldY);
		return;
	}
	XtVaSetValues(hexagons,
		      XtNsizeX, sizeX,
		      XtNsizeY, sizeY, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	PrintRecord(sizeX, sizeY,
		XmToggleButtonGetState(cornersSwitch));
}

static void
CornersToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         sizeX, sizeY;
	Boolean     corners = cbs->set;

	XtVaGetValues(hexagons,
		      XtNsizeX, &sizeX,
		      XtNsizeY, &sizeY, NULL);
	if (corners && sizeY == 1) {
		XmToggleButtonSetState(cornersSwitch, !corners, True);
		return;
	}
	XtVaSetValues(hexagons,
		      XtNcorners, corners, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	PrintRecord(sizeX, sizeY, corners);
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2)
		exit(0);
	XtSetArg(arg[0], XtNmenu, val);
	XtSetValues(hexagons, arg, 1);
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;
	XtSetArg(arg[0], XtNmenu, val + 3);  /* GWQ */
	XtSetValues(hexagons, arg, 1);
}

static const char aboutHelp[] = {
	"Version 5.6.1 by David Bagley <bagleyd@tux.org> http://www.tux.org/~bagleyd"
};

static const char helpHelp[] = {
	"Consult man page.  :)"
};

static void
helpCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		(void) printf("%s: %s\n", progDsp, aboutHelp);
	else
		(void) printf("%s: %s\n", progDsp, helpHelp);
}
#endif

int
main(int argc, char **argv)
{
	Widget      topLevel;
#ifdef HAVE_MOTIF
	Widget      menubar;
	Widget      panel, panel2, panel3, rowcol, rowcol2, rowcol3;
	Pixmap      mouseLeftCursor, mouseRightCursor;
	Pixel       fg, bg;
	XmString    file, play, help, about;
	XmString    get, write, quit;
	XmString    clear, undo, randomize, solve, mode;
#endif

	progDsp = argv[0];

	topLevel = XtInitialize(argv[0], "Hexagons",
				options, XtNumber(options), &argc, argv);
	if (argc != 1)
		Usage(argv[0]);

	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(topLevel),
				      RootWindowOfScreen(XtScreen(topLevel)),
		   (char *) hexagons_bits, hexagons_width, hexagons_height));
#ifdef HAVE_MOTIF
	XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER);	/* not XmEXPLICIT */
	XtSetValues(topLevel, arg, 2);
	panel = XtVaCreateManagedWidget("panel",
				xmPanedWindowWidgetClass, topLevel,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	file = XmStringCreateSimple((char *) "File");
	play = XmStringCreateSimple((char *) "Play");
	help = XmStringCreateSimple((char *) "Help");
	menubar = XmVaCreateSimpleMenuBar(panel, (char *) "menubar",
		XmVaCASCADEBUTTON, file, 'F',
		XmVaCASCADEBUTTON, play, 'P',
		XmVaCASCADEBUTTON, help, 'H',
		NULL);
	XmStringFree(file);
	XmStringFree(play);
	get = XmStringCreateSimple((char *) "Get");
	write = XmStringCreateSimple((char *) "Write");
	quit = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "file_menu", 0, fileCB,
		XmVaPUSHBUTTON, get, 'G', NULL, NULL,
		XmVaPUSHBUTTON, write, 'W', NULL, NULL,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quit, 'Q', NULL, NULL,
		NULL);
	XmStringFree(get);
	XmStringFree(write);
	XmStringFree(quit);
	clear = XmStringCreateSimple((char *) "Clear");
	undo = XmStringCreateSimple((char *) "Undo");
	randomize = XmStringCreateSimple((char *) "Randomize");
	solve = XmStringCreateSimple((char *) "(Solve)");
	mode = XmStringCreateSimple((char *) "Mode");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, clear, 'C', NULL, NULL,
		XmVaPUSHBUTTON, undo, 'U', NULL, NULL,
		XmVaPUSHBUTTON, randomize, 'R', NULL, NULL,
		XmVaPUSHBUTTON, solve, 'S', NULL, NULL,
		XmVaPUSHBUTTON, mode, 'M', NULL, NULL,
		NULL);
	XmStringFree(clear);
	XmStringFree(undo);
	XmStringFree(randomize);
	XmStringFree(solve);
	XmStringFree(mode);
	about = XmStringCreateSimple((char *) "About");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "help_menu", 2, helpCB,
		XmVaPUSHBUTTON, about, 'A', NULL, NULL,
		XmVaPUSHBUTTON, help, 'H', NULL, NULL,
		NULL);
	XmStringFree(about);
	XmStringFree(help);
	XtManageChild(menubar);
	panel2 = XtCreateManagedWidget("panel2",
				       xmPanedWindowWidgetClass, panel,
				       NULL, 0);
	panel3 = XtVaCreateManagedWidget("panel3",
					 xmPanedWindowWidgetClass, panel2,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	rowcol = XtVaCreateManagedWidget("Rowcol", xmRowColumnWidgetClass, panel3,
					 XmNnumColumns, 2,
					 XmNorientation, XmHORIZONTAL,
					 XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(rowcol,
		      XmNforeground, &fg,
		      XmNbackground, &bg, NULL);
	mouseLeftCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	      RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_left_bits,
				 mouse_left_width, mouse_left_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	mouseRightCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	     RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_right_bits,
			       mouse_right_width, mouse_right_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	(void) XtVaCreateManagedWidget("mouseLeftText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Move tile", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseLeft",
				xmLabelGadgetClass, rowcol,
	      XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseLeftCursor, NULL);
	(void) XtVaCreateManagedWidget("mouseRightText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Randomize", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseRight",
				xmLabelGadgetClass, rowcol,
	     XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseRightCursor, NULL);
	(void) XtVaCreateManagedWidget("movesText",
				xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Moves", 6, NULL);
	moves = XtVaCreateManagedWidget("0",
					xmLabelWidgetClass, rowcol, NULL);
	(void) XtVaCreateManagedWidget("recordText",
				xmLabelGadgetClass, rowcol,
		 XtVaTypedArg, XmNlabelString, XmRString, "Record", 7, NULL);
	record = XtVaCreateManagedWidget("0",
					 xmLabelWidgetClass, rowcol, NULL);

	rowcol2 = XtVaCreateManagedWidget("Rowcol2",
				       xmRowColumnWidgetClass, panel3,
				       XmNnumColumns, 1,
				       XmNorientation, XmHORIZONTAL,
				       XmNpacking, XmPACK_COLUMN, NULL);
	hexX = XtVaCreateManagedWidget("hexX",
				      xmScaleWidgetClass, rowcol2,
			  XtVaTypedArg, XmNtitleString, XmRString, "Size X", 7,
				      XmNminimum, MINHEXAGONS,
				      XmNmaximum, MAXHEXAGONS,
				      XmNvalue, MINHEXAGONS,
				      XmNshowValue, True,
				      XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(hexX,
		      XmNvalueChangedCallback, (XtCallbackProc) HexXSlider, (XtPointer) NULL);
	hexY = XtVaCreateManagedWidget("hexY",
				      xmScaleWidgetClass, rowcol2,
			  XtVaTypedArg, XmNtitleString, XmRString, "Size Y", 7,
				      XmNminimum, MINHEXAGONS,
				      XmNmaximum, MAXHEXAGONS,
				      XmNvalue, MINHEXAGONS,
				      XmNshowValue, True,
				      XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(hexY,
		      XmNvalueChangedCallback, (XtCallbackProc) HexYSlider, (XtPointer) NULL);
	cornersSwitch = XtVaCreateManagedWidget("Corners",
				  xmToggleButtonWidgetClass, rowcol2, NULL);
	XtAddCallback(cornersSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) CornersToggle, (XtPointer) NULL);
	rowcol3 = XtVaCreateManagedWidget("Rowcol3",
				       xmRowColumnWidgetClass, panel3, NULL);
	message = XtVaCreateManagedWidget("Play Hexagons! (use mouse or keypad)",
					  xmLabelWidgetClass, rowcol3, NULL);
	hexagons = XtCreateManagedWidget("hexagons",
					 hexagonsWidgetClass, panel3, NULL, 0);
#else
        XtSetArg(arg[1], XtNinput, True);
        XtSetValues(topLevel, arg, 2);
        hexagons = XtCreateManagedWidget("hexagons",
				      hexagonsWidgetClass, topLevel, NULL, 0);
#endif
	XtAddCallback(hexagons,
		      XtNselectCallback, (XtCallbackProc) CallbackHexagons, (XtPointer) NULL);
	Initialize(hexagons);
	XtRealizeWidget(topLevel);
	XGrabButton(XtDisplay(hexagons), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(hexagons), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(hexagons),
		    XCreateFontCursor(XtDisplay(hexagons), XC_crosshair));
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
