/*-
# X-BASED MISSING LINK(tm)
#
#  Mlink.c
#
###
#
#  Copyright (c) 1994 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Mlink */

#include <stdio.h>
#include <stdlib.h>
#ifdef VMS
#include <unixlib.h>
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "MlinkP.h"

#ifndef DATAFILE
#define DATAFILE "/usr/games/lib/mlink.dat"
#endif

static void InitializeMlink(Widget request, Widget renew);
static void ExposeMlink(Widget renew, XEvent * event, Region region);
static void ResizeMlink(MlinkWidget w);
static void DestroyMlink(Widget old);
static Boolean SetValuesMlink(Widget current, Widget request, Widget renew);
static void QuitMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkTop(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkLeft(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkRight(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkBottom(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MoveMlinkInput(MlinkWidget w, int x, int direction, int shift, int control);
static void SelectMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void ReleaseMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void RandomizeMlinkMaybe(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void GetMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void WriteMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void ClearMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void UndoMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void SolveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void OrientizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void MiddleMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void EnterMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void LeaveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs);
static void ControlMlink(MlinkWidget w, int direction);
static void RotateMlink(MlinkWidget w, int direction, int tile);
static void SlideMlink(MlinkWidget w, int direction);

static int  SelectTiles(MlinkWidget w, int x, int y, int *i, int *j);
static void SelectSlideTiles(MlinkWidget w, int pos);
static void SetAllColors(MlinkWidget w, Boolean init);
static void GetColor(MlinkWidget w, int face, Boolean init);
static void MoveShiftCb(MlinkWidget w, int direction);
static void CheckTiles(MlinkWidget w);
static void ResetTiles(MlinkWidget w);
static void ResizeTiles(MlinkWidget w);
static void MoveNoTiles(MlinkWidget w);
static Boolean MoveTilesDir(MlinkWidget w, int direction, int tile);
static void GetTiles(MlinkWidget w);
static void WriteTiles(MlinkWidget w);
static void ClearTiles(MlinkWidget w);
static void UndoTiles(MlinkWidget w);
static void RandomizeTiles(MlinkWidget w);
static void OrientizeTiles(MlinkWidget w);
static void MiddleTiles(MlinkWidget w);
static void MoveTiles(MlinkWidget w, int from);
static void RotateTiles(MlinkWidget w, int c, int direction);
static int  ExchangeTiles(MlinkWidget w, int pos1, int pos2);
static void DrawFrame(MlinkWidget w, GC gc);
static void DrawTile(MlinkWidget w, int pos, Boolean blank, Boolean erase, int offset);
static void DrawLink(MlinkWidget w, GC gc, int pos, int offset);
static Pixel brighter(MlinkWidget w, Pixel pixel);
static Pixel darker(MlinkWidget w, Pixel pixel);
static void fill3DRect(MlinkWidget w, GC gc,
 GC darkerGC, GC brighterGC,
 int x, int y, int width, int height, Boolean raised);
static Boolean PositionTile(MlinkWidget w, int x, int *r);
static int  Column(MlinkWidget w, int pos);
static int  Row(MlinkWidget w, int pos);
static int  int2String(char *buf, int number, int base, Boolean capital);

static char defaultTranslationsMlink[] =
"<KeyPress>q: Quit()\n\
   Ctrl<KeyPress>C: Quit()\n\
   <KeyPress>Up: MoveTop()\n\
   <KeyPress>osfUp: MoveTop()\n\
   <KeyPress>KP_Up: MoveTop()\n\
   <KeyPress>KP_8: MoveTop()\n\
   <KeyPress>R8: MoveTop()\n\
   <KeyPress>Left: MoveLeft()\n\
   <KeyPress>osfLeft: MoveLeft()\n\
   <KeyPress>KP_Left: MoveLeft()\n\
   <KeyPress>KP_4: MoveLeft()\n\
   <KeyPress>R10: MoveLeft()\n\
   <KeyPress>Right: MoveRight()\n\
   <KeyPress>osfRight: MoveRight()\n\
   <KeyPress>KP_Right: MoveRight()\n\
   <KeyPress>KP_6: MoveRight()\n\
   <KeyPress>R12: MoveRight()\n\
   <KeyPress>Down: MoveBottom()\n\
   <KeyPress>osfDown: MoveBottom()\n\
   <KeyPress>KP_Down: MoveBottom()\n\
   <KeyPress>KP_2: MoveBottom()\n\
   <KeyPress>R14: MoveBottom()\n\
   <Btn1Down>: Select()\n\
   <Btn1Up>: Release()\n\
   <KeyPress>r: Randomize()\n\
   <Btn3Down>(2+): Randomize()\n\
   <Btn3Down>: RandomizeMaybe()\n\
   <KeyPress>g: Get()\n\
   <KeyPress>w: Write()\n\
   <KeyPress>c: Clear()\n\
   <KeyPress>u: Undo()\n\
   <KeyPress>s: Solve()\n\
   <KeyPress>o: Orientize()\n\
   <KeyPress>m: Middle()\n\
   <EnterWindow>: Enter()\n\
   <LeaveWindow>: Leave()";

static XtActionsRec actionsListMlink[] =
{
	{(char *) "Quit", (XtActionProc) QuitMlink},
	{(char *) "MoveTop", (XtActionProc) MoveMlinkTop},
	{(char *) "MoveLeft", (XtActionProc) MoveMlinkLeft},
	{(char *) "MoveRight", (XtActionProc) MoveMlinkRight},
	{(char *) "MoveBottom", (XtActionProc) MoveMlinkBottom},
	{(char *) "Select", (XtActionProc) SelectMlink},
	{(char *) "Release", (XtActionProc) ReleaseMlink},
	{(char *) "Randomize", (XtActionProc) RandomizeMlink},
	{(char *) "RandomizeMaybe", (XtActionProc) RandomizeMlinkMaybe},
	{(char *) "Get", (XtActionProc) GetMlink},
	{(char *) "Write", (XtActionProc) WriteMlink},
	{(char *) "Clear", (XtActionProc) ClearMlink},
	{(char *) "Undo", (XtActionProc) UndoMlink},
	{(char *) "Solve", (XtActionProc) SolveMlink},
	{(char *) "Orientize", (XtActionProc) OrientizeMlink},
	{(char *) "Middle", (XtActionProc) MiddleMlink},
	{(char *) "Enter", (XtActionProc) EnterMlink},
	{(char *) "Leave", (XtActionProc) LeaveMlink}
};

static XtResource resourcesMlink[] =
{
	{XtNuserName, XtCUserName, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.username),
	 XtRString, (caddr_t) "nobody"},
	{XtNfaceColor0, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[0]),
	 XtRString, (caddr_t) "White"},
	{XtNfaceColor1, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[1]),
	 XtRString, (caddr_t) "Yellow"},
	{XtNfaceColor2, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[2]),
	 XtRString, (caddr_t) "Green"},
	{XtNfaceColor3, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[3]),
	 XtRString, (caddr_t) "Red"},
	{XtNfaceColor4, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[4]),
	 XtRString, (caddr_t) "Blue"},
	{XtNfaceColor5, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[5]),
	 XtRString, (caddr_t) "Magenta"},
	{XtNfaceColor6, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[6]),
	 XtRString, (caddr_t) "Cyan"},
	{XtNfaceColor7, XtCLabel, XtRString, sizeof (String),
	 XtOffset(MlinkWidget, mlink.faceName[7]),
	 XtRString, (caddr_t) "Orange"},
	{XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.foreground),
	 XtRString, (caddr_t) "Cyan" /*XtDefaultForeground*/},
	{XtNbackground, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.background),
	 XtRString, (caddr_t) "NavyBlue" /*XtDefaultBackground*/},
	{XtNtileColor, XtCBackground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.tileColor),
	 XtRString, (caddr_t) "Black" /*XtDefaultForeground*/},
	{XtNtileBorder, XtCForeground, XtRPixel, sizeof (Pixel),
	 XtOffset(MlinkWidget, mlink.borderColor),
	 XtRString, (caddr_t) "gray25" /*XtDefaultBackground*/},
	{XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
	 XtOffset(MlinkWidget, core.width),
	 XtRString, (caddr_t) "200"},
	{XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
	 XtOffset(MlinkWidget, core.height),
	 XtRString, (caddr_t) "200"},
	{XtNtiles, XtCTiles, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.tiles),
	 XtRString, (caddr_t) "4"},	/* DEFAULTTILES */
	{XtNfaces, XtCFaces, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.faces),
	 XtRString, (caddr_t) "4"},	/* DEFAULTFACES */
	{XtNorient, XtCOrient, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.orient),
	 XtRString, (caddr_t) "FALSE"},	/*DEFAULTORIENT */
	{XtNmiddle, XtCMiddle, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.middle),
	 XtRString, (caddr_t) "TRUE"},	/*DEFAULTMIDDLE */
	{XtNmono, XtCMono, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.mono),
	 XtRString, (caddr_t) "FALSE"},
	{XtNreverse, XtCReverse, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.reverse),
	 XtRString, (caddr_t) "FALSE"},
	{XtNbase, XtCBase, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.base),
	 XtRString, (caddr_t) "10"},
	{XtNstart, XtCBoolean, XtRBoolean, sizeof (Boolean),
	 XtOffset(MlinkWidget, mlink.started),
	 XtRString, (caddr_t) "FALSE"},
	{XtNmenu, XtCMenu, XtRInt, sizeof (int),
	 XtOffset(MlinkWidget, mlink.menu),
	 XtRString, (caddr_t) "-1"},
	{XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
	 XtOffset(MlinkWidget, mlink.select),
	 XtRCallback, (caddr_t) NULL}
};

MlinkClassRec mlinkClassRec =
{
	{
		(WidgetClass) & widgetClassRec,		/* superclass */
		(char *) "Mlink",	/* class name */
		sizeof (MlinkRec),	/* widget size */
		NULL,		/* class initialize */
		NULL,		/* class part initialize */
		FALSE,		/* class inited */
		(XtInitProc) InitializeMlink,	/* initialize */
		NULL,		/* initialize hook */
		XtInheritRealize,	/* realize */
		actionsListMlink,	/* actions */
		XtNumber(actionsListMlink),	/* num actions */
		resourcesMlink,	/* resources */
		XtNumber(resourcesMlink),	/* num resources */
		NULLQUARK,	/* xrm class */
		TRUE,		/* compress motion */
		TRUE,		/* compress exposure */
		TRUE,		/* compress enterleave */
		TRUE,		/* visible interest */
		(XtWidgetProc) DestroyMlink,	/* destroy */
		(XtWidgetProc) ResizeMlink,	/* resize */
		(XtExposeProc) ExposeMlink,	/* expose */
		(XtSetValuesFunc) SetValuesMlink,	/* set values */
		NULL,		/* set values hook */
		XtInheritSetValuesAlmost,	/* set values almost */
		NULL,		/* get values hook */
		NULL,		/* accept focus */
		XtVersion,	/* version */
		NULL,		/* callback private */
		defaultTranslationsMlink,	/* tm table */
		NULL,		/* query geometry */
		NULL,		/* display accelerator */
		NULL		/* extension */
	},
	{
		0		/* ignore */
	}
};

WidgetClass mlinkWidgetClass = (WidgetClass) & mlinkClassRec;

static void
InitializeMlink(Widget request, Widget renew)
{
	MlinkWidget w = (MlinkWidget) renew;

	w->mlink.tileOfPosition = NULL;
	CheckTiles(w);
	InitMoves();
	ResetTiles(w);
	(void) SRAND(getpid());
	w->mlink.depth = DefaultDepthOfScreen(XtScreen(w));
	SetAllColors(w, True);
	ResizeMlink(w);
}

static void
DestroyMlink(Widget old)
{
	MlinkWidget w = (MlinkWidget) old;
	int         face;

	for (face = 0; face < MAXFACES; face++)
		XtReleaseGC(old, w->mlink.faceGC[face]);
	XtReleaseGC(old, w->mlink.tileGC);
	XtReleaseGC(old, w->mlink.tileBrighterGC);
	XtReleaseGC(old, w->mlink.tileDarkerGC);
	XtReleaseGC(old, w->mlink.puzzleGC);
	XtReleaseGC(old, w->mlink.inverseGC);
	XtRemoveCallbacks(old, XtNselectCallback, w->mlink.select);
}

static void
ResizeMlink(MlinkWidget w)
{
	w->mlink.delta.x = 3;
	w->mlink.delta.y = 3;
	w->mlink.offset.x = MAX(((int) w->core.width - w->mlink.delta.x) /
				w->mlink.tiles, 0);
	w->mlink.offset.y = MAX(((int) w->core.height - w->mlink.delta.y) /
				w->mlink.faces, 0);
	w->mlink.faceSize.x = w->mlink.offset.x * w->mlink.tiles +
		w->mlink.delta.x;
	w->mlink.faceSize.y = w->mlink.offset.y + w->mlink.delta.y;
	w->mlink.puzzleSize.x = w->mlink.faceSize.x + w->mlink.delta.x;
	w->mlink.puzzleSize.y = (w->mlink.faceSize.y - w->mlink.delta.y) *
		w->mlink.faces + w->mlink.delta.y;
	w->mlink.puzzleOffset.x = ((int) w->core.width -
				   w->mlink.puzzleSize.x + 2) / 2;
	w->mlink.puzzleOffset.y = ((int) w->core.height -
				   w->mlink.puzzleSize.y + 2) / 2;
	w->mlink.tileSize.x = MAX(w->mlink.offset.x - w->mlink.delta.x, 0);
	w->mlink.tileSize.y = MAX(w->mlink.offset.y - w->mlink.delta.y, 0);
	ResizeTiles(w);
}

static void
ExposeMlink(Widget renew, XEvent * event, Region region)
{
	MlinkWidget w = (MlinkWidget) renew;

	if (w->core.visible) {
		XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.inverseGC,
			0, 0, w->core.width, w->core.height);
		DrawFrame(w, (w->mlink.focus) ? w->mlink.puzzleGC :
			w->mlink.inverseGC);
		DrawAllTiles(w);
	}
}

static      Boolean
SetValuesMlink(Widget current, Widget request, Widget renew)
{
	MlinkWidget c = (MlinkWidget) current, w = (MlinkWidget) renew;
	Boolean     redraw = False, setColors = False;
	Boolean     redrawTiles = False;
	int         face;

	CheckTiles(w);
	for (face = 0; face < MAXFACES; face++) {
		if (strcmp(w->mlink.faceName[face], c->mlink.faceName[face])) {
			setColors = True;
			break;
		}
	}
	if (w->mlink.background != c->mlink.background ||
	    w->mlink.foreground != c->mlink.foreground ||
	    w->mlink.borderColor != c->mlink.borderColor ||
	    w->mlink.tileColor != c->mlink.tileColor ||
	    w->mlink.reverse != c->mlink.reverse ||
	    w->mlink.mono != c->mlink.mono ||
	    setColors) {
		SetAllColors(w, False);
		redrawTiles = True;
	}
	if (w->mlink.tiles != c->mlink.tiles ||
	    w->mlink.faces != c->mlink.faces ||
	    w->mlink.orient != c->mlink.orient ||
	    w->mlink.middle != c->mlink.middle ||
	    w->mlink.base != c->mlink.base) {
		ResetTiles(w);
		ResizeMlink(w);
		redraw = True;
	} else if (w->mlink.offset.x != c->mlink.offset.x ||
		   w->mlink.offset.y != c->mlink.offset.y) {
		ResizeMlink(w);
		redraw = True;
	}
	if (w->mlink.menu != -1) {
		switch (w->mlink.menu) {
		case 0:
			w->mlink.menu = -1;
			GetTiles(w);
			break;
		case 1:
			w->mlink.menu = -1;
			WriteTiles(w);
			break;
		case 3:
			w->mlink.menu = -1;
			ClearTiles(w);
			break;
		case 4:
			w->mlink.menu = -1;
			UndoTiles(w);
			break;
		case 5:
			w->mlink.menu = -1;
			RandomizeTiles(w);
			break;
		case 6:
			w->mlink.menu = -1;
#if 0
			SolveTiles(w);	/* Sorry, this is not implemented */
#endif
			break;
		case 7:
			w->mlink.menu = -1;
			OrientizeTiles(w);
			break;
		case 8:
			w->mlink.menu = -1;
			MiddleTiles(w);
			break;
		default:
			w->mlink.menu = -1;
			break;
		}
	}
	if (redrawTiles && !redraw && XtIsRealized(renew) && renew->core.visible) {
		DrawFrame(w, c->mlink.inverseGC);
		if (w->mlink.focus)
			DrawFrame(w, w->mlink.puzzleGC);
		DrawAllTiles(w);
	}
	return (redraw);
}

static void
QuitMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	XtCloseDisplay(XtDisplay(w));
	exit(0);
}

static void
SelectMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	int         i, j, pos, shift;

	pos = SelectTiles(w, event->xbutton.x, event->xbutton.y, &i, &j);
	if (-w->mlink.tileFaces != pos) {
		w->mlink.currentTile = i;
		w->mlink.currentFace = j;
		w->mlink.currentRef = pos;
		shift = (int) (event->xkey.state & (ShiftMask | LockMask));
		if (shift || !CheckSolved(w)) {
			pos = w->mlink.currentTile + w->mlink.currentFace * w->mlink.tiles;
			DrawTile(w, pos, (w->mlink.tileOfPosition[i + j * w->mlink.tiles] <= 0),
				 False, TRUE);
		}
	} else
		w->mlink.currentRef = -w->mlink.tileFaces;
}

static void
ReleaseMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	int         i, j, pos, diff, shift;

	if (w->mlink.currentRef == -w->mlink.tileFaces)
		return;
	pos = w->mlink.currentTile + w->mlink.currentFace * w->mlink.tiles;
	DrawTile(w, pos, (w->mlink.tileOfPosition[pos] <= 0), True, TRUE);
	if (w->mlink.tileOfPosition[pos] > 0) {
		DrawTile(w, pos, False, False, FALSE);
	}
	shift = (int) (event->xkey.state & (ShiftMask | LockMask));
	if (!shift && CheckSolved(w))
		MoveNoTiles(w);
	else {
		pos = SelectTiles(w, event->xbutton.x, event->xbutton.y, &i, &j);
		if (-w->mlink.tileFaces != pos) {
			if (j == w->mlink.currentFace) {
				pos = w->mlink.currentRef;
				if (pos / w->mlink.tiles == 0 &&
				    j == Row(w, w->mlink.spacePosition) && pos != 0) {
					if (shift && CheckSolved(w))
						MoveNoTiles(w);
					else {
						SelectSlideTiles(w, pos);
						w->mlink.currentTile = w->mlink.tiles;
						if (CheckSolved(w)) {
							mlinkCallbackStruct cb;

							cb.reason = MLINK_SOLVED;
							XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
						}
					}
				}
			} else {
				diff = (w->mlink.currentFace - j + w->mlink.faces) % w->mlink.faces;
				if (diff > w->mlink.faces / 2)
					for (i = 0; i < w->mlink.faces - diff; i++)
						(void) MoveMlink(w, BOTTOM, w->mlink.currentTile, shift);
				else
					for (i = 0; i < diff; i++)
						(void) MoveMlink(w, TOP, w->mlink.currentTile, shift);
				if (!shift && CheckSolved(w)) {
					mlinkCallbackStruct cb;

					cb.reason = MLINK_SOLVED;
					XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
				}
			}
		}
	}
}

static void
RandomizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	RandomizeTiles(w);
}

static void
RandomizeMlinkMaybe(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	if (!w->mlink.started)
		RandomizeTiles(w);
}

static void
GetMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	GetTiles(w);
}

static void
WriteMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	WriteTiles(w);
}

static void
ClearMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	ClearTiles(w);
}

static void
UndoMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	UndoTiles(w);
}

static void
SolveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
#if 0
	SolveTiles(w);		/* Sorry, this is not implemented */
#endif
}

static void
OrientizeMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	OrientizeTiles(w);
}

static void
MiddleMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MiddleTiles(w);
}

static void
EnterMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	w->mlink.focus = True;
	DrawFrame(w, w->mlink.puzzleGC);
}

static void
LeaveMlink(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	w->mlink.focus = False;
	DrawFrame(w, w->mlink.inverseGC);
}

static void
MoveMlinkTop(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, TOP,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkLeft(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, LEFT,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkRight(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, RIGHT,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkBottom(MlinkWidget w, XEvent * event, char **args, int nArgs)
{
	MoveMlinkInput(w, event->xbutton.x, BOTTOM,
		       (int) (event->xkey.state & (ShiftMask | LockMask)),
		       (int) (event->xkey.state & ControlMask));
}

static void
MoveMlinkInput(MlinkWidget w, int x, int direction, int shift, int control)
{
	mlinkCallbackStruct cb;
	int         r;

	if (control)
		ControlMlink(w, direction);
	else if (shift && (direction == TOP || direction == BOTTOM)) {
		MoveShiftCb(w, direction);
		PutMove(direction, 0, 1);
	} else if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	} else if (direction == LEFT || direction == RIGHT) {
		SlideMlink(w, direction);
		if (CheckSolved(w)) {
			cb.reason = MLINK_SOLVED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	} else {
		if (!PositionTile(w, x, &r))
			return;
		RotateMlink(w, direction, r);
		if (CheckSolved(w)) {
			cb.reason = MLINK_SOLVED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	}
}

Boolean
MoveMlink(MlinkWidget w, int direction, int tile, int shift)
{
	mlinkCallbackStruct cb;

	if (shift && (direction == TOP || direction == BOTTOM)) {
		MoveShiftCb(w, direction);
		return True;
	} else if (MoveTilesDir(w, direction, tile)) {
		cb.reason = MLINK_MOVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		PutMove(direction, tile, 0);
		return True;
	}
	return False;
}

static void
ControlMlink(MlinkWidget w, int direction)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_IGNORE;
	switch (direction) {
		case TOP:
			if (w->mlink.faces <= MINFACES)
				return;
			cb.reason = MLINK_DEC_Y;
			break;
		case RIGHT:
			cb.reason = MLINK_INC_X;
			break;
		case BOTTOM:
			if (w->mlink.faces >= MAXFACES)
				return;
			cb.reason = MLINK_INC_Y;
			break;
		case LEFT:
			if (w->mlink.tiles <= MINTILES)
				return;
			cb.reason = MLINK_DEC_X;
			break;
		default:
			(void) printf("ControlMlink: direction %d\n", direction);
	}
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
RotateMlink(MlinkWidget w, int direction, int tile)
{
	mlinkCallbackStruct cb;

	if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	}
	(void) MoveMlink(w, direction, tile, FALSE);
	if (CheckSolved(w)) {
		cb.reason = MLINK_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	return;
}

static void
SlideMlink(MlinkWidget w, int direction)
{
	mlinkCallbackStruct cb;

	if (CheckSolved(w)) {
		MoveNoTiles(w);
		return;
	}
	if (!MoveMlink(w, direction, 0, FALSE)) {
		cb.reason = MLINK_BLOCKED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		return;
	}
	if (CheckSolved(w)) {
		cb.reason = MLINK_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	return;
}

static int
SelectTiles(MlinkWidget w, int x, int y, int *i, int *j)
{
	*i = (x - w->mlink.delta.x / 2 - w->mlink.puzzleOffset.x) /
		w->mlink.offset.x;
	*j = ((y - w->mlink.delta.y / 2 - w->mlink.puzzleOffset.y) %
	      (w->mlink.faces * w->mlink.offset.y + w->mlink.delta.y - 1)) /
		w->mlink.offset.y;
	if (*i >= 0 && *j >= 0 &&
	    *i < w->mlink.tiles && *j < w->mlink.faces)
		return (*i + w->mlink.tiles * *j -
			w->mlink.spacePosition % w->mlink.tileFaces);
	return -w->mlink.tileFaces;
}

static void
SelectSlideTiles(MlinkWidget w, int pos)
{
	mlinkCallbackStruct cb;
	int         n;

	if (pos < 0) {
		for (n = 1; n <= -pos % w->mlink.tiles; n++) {
			MoveTiles(w, w->mlink.spacePosition - 1);
			cb.reason = MLINK_MOVED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			PutMove(RIGHT, 0, 0);
		}
	} else if (pos > 0) {
		for (n = 1; n <= pos % w->mlink.tiles; n++) {
			MoveTiles(w, w->mlink.spacePosition + 1);
			cb.reason = MLINK_MOVED;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			PutMove(LEFT, 0, 0);
		}
	}
}

#define FACTOR 0.7
#define MAXINTENSITY 0xFFFF

static Pixel
brighter(MlinkWidget w, Pixel pixel)
{
	XColor color;
	int i = (int) ((1 - FACTOR) * MAXINTENSITY);

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormap(XtDisplay(w), XtWindow(w)),
		&color);
	if (color.red < i)
		color.red = i;
	if (color.green < i)
		color.green = i;
	if (color.blue < i)
		color.blue = i;
	color.red = (unsigned short) MIN(color.red / FACTOR, MAXINTENSITY);
	color.green = (unsigned short) MIN(color.green / FACTOR, MAXINTENSITY);
	color.blue = (unsigned short) MIN(color.blue / FACTOR, MAXINTENSITY);
	if (XAllocColor(XtDisplay(w),
			DefaultColormap(XtDisplay(w), XtWindow(w)),
			&color))
		return color.pixel;
	return pixel;
}

static Pixel
darker(MlinkWidget w, Pixel pixel)
{
	XColor color;

	color.pixel = pixel;
	XQueryColor(XtDisplay(w), DefaultColormap(XtDisplay(w), XtWindow(w)),
		&color);
	color.red = (unsigned short) (color.red * FACTOR);
	color.green = (unsigned short) (color.green * FACTOR);
	color.blue = (unsigned short) (color.blue * FACTOR);
	if (XAllocColor(XtDisplay(w),
			DefaultColormap(XtDisplay(w), XtWindow(w)),
			&color))
		return color.pixel;
	return pixel;
}

static void
SetAllColors(MlinkWidget w, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	int         face;

	valueMask = GCForeground | GCBackground;

	if (w->mlink.reverse) {
		values.foreground = w->mlink.foreground;
		values.background = w->mlink.background;
	} else {
		values.foreground = w->mlink.background;
		values.background = w->mlink.foreground;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.inverseGC);
	w->mlink.inverseGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.reverse) {
		values.foreground = w->mlink.background;
		values.background = w->mlink.foreground;
	} else {
		values.foreground = w->mlink.foreground;
		values.background = w->mlink.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.puzzleGC);
	w->mlink.puzzleGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth < 2 || w->mlink.mono) {
		if (w->mlink.reverse) {
			values.background = w->mlink.foreground;
			values.foreground = w->mlink.background;
		} else {
			values.foreground = w->mlink.foreground;
			values.background = w->mlink.background;
		}
	} else {
		values.foreground = w->mlink.tileColor;
		values.background = w->mlink.borderColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.tileGC);
	w->mlink.tileGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth >= 2 && !w->mlink.mono) {
		values.foreground = darker(w, w->mlink.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.tileBrighterGC);
	w->mlink.tileBrighterGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth >= 2 && !w->mlink.mono) {
		values.foreground = brighter(w, w->mlink.tileColor);
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.tileDarkerGC);
	w->mlink.tileDarkerGC = XtGetGC((Widget) w, valueMask, &values);
	if (w->mlink.depth < 2 || w->mlink.mono) {
		if (w->mlink.reverse) {
			values.foreground = w->mlink.foreground;
			values.background = w->mlink.background;
		} else {
			values.background = w->mlink.foreground;
			values.foreground = w->mlink.background;
		}
	} else {
		values.foreground = w->mlink.borderColor;
		values.background = w->mlink.tileColor;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.borderGC);
	w->mlink.borderGC = XtGetGC((Widget) w, valueMask, &values);
	for (face = 0; face < MAXFACES; face++)
		GetColor(w, face, init);
}

static void
GetColor(MlinkWidget w, int face, Boolean init)
{
	XGCValues   values;
	XtGCMask    valueMask;
	XColor      colorCell, rgb;

	valueMask = GCForeground | GCBackground;
	if (w->mlink.reverse) {
		values.background = w->mlink.foreground;
	} else {
		values.background = w->mlink.background;
	}
	if (w->mlink.depth > 1 && !w->mlink.mono) {
		if (XAllocNamedColor(XtDisplay(w),
				  DefaultColormap(XtDisplay(w), XtWindow(w)),
				w->mlink.faceName[face], &colorCell, &rgb)) {
			values.foreground = w->mlink.faceColor[face] = colorCell.pixel;
			if (!init)
				XtReleaseGC((Widget) w, w->mlink.faceGC[face]);
			w->mlink.faceGC[face] = XtGetGC((Widget) w, valueMask, &values);
			return;
		} else {
			char        buf[121];

			(void) sprintf(buf, "Color name \"%s\" is not defined",
				       w->mlink.faceName[face]);
			XtWarning(buf);
		}
	}
	if (w->mlink.reverse) {
		values.background = w->mlink.background;
		values.foreground = w->mlink.foreground;
	} else {
		values.background = w->mlink.foreground;
		values.foreground = w->mlink.background;
	}
	if (!init)
		XtReleaseGC((Widget) w, w->mlink.faceGC[face]);
	w->mlink.faceGC[face] = XtGetGC((Widget) w, valueMask, &values);
}

static void
MoveShiftCb(MlinkWidget w, int direction)
{
	mlinkCallbackStruct cb;

	if (w->mlink.middle) {
		(void) MoveTilesDir(w, direction, 0);
		(void) MoveTilesDir(w, direction, 1);
		(void) MoveTilesDir(w, direction, w->mlink.tiles - 1);
	} else {
		int         r;

		for (r = 0; r < w->mlink.tiles; r++)
			(void) MoveTilesDir(w, direction, r);
	}
	cb.reason = MLINK_CONTROL;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
CheckTiles(MlinkWidget w)
{
	char        buf[121];

	if (w->mlink.tiles < MINTILES) {
		(void) sprintf(buf,
			       "Number of Mlink in X direction out of bounds, use %d..MAXINT",
			       MINTILES);
		XtWarning(buf);
		w->mlink.tiles = DEFAULTTILES;
	}
	if (w->mlink.faces < MINFACES || w->mlink.faces > MAXFACES) {
		(void) sprintf(buf,
		  "Number of Mlink in Y direction out of bounds, use %d..%d",
			       MINFACES, MAXFACES);
		XtWarning(buf);
		w->mlink.faces = DEFAULTFACES;
	}
	if (w->mlink.base > 36) {
		/* 10 numbers + 26 letters (ASCII or EBCDIC) */
		XtWarning("Base must be less than or equal to 36");
		w->mlink.base = 10;
	} else if (w->mlink.base <= 1) {	/* Base 1 is rediculous :) */
		XtWarning("Base must be greater than 1");
		w->mlink.base = 10;
	}
}

static void
ResetTiles(MlinkWidget w)
{
	int         i;

	w->mlink.tileFaces = w->mlink.tiles * w->mlink.faces;
	if (w->mlink.tileOfPosition)
		free(w->mlink.tileOfPosition);
	if (!(w->mlink.tileOfPosition = (int *)
	      malloc(sizeof (int) * w->mlink.tileFaces)))
		            XtError("Not enough memory, exiting.");

	if (startPosition)
		free(startPosition);
	if (!(startPosition = (int *)
	      malloc(sizeof (int) * w->mlink.tileFaces)))
		            XtError("Not enough memory, exiting.");

	w->mlink.spacePosition = w->mlink.tileFaces - 1;
	w->mlink.tileOfPosition[w->mlink.tileFaces - 1] = 0;
	for (i = 1; i < w->mlink.tileFaces; i++)
		w->mlink.tileOfPosition[i - 1] = i;
	FlushMoves(w);
	w->mlink.started = False;
}

static void
ResizeTiles(MlinkWidget w)
{
	w->mlink.digitOffset.x = 3;
	w->mlink.digitOffset.y = 4;
}

static void
MoveNoTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_IGNORE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static      Boolean
MoveTilesDir(MlinkWidget w, int direction, int tile)
{
	switch (direction) {
	case TOP:
		RotateTiles(w, tile, TOP);
		return True;
	case RIGHT:
		if (Column(w, w->mlink.spacePosition) > 0) {
			MoveTiles(w, w->mlink.spacePosition - 1);
			return True;
		}
		break;
	case BOTTOM:
		RotateTiles(w, tile, BOTTOM);
		return True;
	case LEFT:
		if (Column(w, w->mlink.spacePosition) < w->mlink.tiles - 1) {
			MoveTiles(w, w->mlink.spacePosition + 1);
			return True;
		}
		break;
	default:
		(void) printf("MoveTilesDir: direction %d\n", direction);
	}
	return False;
}

static void
GetTiles(MlinkWidget w)
{
	FILE       *fp;
	int         c, i, tiles, faces, middle, orient, moves;
	mlinkCallbackStruct cb;

	if ((fp = fopen(DATAFILE, "r")) == NULL)
		(void) printf("Can not read %s for get.\n", DATAFILE);
	else {
		FlushMoves(w);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &tiles);
		if (tiles >= MINTILES) {
			for (i = w->mlink.tiles; i < tiles; i++) {
				cb.reason = MLINK_INC_X;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
			for (i = w->mlink.tiles; i > tiles; i--) {
				cb.reason = MLINK_DEC_X;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: tiles %d should be between %d and MAXINT\n",
				      DATAFILE, tiles, MINTILES);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &faces);
		if (faces >= MINFACES && faces <= MAXFACES) {
			for (i = w->mlink.faces; i < faces; i++) {
				cb.reason = MLINK_INC_Y;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
			for (i = w->mlink.faces; i > faces; i--) {
				cb.reason = MLINK_DEC_Y;
				XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
			}
		} else
			(void) printf("%s corrupted: faces %d should be between %d and %d\n",
				      DATAFILE, faces, MINFACES, MAXFACES);
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &middle);
		if (w->mlink.middle != (Boolean) middle) {
			cb.reason = MLINK_MIDDLE;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &orient);
		if (w->mlink.orient != (Boolean) orient) {
			cb.reason = MLINK_ORIENT;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
		while ((c = getc(fp)) != EOF && c != SYMBOL);
		(void) fscanf(fp, "%d", &moves);
		ScanStartPosition(fp, w);
		cb.reason = MLINK_RESTORE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		SetStartPosition(w);
		ScanMoves(fp, w, moves);
		(void) fclose(fp);
		(void) printf("%s: tiles %d, faces %d, middle %d, orient %d, moves %d.\n",
			      DATAFILE, tiles, faces, middle, orient, moves);
	}
}

static void
WriteTiles(MlinkWidget w)
{
	FILE       *fp;

	if ((fp = fopen(DATAFILE, "w")) == NULL)
		(void) printf("Can not write to %s.\n", DATAFILE);
	else {
		(void) fprintf(fp, "tiles%c %d\n", SYMBOL, w->mlink.tiles);
		(void) fprintf(fp, "faces%c %d\n", SYMBOL, w->mlink.faces);
		(void) fprintf(fp, "middle%c %d\n", SYMBOL, (w->mlink.middle) ? 1 : 0);
		(void) fprintf(fp, "orient%c %d\n", SYMBOL, (w->mlink.orient) ? 1 : 0);
		(void) fprintf(fp, "moves%c %d\n", SYMBOL, NumMoves());
		PrintStartPosition(fp, w);
		PrintMoves(fp);
		(void) fclose(fp);
		(void) printf("Saved to %s.\n", DATAFILE);
	}
}

static void
ClearTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	ResetTiles(w);
	DrawAllTiles(w);
	cb.reason = MLINK_RESET;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
UndoTiles(MlinkWidget w)
{
	if (MadeMoves()) {
		int         direction, tile, shift;
		mlinkCallbackStruct cb;

		GetMove(&direction, &tile, &shift);
		direction = (direction + COORD / 2) % COORD;
		if (shift && (direction == TOP || direction == BOTTOM))
			MoveShiftCb(w, direction);
		else if (MoveTilesDir(w, direction, tile)) {
			cb.reason = MLINK_UNDO;
			XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
		}
	}
}

static void
RandomizeTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	/* First interchange tiles an even number of times */
	if (w->mlink.tiles > 1 && w->mlink.faces > 1 && w->mlink.tileFaces > 4) {
		int         pos, count = 0;

		for (pos = 0; pos < w->mlink.tileFaces; pos++) {
			int         randomPos = pos;

			while (randomPos == pos)
				randomPos = NRAND(w->mlink.tileFaces);
			count += ExchangeTiles(w, pos, randomPos);
		}
		if (count % 2 &&
		    !ExchangeTiles(w, 0, 1) &&
		    !ExchangeTiles(w, w->mlink.tileFaces - 2,
		               w->mlink.tileFaces - 1))
			(void) printf("RandomizeTiles: should not get here\n");
		DrawAllTiles(w);
	}
	/* Now move the space around randomly */
	if (w->mlink.tiles > 1 || w->mlink.faces > 1) {
		int         big = w->mlink.tileFaces + NRAND(2);
		int         lastDirection = -1;
		int         randomDirection;

		cb.reason = MLINK_RESET;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);

#ifdef DEBUG
		big = 3;
#endif
		if (big > 100)
			big = 100;
		while (big--) {
			randomDirection = NRAND(COORD);

#ifdef DEBUG
			sleep(1);
#endif

			if ((randomDirection + COORD / 2) % COORD != lastDirection ||
			    w->mlink.tiles == 1 || w->mlink.faces == 1) {
				if (MoveMlink(w, randomDirection, NRAND(w->mlink.tiles), FALSE))
					lastDirection = randomDirection;
				else
					big++;
			}
		}
		FlushMoves(w);
		cb.reason = MLINK_RANDOMIZE;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
	if (CheckSolved(w)) {
		cb.reason = MLINK_SOLVED;
		XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
	}
}

static void
OrientizeTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_ORIENT;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
MiddleTiles(MlinkWidget w)
{
	mlinkCallbackStruct cb;

	cb.reason = MLINK_MIDDLE;
	XtCallCallbacks((Widget) w, XtNselectCallback, &cb);
}

static void
MoveTiles(MlinkWidget w, int from)
{
	int         tempTile;

	tempTile = w->mlink.tileOfPosition[from];
	w->mlink.tileOfPosition[from] =
		w->mlink.tileOfPosition[w->mlink.spacePosition];
	w->mlink.tileOfPosition[w->mlink.spacePosition] = tempTile;
	DrawTile(w, w->mlink.spacePosition, False, False, FALSE);
	w->mlink.spacePosition = from;
	DrawTile(w, w->mlink.spacePosition, True, True, FALSE);
}

static void
RotateTiles(MlinkWidget w, int c, int direction)
{
	int         newR, k, tempTile, currTile, pos, newPos;
	int         r = 0, space = -1, loop = FALSE;

	if (w->mlink.middle && c > 0 && c < w->mlink.tiles - 1) {
		loop = TRUE;
		c = 1;
	}
	do {
		currTile = w->mlink.tileOfPosition[r * w->mlink.tiles + c];
		for (k = 1; k <= w->mlink.faces; k++) {
			newR = (direction == TOP) ? (r + w->mlink.faces - 1) % w->mlink.faces :
				(r + 1) % w->mlink.faces;
			pos = r * w->mlink.tiles + c;
			newPos = newR * w->mlink.tiles + c;
			tempTile = w->mlink.tileOfPosition[newPos];
			w->mlink.tileOfPosition[newPos] = currTile;
			currTile = tempTile;
			if (w->mlink.spacePosition == pos) {
				DrawTile(w, newPos, True, True, FALSE);
				space = newPos;
			} else
				DrawTile(w, newPos, False, False, FALSE);
			r = newR;
		}
		if (space >= 0)
			w->mlink.spacePosition = space;
	} while (loop && ++c < w->mlink.tiles - 1);
}

static int
ExchangeTiles(MlinkWidget w, int pos1, int pos2)
{
	int         tempTile;

	if (w->mlink.tileOfPosition[pos1] <= 0)
		return FALSE;
	else if (w->mlink.tileOfPosition[pos2] <= 0)
		return FALSE;
	tempTile = w->mlink.tileOfPosition[pos1];
	w->mlink.tileOfPosition[pos1] = w->mlink.tileOfPosition[pos2];
	w->mlink.tileOfPosition[pos2] = tempTile;
	return TRUE;
}

static void
DrawFrame(MlinkWidget w, GC gc)
{
	int         sumX, sumY, offsetX, offsetY, r;

	offsetX = w->mlink.puzzleOffset.x;
	offsetY = w->mlink.puzzleOffset.y;
	sumX = w->mlink.tiles * w->mlink.offset.x + w->mlink.delta.x / 2 + 1;
	sumY = w->mlink.offset.y;
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		0, 0, w->core.width, offsetY);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		0, w->core.height - offsetY - 1,
		w->core.width, offsetY + 2);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		0, 0, offsetX, w->core.height);
	XFillRectangle(XtDisplay(w), XtWindow(w), w->mlink.borderGC,
		offsetX + sumX, 0,
		w->core.width - sumX - offsetX, w->core.height);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		offsetX, offsetY, sumX, 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		offsetX, offsetY, 1, w->core.height - 2 * offsetY - 1);
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
		sumX + offsetX, offsetY, 1, w->core.height - 2 * offsetY - 1);
	for (r = 1; r < w->mlink.faces; r++) {
		offsetY += sumY;
		XFillRectangle(XtDisplay(w), XtWindow(w), gc,
			       offsetX, offsetY, sumX, 1);
	}
	XFillRectangle(XtDisplay(w), XtWindow(w), gc,
	       offsetX, offsetY + sumY, sumX + 1, 1);
}

void
DrawAllTiles(MlinkWidget w)
{
	int         k;

	for (k = 0; k < w->mlink.tileFaces; k++)
		DrawTile(w, k, (w->mlink.tileOfPosition[k] <= 0),
			 (w->mlink.tileOfPosition[k] <= 0), FALSE);
}

static void
DrawTile(MlinkWidget w, int pos, Boolean blank, Boolean erase, int offsetPosition)
{
	int         dx, dy, tile = w->mlink.tileOfPosition[pos];
	GC          linkGC, tileGC, borderGC;

	if (erase) {
		tileGC = w->mlink.inverseGC;
		borderGC = w->mlink.inverseGC;
	} else if (offsetPosition) {
		tileGC = w->mlink.borderGC;
		borderGC = w->mlink.tileGC;
	} else {
		tileGC = w->mlink.tileGC;
		borderGC = w->mlink.borderGC;
	}
	if ((w->mlink.depth < 2 || w->mlink.mono) && offsetPosition) {
		linkGC = borderGC;
	} else {
		int         face = ((tile - 1) / w->mlink.tiles + 1) % w->mlink.faces;

		linkGC = w->mlink.faceGC[face];
		if (w->mlink.depth >= 2 && !w->mlink.mono) {
			if (((w->mlink.faceColor[face] == w->mlink.borderColor) && offsetPosition) ||
			    ((w->mlink.faceColor[face] == w->mlink.tileColor) && !offsetPosition))
				linkGC = borderGC;
		}
	}
	dx = Column(w, pos) * w->mlink.offset.x + w->mlink.delta.x +
		w->mlink.puzzleOffset.x + offsetPosition - 1;
	dy = Row(w, pos) * w->mlink.offset.y + w->mlink.delta.y +
		w->mlink.puzzleOffset.y + offsetPosition - 1;
	if (blank)
		XFillRectangle(XtDisplay(w), XtWindow(w), tileGC, dx, dy,
		       w->mlink.tileSize.x + 1, w->mlink.tileSize.y - 1);

	else
		fill3DRect(w, tileGC, w->mlink.tileDarkerGC, w->mlink.tileBrighterGC,
                        dx, dy, w->mlink.tileSize.x + 1, w->mlink.tileSize.y - 1,
                        offsetPosition == 0);
		/* XFillRectangle(XtDisplay(w), XtWindow(w), tileGC,
		       dx + 1, dy + 1,
		       w->mlink.tileSize.x, w->mlink.tileSize.y - 2); */
	if (!blank) {
		int         i = 0, offsetX = 0, t = tile;
		char        buf[121];

		buf[0] = '\0';
		if (w->mlink.orient || w->mlink.depth < 2 || w->mlink.mono) {
			if (w->mlink.depth < 2 || w->mlink.mono) {
				if (w->mlink.orient) {
					int         last;

					last = int2String(buf, tile, w->mlink.base, False);
					buf[last] = w->mlink.faceName[((tile - 1) / w->mlink.tiles + 1) %
							  w->mlink.faces][0];
					buf[last + 1] = '\0';
					t *= w->mlink.base;
				} else {
					buf[0] = w->mlink.faceName[((tile - 1) /
								    w->mlink.tiles + 1) % w->mlink.faces][0];
					buf[1] = '\0';
					t = 1;
				}
			} else {
				(void) int2String(buf, tile, w->mlink.base, True);
			}
			while (t >= 1) {
				t /= w->mlink.base;
				offsetX += w->mlink.digitOffset.x;
				i++;
			}
			(void) XDrawString(XtDisplay(w), XtWindow(w), linkGC,
				    dx + w->mlink.tileSize.x / 2 - offsetX,
				    dy + w->mlink.tileSize.y / 2 + w->mlink.digitOffset.y, buf, i);
		}
	}
	if (tile != 0 &&
	    ((tile != w->mlink.tileFaces - 1 && w->mlink.tiles > 1) ||
	     w->mlink.tiles > 2))
		DrawLink(w, linkGC, pos, offsetPosition);
	XDrawRectangle(XtDisplay(w), XtWindow(w), borderGC, dx, dy,
		       w->mlink.tileSize.x + 1, w->mlink.tileSize.y - 1);
}

#define MULT 64
#define THICKNESS 7
static void
DrawLink(MlinkWidget w, GC gc, int pos, int offset)
{
	int         dx, dy, sizex, sizey, tile = w->mlink.tileOfPosition[pos],
	            thick;

	sizex = w->mlink.offset.x * 3 / 2 - 3;
	sizey = w->mlink.offset.y * 3 / 4 - 3;
	dx = Column(w, pos) * w->mlink.offset.x + w->mlink.delta.x +
		w->mlink.puzzleOffset.x - sizex / 2 + offset;
	dy = Row(w, pos) * w->mlink.offset.y + w->mlink.delta.y +
		w->mlink.puzzleOffset.y + w->mlink.tileSize.y / 2 - sizey / 2 - 1 + offset;
	thick = MIN(sizey, sizex) / 3;
	thick = MIN(thick, THICKNESS);
	if (thick > 1) {
		XSetLineAttributes(XtDisplay(w), gc, thick,
				   LineSolid, CapNotLast, JoinRound);
		if (tile % w->mlink.tiles == 0 || tile == w->mlink.tileFaces - 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -179 * MULT);
		else if (tile % w->mlink.tiles == 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx + w->mlink.tileSize.x - 1, dy,
				 sizex, sizey, 90 * MULT, 180 * MULT);
		else {
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -124 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, 30 * MULT);
			dx += w->mlink.tileSize.x - 1;
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 90 * MULT, 30 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, -125 * MULT);
		}
		XSetLineAttributes(XtDisplay(w), gc, 1, LineSolid, CapNotLast, JoinRound);
	}
}

#if 0
#define MULT 64
#define THICKNESS 8
static void
DrawLink(w, gc, pos, offset)
	MlinkWidget w;
	GC          gc;
	int         pos;
{
	int         dx, dy, sizex, sizey, i, tile = w->mlink.tileOfPosition[pos];

	for (i = 0; i < THICKNESS; i++) {
		sizex = w->mlink.offset.x * 3 / 2 - i;
		sizey = w->mlink.offset.y * 3 / 4 - i;
		dx = Column(w, pos) * w->mlink.offset.x + w->mlink.delta.x +
			w->mlink.puzzleOffset.x - sizex / 2 + offset;
		dy = Row(w, pos) * w->mlink.offset.y + w->mlink.delta.y +
			w->mlink.puzzleOffset.y + w->mlink.tileSize.y / 2 - sizey / 2 + offset;
		if (tile % w->mlink.tiles == 0 || tile == w->mlink.tileFaces - 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -179 * MULT);
		else if (tile % w->mlink.tiles == 1)
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx + w->mlink.tileSize.x - 1, dy,
				 sizex, sizey, 90 * MULT, 180 * MULT);
		else {
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 89 * MULT, -32 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, 128 * MULT);
			dx += w->mlink.tileSize.x - 1;
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, 90 * MULT, 128 * MULT);
			XDrawArc(XtDisplay(w), XtWindow(w), gc, dx, dy,
				 sizex, sizey, -90 * MULT, -33 * MULT);
		}
	}
}
#endif

static void
fill3DRect(MlinkWidget w, GC gc, GC darkerGC, GC brighterGC,
         int x, int y, int width, int height, Boolean raised)
{
        GC currentGC = (raised) ? gc : darkerGC;

        if (width > 2 && height > 2)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + 1, y + 1, width - 2, height - 2);
        currentGC = (raised) ? brighterGC : darkerGC;
        XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                x, y, 1, height);
        if (width > 1)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + 1, y, width - 2, 1);
        currentGC = (raised) ? darkerGC : gc;
        if (width > 1 && height > 1)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + 1, y + height - 1, width - 1, 1);
        if (width > 1 && height > 1)
                XFillRectangle(XtDisplay(w), XtWindow(w), currentGC,
                        x + width - 1, y, 1, height - 1);
}

static      Boolean
PositionTile(MlinkWidget w, int x, int *r)
{
	*r = (x - w->mlink.delta.x / 2 - w->mlink.puzzleOffset.x) / w->mlink.offset.x;
	if (*r < 0 || *r >= w->mlink.tiles)
		return False;
	return True;
}

static int
Column(MlinkWidget w, int pos)
{
	return (pos % w->mlink.tiles);
}

static int
Row(MlinkWidget w, int pos)
{
	return (pos / w->mlink.tiles);
}

static int
int2String(char *buf, int number, int base, Boolean capital)
{
	int         digit, mult = base, last, position;
	int         a, i, j, s, r;

	if (capital) {
		a = 'A', i = 'I', j = 'J', s = 'S', r = 'R';
	} else {
		a = 'a', i = 'i', j = 'j', s = 's', r = 'r';
	}
	if (number < 0) {
		(void) printf("number %d < 0\n", number);
		return 0;
	}
	last = 1;
	while (number >= mult) {
		last++;
		mult *= base;
	}
	for (position = 0; position < last; position++) {
		mult /= base;
		digit = number / mult;
		number -= digit * mult;
		buf[position] = digit + '0';
		if (buf[position] > '9') {	/* ASCII */
			buf[position] += (a - '9' - 1);
		} else if (buf[position] < '0') {	/* EBCDIC */
			buf[position] += (a - '9' - 1);
			if (buf[position] > i)
				buf[position] += (j - i - 1);
			if (buf[position] > r)
				buf[position] += (s - r - 1);
		}
	}
	buf[last] = '\0';
	return last;
}

Boolean
CheckSolved(MlinkWidget w)
{
	int         i, j, firstTile, lastTile, midTile;

	if (w->mlink.tiles < 2)
		return True;
	if (w->mlink.orient) {
		firstTile = w->mlink.tileOfPosition[0];
		if (firstTile != 0 && (firstTile - 1) % w->mlink.tiles == 0) {
			for (i = 1; i < w->mlink.tileFaces; i++) {
				midTile = w->mlink.tileOfPosition[i];
				if (midTile && (midTile - firstTile + w->mlink.tileFaces) %
				    w->mlink.tileFaces != i)
					return False;
			}
		} else
			return False;
	} else {
		for (i = 0; i < w->mlink.faces; i++) {
			firstTile = w->mlink.tileOfPosition[i * w->mlink.tiles];
			if (firstTile == 0)
				firstTile = w->mlink.tileOfPosition[i * w->mlink.tiles + 1];
			lastTile = w->mlink.tileOfPosition[(i + 1) * w->mlink.tiles - 1];
			if (lastTile == 0)
				lastTile = w->mlink.tileOfPosition[(i + 1) * w->mlink.tiles - 2];
			if (firstTile % w->mlink.tiles != 1 ||
			    (lastTile % w->mlink.tiles != 0 &&
			     lastTile != w->mlink.tileFaces - 1) ||
			    (lastTile - 1) / w->mlink.tiles !=
			    (firstTile - 1) / w->mlink.tiles)
				return False;
			for (j = 1; j < w->mlink.tiles - 1; j++) {
				midTile = w->mlink.tileOfPosition[i * w->mlink.tiles + j];
				if ((midTile - 1) / w->mlink.tiles !=
				    (firstTile - 1) / w->mlink.tiles || midTile == 0)
					return False;
			}
		}
	}
	return True;
}
