/*-
# MOTIF/X-BASED OCTAHEDRON
#
#  xoct.c
#
###
#
#  Copyright (c) 1993 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "useful",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 5: 95/10/04 Xt/Motif
  Version 4: 94/05/31 Xt
  Version 3: 93/04/01 Motif
  Version 2: 92/02/01 XView
  Version 1: 91/06/10 SunView
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifdef VMS
#include <unixlib.h>
#define getlogin() cuserid(NULL)
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#if HAVE_FCNTL_H
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include "mouse-l.xbm"
#include "mouse-r.xbm"
#endif
#include "Oct.h"
#include "oct.xbm"

#ifndef SCOREFILE
#define SCOREFILE "/usr/games/lib/oct.scores"
#endif

/* The following are in OctP.h also */
#define MINOCTAS 1
#define PERIOD3 3
#define PERIOD4 4
#define BOTH 5
#define MAXMODES 3

#define MAXOCTAS 6
#define NEVER -1
#define FILENAMELEN 1024
#define USERNAMELEN 128
#define NOACCESS "noaccess"
#define NOBODY "nobody"

typedef struct {
	int         score;
	char        name[USERNAMELEN];
} GameRecord;

#ifdef HAVE_MOTIF
static Widget moves, record, message, modes[MAXMODES], orientSwitch,
	stickySwitch, practiceSwitch, octas;
static char buff[16];
static const char *modeString[] =
{
	"Period 3", "Period 4", "Both"
};
#else
static char titleDsp[256] = "";
#endif

static Widget oct;
static Arg  arg[5];
static GameRecord octRecord[MAXMODES][2][MAXOCTAS - MINOCTAS + 2];
static int  movesDsp = 0;
static char *progDsp;
static char messageDsp[128] = "Welcome";
static int  oldSize;
static char recordDsp[16] = "NOT RECORDED";
static char usernameDsp[USERNAMELEN] = "";

static void
Usage(char * programName)
{
	(void) fprintf(stderr, "usage: %s\n", programName);
	(void) fprintf(stderr,
	     "\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
	(void) fprintf(stderr,
	   "\t[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n");
	(void) fprintf(stderr,
		"\t[-{foreground|fg} {color}] [-{background|bg} {color}]\n");
	(void) fprintf(stderr,
	      "\t[-{border|bd} {color}] [-face{0|1|2|3|4|5|6|7} {color}]\n");
	(void) fprintf(stderr,
		       "\t[-{size {int} | sticky}] [-{mode {int} | both}]\n");
	(void) fprintf(stderr,
		   "\t[-[no]orient] [-[no]practice] [-username {string}]\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*oct.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*oct.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*oct.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*oct.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*oct.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*oct.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*oct.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*oct.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*oct.pieceBorder", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*oct.pieceBorder", XrmoptionSepArg, NULL},
	{(char *) "-face0", (char *) "*oct.faceColor0", XrmoptionSepArg, NULL},
	{(char *) "-face1", (char *) "*oct.faceColor1", XrmoptionSepArg, NULL},
	{(char *) "-face2", (char *) "*oct.faceColor2", XrmoptionSepArg, NULL},
	{(char *) "-face3", (char *) "*oct.faceColor3", XrmoptionSepArg, NULL},
	{(char *) "-face4", (char *) "*oct.faceColor4", XrmoptionSepArg, NULL},
	{(char *) "-face5", (char *) "*oct.faceColor5", XrmoptionSepArg, NULL},
	{(char *) "-face6", (char *) "*oct.faceColor6", XrmoptionSepArg, NULL},
	{(char *) "-face7", (char *) "*oct.faceColor7", XrmoptionSepArg, NULL},
	{(char *) "-size", (char *) "*oct.size", XrmoptionSepArg, NULL},
	{(char *) "-sticky", (char *) "*oct.sticky", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-mode", (char *) "*oct.mode", XrmoptionSepArg, NULL},
	{(char *) "-both", (char *) "*oct.mode", XrmoptionNoArg, (char *) "4"},
	{(char *) "-orient", (char *) "*oct.orient", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noorient", (char *) "*oct.orient", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-practice", (char *) "*oct.practice", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nopractice", (char *) "*oct.practice", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-username", (char *) "*oct.userName", XrmoptionSepArg, NULL}
};

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *message)
{
	Arg         wargs[1];
	XmString    xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintStatef() requires a Label Widget");
	xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(wargs[0], XmNlabelString, xmstr);
	XtSetValues(w, wargs, 1);
}
#else
static void
PrintState(Widget w, int mode, int size, Boolean sticky, int moves, char *message)
{
	char        ss[10], mb[10];

	if (sticky)
		(void) strcpy(ss, "sticky");
	else
		(void) sprintf(ss, "%d", size);
	if (mode == BOTH)
		(void) strcpy(mb, "both");
	else
		(void) sprintf(mb, "%d", mode);
	(void) sprintf(titleDsp, "%s.%s: %s@ (%d/%s) - %s",
		progDsp, mb, ss, moves, recordDsp, message);
	XtSetArg(arg[0], XtNtitle, titleDsp);
	XtSetValues(w, arg, 1);
}
#endif

static void
InitRecords(void)
{
	int         i, mode, orient;

	for (mode = 0; mode < MAXMODES; mode++)
		for (orient = 0; orient < 2; orient++)
			for (i = 0; i <= MAXOCTAS - MINOCTAS + 1; i++) {
				octRecord[mode][orient][i].score = NEVER;
				(void) strcpy(octRecord[mode][orient][i].name, NOACCESS);
			}
}

static void
ReadRecords(void)
{
	FILE       *fp;
	int         i, n, mode, orient;
	char        username[USERNAMELEN];

	if ((fp = fopen(SCOREFILE, "r")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
		for (mode = 0; mode < MAXMODES; mode++)
			for (orient = 0; orient < 2; orient++)
				for (i = 0; i <= MAXOCTAS - MINOCTAS + 1; i++) {
					(void) fscanf(fp, "%d %s\n", &n, username);
					if (n <= octRecord[mode][orient][i].score ||
					    octRecord[mode][orient][i].score <= NEVER) {
						octRecord[mode][orient][i].score = n;
						(void) strcpy(octRecord[mode][orient][i].name, username);
					}
				}
		(void) fclose(fp);
	}
}

static void
WriteRecords(void)
{
	FILE       *fp;
	int         i, mode, orient;

	if ((fp = fopen(SCOREFILE, "w")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif

	} else {
#if HAVE_FCNTL_H
		int         lfd;
		char        lockfile[FILENAMELEN];

		(void) strcpy(lockfile, SCOREFILE);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
		       errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr, "Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr, "Lock file exists... score not recorded - sorry.\n"
				);
			return;
#endif
		}
#endif
		for (mode = 0; mode < MAXMODES; mode++) {
			for (orient = 0; orient < 2; orient++) {
				for (i = 0; i <= MAXOCTAS - MINOCTAS + 1; i++)
					(void) fprintf(fp, "%d %s\n",
						       octRecord[mode][orient][i].score, octRecord[mode][orient][i].name);
				(void) fprintf(fp, "\n");
			}
			(void) fprintf(fp, "\n");
		}
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
}

static void
PrintRecord(int size, int mode, Boolean orient, Boolean sticky, Boolean practice)
{
	int         i = mode - PERIOD3;
	int         j = (orient) ? 1 : 0;
	int         k = (sticky) ? MAXOCTAS - MINOCTAS + 1 : size - MINOCTAS;

	if (practice)
		(void) strcpy(recordDsp, "practice");
	else if (!sticky && size > MAXOCTAS)
		(void) strcpy(recordDsp, "NOT RECORDED");
	else if (octRecord[i][j][k].score <= NEVER) {
		(void) sprintf(recordDsp, "NEVER %s", NOACCESS);
	} else {
		(void) sprintf(recordDsp, "%d %s",
			  octRecord[i][j][k].score, octRecord[i][j][k].name);
	}
#ifdef HAVE_MOTIF
	PrintState(record, recordDsp);
#endif
}

static      Boolean
HandleSolved(int counter, int size, int mode, Boolean orient, Boolean sticky)
{
	int         i = mode - PERIOD3;
	int         j = (orient) ? 1 : 0;
	int         k = (sticky) ? MAXOCTAS - MINOCTAS + 1 : size - MINOCTAS;

	if ((sticky || size <= MAXOCTAS) && (counter < octRecord[i][j][k].score ||
	    octRecord[i][j][k].score <= NEVER)) {
		ReadRecords();	/* Maybe its been updated by another */
		octRecord[i][j][k].score = counter;
		(void) strcpy(octRecord[i][j][k].name, usernameDsp);
		if ((size < 2 && mode != PERIOD4) || (size < 4 && mode == PERIOD4) ||
		    (orient && (counter < octRecord[i][!j][k].score ||
		    octRecord[i][!j][k].score <= NEVER))) {
			octRecord[i][!j][k].score = counter;
			(void) strcpy(octRecord[i][!j][k].name, usernameDsp);
		}
		WriteRecords();
		PrintRecord(size, mode, orient, sticky, False);
		return True;
	}
	return False;
}

static void
Initialize(Widget w)
{
	int         size, mode;
	Boolean     orient, sticky, practice;
	String      username;

	XtVaSetValues(w,
		      XtNstart, False, NULL);
	XtVaGetValues(w,
		      XtNuserName, &username,
		      XtNsize, &size,
		      XtNmode, &mode,
		      XtNorient, &orient,
		      XtNsticky, &sticky,
		      XtNpractice, &practice, NULL);
#ifdef HAVE_MOTIF
	if (size > MAXOCTAS)
		XtVaSetValues(octas, XmNmaximum, size, NULL);
	XmScaleSetValue(octas, size);
	XmToggleButtonSetState(modes[mode - PERIOD3], True, False);
	XmToggleButtonSetState(orientSwitch, orient, True);
	XmToggleButtonSetState(stickySwitch, sticky, True);
	XmToggleButtonSetState(practiceSwitch, practice, True);
#endif
	InitRecords();
	ReadRecords();
	(void) strcpy(usernameDsp, username);
	if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
	    !strcmp(usernameDsp, NOACCESS) || !strcmp(usernameDsp, NOBODY)) {
		/* The NOACCESS is not necessary, but it stops people from being cute. */
		(void) sprintf(usernameDsp, "%s", getlogin());
		if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
		    !strcmp(usernameDsp, NOACCESS))
			(void) sprintf(usernameDsp, "%s", NOBODY);	/* It really IS nobody */
	}
	PrintRecord(size, mode, orient, sticky, practice);
	oldSize = size;
#ifndef HAVE_MOTIF
	PrintState(XtParent(w), mode, size, sticky, movesDsp, messageDsp);
#endif
}

static void
CallbackOct(Widget w, caddr_t clientData, octCallbackStruct * callData)
{
	int         size, mode;
	Boolean     orient, sticky, practice, start;

	XtVaGetValues(w,
		      XtNsize, &size,
		      XtNmode, &mode,
		      XtNorient, &orient,
		      XtNsticky, &sticky,
		      XtNpractice, &practice,
		      XtNstart, &start, NULL);
	(void) strcpy(messageDsp, "");
	switch (callData->reason) {
		case OCT_RESTORE:
			if (practice) {
				(void) strcpy(recordDsp, "practice");
#ifdef HAVE_MOTIF
				PrintState(record, recordDsp);
#endif
			}
			movesDsp = 0;
			break;
		case OCT_RESET:
			movesDsp = 0;
			break;
		case OCT_AMBIGUOUS:
			(void) strcpy(messageDsp, "Ambiguous move");
			break;
		case OCT_ILLEGAL:
			if (practice || start)
				(void) strcpy(messageDsp, "Illegal move");
			else
				(void) strcpy(messageDsp, "Randomize to start");
			break;
		case OCT_MOVED:
			movesDsp++;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
		case OCT_CONTROL:
			return;
		case OCT_SOLVED:
			if (practice)
				movesDsp = 0;
			else {
				if (HandleSolved(movesDsp, size, mode, orient, sticky))
					(void) sprintf(messageDsp, "Congratulations %s!!", usernameDsp);
				else
					(void) strcpy(messageDsp, "Solved!");
			}
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case OCT_PRACTICE:
			movesDsp = 0;
			practice = !practice;
			if (!practice)
				(void) strcpy(messageDsp, "Randomize to start");
			XtSetArg(arg[0], XtNpractice, practice);
			XtSetArg(arg[1], XtNstart, False);
			XtSetValues(w, arg, 2);
			PrintRecord(size, mode, orient, sticky, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(practiceSwitch, practice, True);
#endif
			break;
		case OCT_RANDOMIZE:
			movesDsp = 0;
			XtSetArg(arg[0], XtNpractice, False);
			XtSetArg(arg[1], XtNstart, False);
			XtSetValues(w, arg, 2);
			break;
		case OCT_DEC:
			if (!sticky) {
				movesDsp = 0;
				size--;
				oldSize = size;
				XtSetArg(arg[0], XtNsize, size);
				XtSetValues(w, arg, 1);
				PrintRecord(size, mode, orient, sticky, practice);
#ifdef HAVE_MOTIF
				XmScaleSetValue(octas, size);
				if (size >= MAXOCTAS)
					XtVaSetValues(octas, XmNmaximum, size, NULL);
#endif
			}
			break;
		case OCT_ORIENT:
			movesDsp = 0;
			orient = !orient;
			XtSetArg(arg[0], XtNorient, orient);
			XtSetValues(w, arg, 1);
			PrintRecord(size, mode, orient, sticky, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(orientSwitch, orient, True);
#endif
			break;
		case OCT_INC:
			if (!sticky) {
				movesDsp = 0;
				size++;
				oldSize = size;
				XtSetArg(arg[0], XtNsize, size);
				XtSetValues(w, arg, 1);
				PrintRecord(size, mode, orient, sticky, practice);
#ifdef HAVE_MOTIF
				if (size > MAXOCTAS)
					XtVaSetValues(octas, XmNmaximum, size, NULL);
				XmScaleSetValue(octas, size);
#endif
			}
			break;
		case OCT_PERIOD3:
		case OCT_PERIOD4:
		case OCT_BOTH:
			movesDsp = 0;
			mode = callData->reason - OCT_PERIOD3 + PERIOD3;
			XtSetArg(arg[0], XtNmode, mode);
			XtSetValues(w, arg, 1);
			PrintRecord(size, mode, orient, sticky, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(modes[mode - PERIOD3], True, True);
#endif
			break;
		case OCT_STICKY:
			movesDsp = 0;
			sticky = !sticky;
			if (sticky)
				size = 4;
			else
				size = oldSize;
			XtSetArg(arg[0], XtNsticky, sticky);
			XtSetArg(arg[1], XtNsize, size);
			XtSetValues(w, arg, 2);
			PrintRecord(size, mode, orient, sticky, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(stickySwitch, sticky, True);
			XmScaleSetValue(octas, size);
#endif
			break;
		case OCT_COMPUTED:
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case OCT_UNDO:
			movesDsp--;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
#else
	PrintState(XtParent(w), mode, size, sticky, movesDsp, messageDsp);
#endif
}

#ifdef HAVE_MOTIF
static void
OctaSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         size = cbs->value, mode;
	Boolean     orient, sticky, practice;

	XtVaGetValues(oct,
		      XtNsize, &oldSize,
		      XtNmode, &mode,
		      XtNorient, &orient,
		      XtNsticky, &sticky,
		      XtNpractice, &practice, NULL);
	if (sticky)
		XmScaleSetValue(w, oldSize);
	else if (oldSize != size) {
		XtVaSetValues(oct,
			      XtNsize, size, NULL);
		oldSize = size;
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(size, mode, orient, sticky, practice);
	}
}

static void
ModeToggle(Widget w, int mode, XmToggleButtonCallbackStruct * cbs)
{
	int         size;
	Boolean     orient, sticky, practice;

	if (cbs->set) {
		XtVaGetValues(oct,
			      XtNsize, &size,
			      XtNorient, &orient,
			      XtNsticky, &sticky,
			      XtNpractice, &practice, NULL);
		XtVaSetValues(oct,
			      XtNmode, mode + PERIOD3, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(size, mode + PERIOD3, orient, sticky, practice);
	}
}

static void
OrientToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         size, mode;
	Boolean     orient = cbs->set, sticky, practice;

	XtVaGetValues(oct,
		      XtNsize, &size,
		      XtNmode, &mode,
		      XtNsticky, &sticky,
		      XtNpractice, &practice, NULL);
	XtVaSetValues(oct,
		      XtNorient, orient, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	PrintRecord(size, mode, orient, sticky, practice);
}

static void
StickyToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         size, mode;
	Boolean     orient, sticky = cbs->set, practice;

	XtVaGetValues(oct,
		      XtNsize, &size,
		      XtNmode, &mode,
		      XtNorient, &orient,
		      XtNpractice, &practice, NULL);
	XtVaSetValues(oct,
		      XtNsticky, sticky, NULL);
	if ((sticky && size != 4) || (!sticky && oldSize != 4)) {
		if (sticky)
			size = 4;
		else
			size = oldSize;
		if (size <= MAXOCTAS)
			XmScaleSetValue(octas, size);
		XtVaSetValues(oct,
			      XtNsize, size, NULL);
	}
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	PrintRecord(size, mode, orient, sticky, practice);
}

static void
PracticeToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         size, mode;
	Boolean     orient, sticky, practice = cbs->set;

	XtVaSetValues(oct,
		      XtNpractice, practice,
		      XtNstart, False, NULL);
	XtVaGetValues(oct,
		      XtNsize, &size,
		      XtNmode, &mode,
		      XtNorient, &orient,
		      XtNsticky, &sticky, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	if (!practice)
		(void) strcpy(messageDsp, "Randomize to start");
	PrintRecord(size, mode, orient, sticky, practice);
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2)
		exit(0);
	XtSetArg(arg[0], XtNmenu, val);
	XtSetValues(oct, arg, 1);
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;
	XtSetArg(arg[0], XtNmenu, val + 3);  /* GWQ */
	XtSetValues(oct, arg, 1);
}

static const char aboutHelp[] = {
	"Version 5.6.1 by David Bagley <bagleyd@tux.org> http://www.tux.org/~bagleyd"
};

static const char helpHelp[] = {
	"Consult man page.  :)"
};

static void
helpCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		(void) printf("%s: %s\n", progDsp, aboutHelp);
	else
		(void) printf("%s: %s\n", progDsp, helpHelp);
}
#endif

int
main(int argc, char **argv)
{
	Widget      topLevel;
#ifdef HAVE_MOTIF
	Widget      menubar, panel, panel2, panel3;
	Widget      rowcol, rowcol2, rowcol3, rowcol4, rowcol5;
	Pixmap      mouseLeftCursor, mouseRightCursor;
	Pixel       fg, bg;
	XmString    file, play, help, about;
	XmString    get, write, quit;
	XmString    clear, undo, randomize, solve;
	XmString    increment, decrement, orientize, sticky, practice;
	unsigned int i;
#endif

	progDsp = argv[0];
	topLevel = XtInitialize(argv[0], "Oct",
				options, XtNumber(options), &argc, argv);
	if (argc != 1)
		Usage(argv[0]);

	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(topLevel),
				      RootWindowOfScreen(XtScreen(topLevel)),
				  (char *) oct_bits, oct_width, oct_height));
#ifdef HAVE_MOTIF
	XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER);	/* not XmEXPLICIT */
	XtSetValues(topLevel, arg, 2);
	panel = XtVaCreateManagedWidget("panel",
				xmPanedWindowWidgetClass, topLevel,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	file = XmStringCreateSimple((char *) "File");
	play = XmStringCreateSimple((char *) "Play");
	help = XmStringCreateSimple((char *) "Help");
	menubar = XmVaCreateSimpleMenuBar(panel, (char *) "menubar",
		XmVaCASCADEBUTTON, file, 'F',
		XmVaCASCADEBUTTON, play, 'P',
		XmVaCASCADEBUTTON, help, 'H',
		NULL);
	XmStringFree(file);
	XmStringFree(play);
	get = XmStringCreateSimple((char *) "Get");
	write = XmStringCreateSimple((char *) "Write");
	quit = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "file_menu", 0, fileCB,
		XmVaPUSHBUTTON, get, 'G', NULL, NULL,
		XmVaPUSHBUTTON, write, 'W', NULL, NULL,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quit, 'Q', NULL, NULL,
		NULL);
	XmStringFree(get);
	XmStringFree(write);
	XmStringFree(quit);
	clear = XmStringCreateSimple((char *) "Clear");
	undo = XmStringCreateSimple((char *) "Undo");
	randomize = XmStringCreateSimple((char *) "Randomize");
	solve = XmStringCreateSimple((char *) "(Solve)");
	increment = XmStringCreateSimple((char *) "Increment");
	decrement = XmStringCreateSimple((char *) "Decrement");
	orientize = XmStringCreateSimple((char *) "Orientize");
	sticky = XmStringCreateSimple((char *) "StickY");
	practice = XmStringCreateSimple((char *) "Practice");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, clear, 'C', NULL, NULL,
		XmVaPUSHBUTTON, undo, 'U', NULL, NULL,
		XmVaPUSHBUTTON, randomize, 'R', NULL, NULL,
		XmVaPUSHBUTTON, solve, 'S', NULL, NULL,
		XmVaPUSHBUTTON, increment, 'I', NULL, NULL,
		XmVaPUSHBUTTON, decrement, 'D', NULL, NULL,
		XmVaPUSHBUTTON, orientize, 'O', NULL, NULL,
		XmVaPUSHBUTTON, sticky, 'Y', NULL, NULL,
		XmVaPUSHBUTTON, practice, 'P', NULL, NULL,
		NULL);
	XmStringFree(clear);
	XmStringFree(undo);
	XmStringFree(randomize);
	XmStringFree(solve);
	XmStringFree(increment);
	XmStringFree(decrement);
	XmStringFree(orientize);
	XmStringFree(sticky);
	XmStringFree(practice);
	about = XmStringCreateSimple((char *) "About");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "help_menu", 2, helpCB,
		XmVaPUSHBUTTON, about, 'A', NULL, NULL,
		XmVaPUSHBUTTON, help, 'H', NULL, NULL,
		NULL);
	XmStringFree(about);
	XmStringFree(help);
	XtManageChild(menubar);
	panel2 = XtCreateManagedWidget("panel2",
				       xmPanedWindowWidgetClass, panel,
				       NULL, 0);
	panel3 = XtVaCreateManagedWidget("panel3",
					 xmPanedWindowWidgetClass, panel2,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	rowcol = XtVaCreateManagedWidget("Rowcol",
					 xmRowColumnWidgetClass, panel3,
					 XmNnumColumns, 2,
					 XmNorientation, XmHORIZONTAL,
					 XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(rowcol,
		      XmNforeground, &fg,
		      XmNbackground, &bg, NULL);
	mouseLeftCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	      RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_left_bits,
				 mouse_left_width, mouse_left_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	mouseRightCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	     RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_right_bits,
			       mouse_right_width, mouse_right_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	(void) XtVaCreateManagedWidget("mouseLeftText",
				xmLabelGadgetClass, rowcol,
		   XtVaTypedArg, XmNlabelString, XmRString, "Move", 5, NULL);
	(void) XtVaCreateManagedWidget("mouseLeft",
				xmLabelGadgetClass, rowcol,
	      XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseLeftCursor, NULL);
	(void) XtVaCreateManagedWidget("mouseRightText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Randomize", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseRight",
				xmLabelGadgetClass, rowcol,
	     XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseRightCursor, NULL);
	(void) XtVaCreateManagedWidget("CallbackOct",
				xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Moves", 6, NULL);
	moves = XtVaCreateManagedWidget("0",
					xmLabelWidgetClass, rowcol, NULL);
	(void) XtVaCreateManagedWidget("record_text",
				xmLabelGadgetClass, rowcol,
		 XtVaTypedArg, XmNlabelString, XmRString, "Record", 7, NULL);
	record = XtVaCreateManagedWidget("0",
					 xmLabelWidgetClass, rowcol, NULL);

	rowcol2 = XtVaCreateManagedWidget("Rowcol2",
				       xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	octas = XtVaCreateManagedWidget("octas",
					xmScaleWidgetClass, rowcol2,
			 XtVaTypedArg, XmNtitleString, XmRString, "Size", 6,
					XmNminimum, MINOCTAS,
					XmNmaximum, MAXOCTAS,
					XmNvalue, MINOCTAS,
					XmNshowValue, True,
					XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(octas,
		      XmNvalueChangedCallback, (XtCallbackProc) OctaSlider, (XtPointer) NULL);
	rowcol3 = XtVaCreateManagedWidget("Rowcol3",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN,
					  XmNradioBehavior, True, NULL);
	for (i = 0; i < XtNumber(modeString); i++) {
		modes[i] = XtVaCreateManagedWidget(modeString[i],
					  xmToggleButtonGadgetClass, rowcol3,
					       XmNradioBehavior, True, NULL);
		XtAddCallback(modes[i],
			      XmNvalueChangedCallback, (XtCallbackProc) ModeToggle, (XtPointer) i);
	}
	rowcol4 = XtVaCreateManagedWidget("Rowcol4",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	orientSwitch = XtVaCreateManagedWidget("Oriented",
				   xmToggleButtonWidgetClass, rowcol4, NULL);
	XtAddCallback(orientSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) OrientToggle, (XtPointer) NULL);
	stickySwitch = XtVaCreateManagedWidget("Sticky",
				   xmToggleButtonWidgetClass, rowcol4, NULL);
	XtAddCallback(stickySwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) StickyToggle, (XtPointer) NULL);
	practiceSwitch = XtVaCreateManagedWidget("Practice",
				   xmToggleButtonWidgetClass, rowcol4, NULL);
	XtAddCallback(practiceSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) PracticeToggle, (XtPointer) NULL);
	rowcol5 = XtVaCreateManagedWidget("Rowcol5",
				       xmRowColumnWidgetClass, panel3, NULL);
	message = XtVaCreateManagedWidget("Play Oct! (use mouse and keypad)",
					  xmLabelWidgetClass, rowcol5, NULL);
	oct = XtCreateManagedWidget("oct",
				    octWidgetClass, panel2, NULL, 0);
#else
	XtSetArg(arg[1], XtNinput, True);
	XtSetValues(topLevel, arg, 2);
	oct = XtCreateManagedWidget("oct",
				    octWidgetClass, topLevel, NULL, 0);
#endif
	XtAddCallback(oct,
	  XtNselectCallback, (XtCallbackProc) CallbackOct, (XtPointer) NULL);
	Initialize(oct);
	XtRealizeWidget(topLevel);
	XGrabButton(XtDisplay(oct), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(oct), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(oct),
		    XCreateFontCursor(XtDisplay(oct), XC_crosshair));
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
