/*-
# MOTIF/X-BASED PANEX(tm)
#
#  xpanex.c
#
###
#
#  Copyright (c) 1996 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 5.4: 97/01/01 Xt/Motif
  Version 5.3: 96/04/30 Xt/Motif
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifdef VMS
#include <unixlib.h>
#define getlogin() cuserid(NULL)
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#if HAVE_FCNTL_H
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include "mouse-l.xbm"
#include "mouse-r.xbm"
#endif
#include "Panex.h"
#include "panex.xbm"

#ifndef SCOREFILE
#define SCOREFILE "/usr/games/lib/panex.scores"
#endif

/* The following are in PanexP.h also */
#define MINTILES 1
#define HANOI 0
#define PANEX 1
#define MAXMODES 2

#define MAXTILES 10
#define MINSPEED 1
#define MAXSPEED 50
#define MULTSPEED 20
#define NEVER -1
#define FILENAMELEN 1024
#define USERNAMELEN 128
#define NOACCESS "noaccess"
#define NOBODY "nobody"

typedef struct {
	int         score;
	char        name[USERNAMELEN];
} GameRecord;

#ifdef HAVE_MOTIF
static Widget moves, record, message, modes[MAXMODES], tile, speed;
static char buff[16];
static const char *modeString[] =
{
	"Hanoi", "Panex"
};
#else
static char titleDsp[256] = "";
#endif

static Widget panex;
static Arg  arg[2];
static GameRecord panexRecord[MAXMODES][MAXTILES - MINTILES + 1];
static int  movesDsp = 0;
static char *progDsp;
static char messageDsp[128] = "Welcome";
static char recordDsp[16] = "NOT RECORDED";
static char usernameDsp[USERNAMELEN] = "";

static void
Usage(char * programName)
{
	(void) fprintf(stderr, "usage: %s\n", programName);
	(void) fprintf(stderr,
	     "\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
	(void) fprintf(stderr,
	   "\t[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n");
	(void) fprintf(stderr,
		"\t[-{foreground|fg} {color}] [-{background|bg} {color}]\n");
	(void) fprintf(stderr,
	"\t[-{border|bd} {color}] [-tile {color}] [-pyramid{0|1} {color}]\n");
	(void) fprintf(stderr,
	    "\t[-tiles {int}] [-{mode {int}|hanoi|panex}] [-delay msecs]\n");
	(void) fprintf(stderr,
		       "\t[-username {string}]\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*panex.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*panex.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*panex.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*panex.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*panex.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*panex.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*panex.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*panex.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*panex.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*panex.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-tile", (char *) "*panex.tileColor", XrmoptionSepArg, NULL},
	{(char *) "-pyramid0", (char *) "*panex.pyramidColor0", XrmoptionSepArg, NULL},
	{(char *) "-pyramid1", (char *) "*panex.pyramidColor1", XrmoptionSepArg, NULL},
	{(char *) "-tiles", (char *) "*panex.tiles", XrmoptionSepArg, NULL},
	{(char *) "-mode", (char *) "*panex.mode", XrmoptionSepArg, NULL},
	{(char *) "-hanoi", (char *) "*panex.mode", XrmoptionNoArg, (char *) "0"},
	{(char *) "-panex", (char *) "*panex.mode", XrmoptionNoArg, (char *) "1"},
	{(char *) "-delay", (char *) "*panex.delay", XrmoptionSepArg, NULL},
	{(char *) "-username", (char *) "*userName", XrmoptionSepArg, NULL}
};

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *message)
{
	Arg         wargs[1];
	XmString    xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(wargs[0], XmNlabelString, xmstr);
	XtSetValues(w, wargs, 1);
}
#else
static void
PrintState(Widget w, int tiles, int moves, char *message)
{
        (void) sprintf(titleDsp, "%s: %d@ (%d/%s) - %s",
                progDsp, tiles, moves, recordDsp, message);
        XtSetArg(arg[0], XtNtitle, titleDsp);
        XtSetValues(w, arg, 1);
}
#endif

static void
InitRecords(void)
{
	int         i, mode;

	for (mode = 0; mode < MAXMODES; mode++)
		for (i = 0; i < MAXTILES - MINTILES + 1; i++) {
			panexRecord[mode][i].score = NEVER;
			(void) strcpy(panexRecord[mode][i].name, NOACCESS);
		}
}

static void
ReadRecords(void)
{
	FILE       *fp;
	int         i, mode, n;
	char        username[USERNAMELEN];

	if ((fp = fopen(SCOREFILE, "r")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
		for (mode = 0; mode < MAXMODES; mode++)
			for (i = 0; i < MAXTILES - MINTILES + 1; i++) {
				(void) fscanf(fp, "%d %s\n", &n, username);
				if (n <= panexRecord[mode][i].score ||
	    			    panexRecord[mode][i].score <= NEVER) {
					panexRecord[mode][i].score = n;
					(void) strcpy(panexRecord[mode][i].name, username);
				}
			}
		(void) fclose(fp);
	}
}

static void
WriteRecords(void)
{
	FILE       *fp;
	int         i, mode;

	if ((fp = fopen(SCOREFILE, "w")) == NULL) {
		(void) sprintf(messageDsp, "Can not write to %s.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
#if HAVE_FCNTL_H
		int         lfd;
		char        lockfile[FILENAMELEN];

		(void) strcpy(lockfile, SCOREFILE);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
		       errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr, "Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr, "Lock file exists... score not recorded - sorry.\n");
			return;
#endif
		}
#endif
		for (mode = 0; mode < MAXMODES; mode++) {
			for (i = 0; i < MAXTILES - MINTILES + 1; i++)
				(void) fprintf(fp, "%d %s\n",
					       panexRecord[mode][i].score, panexRecord[mode][i].name);
			(void) fprintf(fp, "\n");
		}
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
}

static void
PrintRecord(int tiles, int mode)
{
	int         i = tiles - MINTILES, j = mode - HANOI;

	if (tiles > MAXTILES)
		(void) strcpy(recordDsp, "NOT RECORDED");
	else if (panexRecord[j][i].score <= NEVER) {
		(void) sprintf(recordDsp, "NEVER %s", NOACCESS);
	} else {
		(void) sprintf(recordDsp, "%d %s",
			    panexRecord[j][i].score, panexRecord[j][i].name);

	}
#ifdef HAVE_MOTIF
	PrintState(record, recordDsp);
#endif
}

static      Boolean
HandleSolved(int counter, int tiles, int mode)
{
	int         i = tiles - MINTILES, j = mode - HANOI;

	if (tiles <= MAXTILES && (counter < panexRecord[j][i].score ||
	    panexRecord[j][i].score <= NEVER)) {
		ReadRecords();	/* Maybe its been updated by another */
		panexRecord[j][i].score = counter;
		(void) strcpy(panexRecord[j][i].name, usernameDsp);
		WriteRecords();
		PrintRecord(tiles, mode);
		return True;
	}
	return False;
}

static void
Initialize(Widget w)
{
	int         tiles, mode, delay;
	String      username;

	/*XtVaSetValues(w,
	   XtNstart, False, NULL); */
	XtVaGetValues(w,
		      XtNuserName, &username,
		      XtNtiles, &tiles,
		      XtNmode, &mode,
		      XtNdelay, &delay, NULL);
#ifdef HAVE_MOTIF
	if (tiles > MAXTILES)
		XtVaSetValues(tile, XmNmaximum, tiles, NULL);
	XmScaleSetValue(tile, tiles);
	XmToggleButtonSetState(modes[mode - HANOI], True, False);
	XmScaleSetValue(speed, MAXSPEED + MINSPEED - delay / MULTSPEED - 1);
#endif
	InitRecords();
	ReadRecords();
	(void) strcpy(usernameDsp, username);
	if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
	    !strcmp(usernameDsp, NOACCESS) || !strcmp(usernameDsp, NOBODY)) {
		/* The NOACCESS is not necessary, but it stops people from being cute. */
		(void) sprintf(usernameDsp, "%s", getlogin());
		if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
		    !strcmp(usernameDsp, NOACCESS))
			(void) sprintf(usernameDsp, "%s", NOBODY);	/* It really IS nobody */
	}
	PrintRecord(tiles, mode);
#ifndef HAVE_MOTIF
	PrintState(XtParent(w), tiles, movesDsp, messageDsp);
#endif

}

static void
CallbackPanex(Widget w, caddr_t clientData, panexCallbackStruct * callData)
{
	int         tiles, mode;

	XtVaGetValues(w,
		      XtNtiles, &tiles,
		      XtNmode, &mode, NULL);
	(void) strcpy(messageDsp, "");
	switch (callData->reason) {
		case PANEX_RESTORE:
		case PANEX_RESET:
			movesDsp = 0;
			break;
		case PANEX_ILLEGAL:
			(void) strcpy(messageDsp, "Illegal move");
			break;
		case PANEX_BLOCKED:
			(void) strcpy(messageDsp, "Blocked");
			break;
		case PANEX_SPACE:
#if 0
			/* Too annoying */
			(void) strcpy(messageDsp, "Spaces can not slide");
#endif
			break;
		case PANEX_IGNORE:
			(void) strcpy(messageDsp, "Clear to start");
			break;
		case PANEX_MOVED:
			movesDsp++;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
		case PANEX_SOLVED:
			if (HandleSolved(movesDsp, tiles, mode))
				(void) sprintf(messageDsp, "Congratulations %s!!", usernameDsp);
			else
				(void) strcpy(messageDsp, "Solved!");
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case PANEX_DEC:
			movesDsp = 0;
			tiles--;
			XtSetArg(arg[0], XtNtiles, tiles);
			XtSetValues(w, arg, 1);
			PrintRecord(tiles, mode);
#ifdef HAVE_MOTIF
			XmScaleSetValue(tile, tiles);
			if (tiles >= MAXTILES)
				XtVaSetValues(tile, XmNmaximum, tiles, NULL);
#endif
			break;
		case PANEX_INC:
			movesDsp = 0;
			tiles++;
			XtSetArg(arg[0], XtNtiles, tiles);
			XtSetValues(w, arg, 1);
			PrintRecord(tiles, mode);
#ifdef HAVE_MOTIF
			if (tiles > MAXTILES)
				XtVaSetValues(tile, XmNmaximum, tiles, NULL);
			XmScaleSetValue(tile, tiles);
#endif
			break;
		case PANEX_MODE:
			movesDsp = 0;
			mode = (mode == PANEX) ? HANOI : PANEX;
			XtSetArg(arg[0], XtNmode, mode);
			XtSetValues(w, arg, 1);
			PrintRecord(tiles, mode);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(modes[mode - HANOI], True, True);
#endif
			break;
		case PANEX_COMPUTED:
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case PANEX_UNDO:
			movesDsp--;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
#else
	PrintState(XtParent(w), tiles, movesDsp, messageDsp);
#endif

}

#ifdef HAVE_MOTIF
static void
TileSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         tiles = cbs->value, mode, old;

	XtVaGetValues(panex,
		      XtNtiles, &old,
		      XtNmode, &mode, NULL);
	if (old != tiles) {
		XtVaSetValues(panex,
			      XtNtiles, tiles, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(tiles, mode);
	}
}

static void
ModeToggle(Widget w, int mode, XmToggleButtonCallbackStruct * cbs)
{
	int         tiles;

	if (cbs->set) {
		XtVaGetValues(panex,
			      XtNtiles, &tiles, NULL);
		XtVaSetValues(panex,
			      XtNmode, mode + HANOI, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(tiles, mode + HANOI);
	}
}

static void
SpeedSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int delay = MULTSPEED * (MAXSPEED + MINSPEED - cbs->value - 1), oldDelay;

	XtVaGetValues(panex,
		      XtNdelay, &oldDelay, NULL);
	if (oldDelay != delay) {
		XtVaSetValues(panex, XtNdelay, delay, NULL);
	}
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2)
		exit(0);
	XtSetArg(arg[0], XtNmenu, val);
	XtSetValues(panex, arg, 1);
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;
	XtSetArg(arg[0], XtNmenu, val + 3);  /* GWQ */
	XtSetValues(panex, arg, 1);
}

static const char aboutHelp[] = {
	"Version 5.6.1 by David Bagley <bagleyd@tux.org> http://www.tux.org/~bagleyd"
};

static const char helpHelp[] = {
	"Consult man page.  :)"
};

static void
helpCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		(void) printf("%s: %s\n", progDsp, aboutHelp);
	else
		(void) printf("%s: %s\n", progDsp, helpHelp);
}
#endif

int
main(int argc, char **argv)
{
	Widget      topLevel;
#ifdef HAVE_MOTIF
	Widget      menubar, panel, panel2, panel3;
	Widget      rowcol, rowcol2, rowcol3, rowcol4;
	Pixmap      mouseLeftCursor, mouseRightCursor;
	Pixel       fg, bg;
	XmString    file, play, help, about;
	XmString    get, write, quit;
	XmString    clear, undo, solve, increment, decrement, mode;
	unsigned int i;
#endif

	progDsp = argv[0];

	topLevel = XtInitialize(argv[0], "Panex",
				options, XtNumber(options), &argc, argv);
	if (argc != 1)
		Usage(argv[0]);

	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(topLevel),
				      RootWindowOfScreen(XtScreen(topLevel)),
			    (char *) panex_bits, panex_width, panex_height));
#ifdef HAVE_MOTIF
	XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER);	/* not XmEXPLICIT */
	XtSetValues(topLevel, arg, 2);
	panel = XtVaCreateManagedWidget("panel",
				xmPanedWindowWidgetClass, topLevel,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	file = XmStringCreateSimple((char *) "File");
	play = XmStringCreateSimple((char *) "Play");
	help = XmStringCreateSimple((char *) "Help");
	menubar = XmVaCreateSimpleMenuBar(panel, (char *) "menubar",
		XmVaCASCADEBUTTON, file, 'F',
		XmVaCASCADEBUTTON, play, 'P',
		XmVaCASCADEBUTTON, help, 'H',
		NULL);
	XmStringFree(file);
	XmStringFree(play);
	get = XmStringCreateSimple((char *) "Get");
	write = XmStringCreateSimple((char *) "Write");
	quit = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "file_menu", 0, fileCB,
		XmVaPUSHBUTTON, get, 'G', NULL, NULL,
		XmVaPUSHBUTTON, write, 'W', NULL, NULL,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quit, 'Q', NULL, NULL,
		NULL);
	XmStringFree(get);
	XmStringFree(write);
	XmStringFree(quit);
	clear = XmStringCreateSimple((char *) "Clear");
	undo = XmStringCreateSimple((char *) "Undo");
	solve = XmStringCreateSimple((char *) "Solve");
	increment = XmStringCreateSimple((char *) "Increment");
	decrement = XmStringCreateSimple((char *) "Decrement");
	mode = XmStringCreateSimple((char *) "Mode");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, clear, 'C', NULL, NULL,
		XmVaPUSHBUTTON, undo, 'U', NULL, NULL,
		XmVaPUSHBUTTON, solve, 'S', NULL, NULL,
		XmVaPUSHBUTTON, increment, 'I', NULL, NULL,
		XmVaPUSHBUTTON, decrement, 'D', NULL, NULL,
		XmVaPUSHBUTTON, mode, 'M', NULL, NULL,
		NULL);
	XmStringFree(clear);
	XmStringFree(undo);
	XmStringFree(solve);
	XmStringFree(increment);
	XmStringFree(decrement);
	XmStringFree(mode);
	about = XmStringCreateSimple((char *) "About");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "help_menu", 2, helpCB,
		XmVaPUSHBUTTON, about, 'A', NULL, NULL,
		XmVaPUSHBUTTON, help, 'H', NULL, NULL,
		NULL);
	XmStringFree(about);
	XmStringFree(help);
	XtManageChild(menubar);
	panel2 = XtCreateManagedWidget("panel2",
				       xmPanedWindowWidgetClass, panel,
				       NULL, 0);
	panel3 = XtVaCreateManagedWidget("panel3",
					 xmPanedWindowWidgetClass, panel2,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	rowcol = XtVaCreateManagedWidget("Rowcol",
					 xmRowColumnWidgetClass, panel3,
					 XmNnumColumns, 2,
					 XmNorientation, XmHORIZONTAL,
					 XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(rowcol,
		      XmNforeground, &fg,
		      XmNbackground, &bg, NULL);
	mouseLeftCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	      RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_left_bits,
				 mouse_left_width, mouse_left_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	mouseRightCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	     RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_right_bits,
			       mouse_right_width, mouse_right_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	(void) XtVaCreateManagedWidget("mouseLeftText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Move tile", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseLeft",
				xmLabelGadgetClass, rowcol,
	      XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseLeftCursor, NULL);
	(void) XtVaCreateManagedWidget("mouseRightText",
				xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Clear", 6, NULL);
	(void) XtVaCreateManagedWidget("mouseRight", xmLabelGadgetClass, rowcol,
	     XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseRightCursor, NULL);
	(void) XtVaCreateManagedWidget("movesText",
				xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Moves", 6, NULL);
	moves = XtVaCreateManagedWidget("0", xmLabelWidgetClass, rowcol, NULL);
	(void) XtVaCreateManagedWidget("recordText",
				xmLabelGadgetClass, rowcol,
		 XtVaTypedArg, XmNlabelString, XmRString, "Record", 7, NULL);
	record = XtVaCreateManagedWidget("0",
					 xmLabelWidgetClass, rowcol, NULL);

	rowcol2 = XtVaCreateManagedWidget("Rowcol2",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN,
					  XmNradioBehavior, True, NULL);
	for (i = 0; i < XtNumber(modeString); i++) {
		modes[i] = XtVaCreateManagedWidget(modeString[i],
					  xmToggleButtonGadgetClass, rowcol2,
					       XmNradioBehavior, True, NULL);
		XtAddCallback(modes[i],
			      XmNvalueChangedCallback, (XtCallbackProc) ModeToggle, (XtPointer) i);
	}
	rowcol3 = XtVaCreateManagedWidget("Rowcol3",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	tile = XtVaCreateManagedWidget("tile",
				       xmScaleWidgetClass, rowcol3,
			 XtVaTypedArg, XmNtitleString, XmRString, "Size", 6,
				       XmNminimum, MINTILES,
				       XmNmaximum, MAXTILES,
				       XmNvalue, MAXTILES,
				       XmNshowValue, True,
				       XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(tile,
		      XmNvalueChangedCallback, (XtCallbackProc) TileSlider, (XtPointer) NULL);
	speed = XtVaCreateManagedWidget("speed",
					 xmScaleWidgetClass, rowcol3,
		      XtVaTypedArg, XmNtitleString, XmRString, "Animation Speed", 8,
					 XmNminimum, MINSPEED,
					 XmNmaximum, MAXSPEED,
					 XmNvalue, MAXSPEED - 10,
					 XmNshowValue, True,
					 XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(speed,
		      XmNvalueChangedCallback, (XtCallbackProc) SpeedSlider, (XtPointer) NULL);

	rowcol4 = XtVaCreateManagedWidget("Rowcol4",
					  xmRowColumnWidgetClass, panel3, NULL);
	message = XtVaCreateManagedWidget("Play Panex!",
					  xmLabelWidgetClass, rowcol4, NULL);

	panex = XtCreateManagedWidget("panex",
				      panexWidgetClass, panel2, NULL, 0);
#else
	XtSetArg(arg[1], XtNinput, True);
	XtSetValues(topLevel, arg, 2);
	panex = XtCreateManagedWidget("panex",
				      panexWidgetClass, topLevel, NULL, 0);
#endif
	XtAddCallback(panex,
	XtNselectCallback, (XtCallbackProc) CallbackPanex, (XtPointer) NULL);
	Initialize(panex);
	XtRealizeWidget(topLevel);
	XGrabButton(XtDisplay(panex), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(panex), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(panex),
		    XCreateFontCursor(XtDisplay(panex), XC_crosshair));
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
