/*-
# MOTIF/X-BASED RUBIK'S CUBE(tm)
#
#  xrubik.c
#
###
#
#  Copyright (c) 1993 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "useful",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 5: 95/10/06 Xt/Motif
  Version 4: 94/04/07 Xt
  Version 3: 93/05/20 Motif
  Version 2: 92/01/16 XView
  Version 1: 91/01/16 SunView
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifdef VMS
#include <unixlib.h>
#define getlogin() cuserid(NULL)
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#if HAVE_FCNTL_H
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include "mouse-l.xbm"
#include "mouse-r.xbm"
#endif
#include "Rubik.h"
#include "Rubik2d.h"
#include "Rubik3d.h"
#include "rubik.xbm"

#ifndef SCOREFILE
#define SCOREFILE "/usr/games/lib/mball.scores"
#endif

/* The following are in RubikP.h also */
#define MINCUBES 1
#define MAXFACES 6

#define MAXCUBES 6
#define MINSPEED 1
#define MAXSPEED 50
#define MULTSPEED 20
#define NEVER -1
#define FILENAMELEN 1024
#define USERNAMELEN 128
#define NOACCESS "noaccess"
#define NOBODY "nobody"

typedef struct {
	int         score;
	char        name[USERNAMELEN];
} GameRecord;

#ifdef HAVE_MOTIF
static Widget moves, record, message, cubesx, cubesy, cubesz,
	orientSwitch, practiceSwitch, speed;
static char buff[16];
#else
static char titleDsp[256] = "";
#endif

static Widget rubik2d, rubik3d;
static Arg  arg[4];
static GameRecord rubikRecord[2][MAXCUBES - MINCUBES + 1][MAXCUBES - MINCUBES + 1][MAXCUBES - MINCUBES + 1];
static int  movesDsp = 0;
static char *progDsp;
static char messageDsp[128] = "Welcome";
static char recordDsp[16] = "NOT RECORDED";
static char usernameDsp[USERNAMELEN] = "";

static void
usage(char * programName)
{
	(void) fprintf(stderr, "usage: %s\n", programName);
	(void) fprintf(stderr,
		"\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
	(void) fprintf(stderr,
		"\t[-display [{host}]:[{vs}]] [-[no]mono] [-[no]{reverse|rv}]\n");
	(void) fprintf(stderr,
		"\t[-{foreground|fg} {color}] [-{background|bg} {color}]\n");
	(void) fprintf(stderr,
		"\t[-{border|bd} {color}] [-face{0|1|2|3|4|5} {color}]\n");
	(void) fprintf(stderr,
		"\t[-sizex {int}] [-sizey {int}] [-sizez {int}]\n");
	(void) fprintf(stderr,
		"\t[-[no]orient] [-[no]practice] [-delay msecs]\n");
	(void) fprintf(stderr,
		"\t[-username {string}]\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*rubik.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*rubik.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*rubik.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*rubik.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*rubik.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*rubik.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "rubik.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "rubik.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*rubik.pieceBorder", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*rubik.pieceBorder", XrmoptionSepArg, NULL},
	{(char *) "-face0", (char *) "*rubik.faceColor0", XrmoptionSepArg, NULL},
	{(char *) "-face1", (char *) "*rubik.faceColor1", XrmoptionSepArg, NULL},
	{(char *) "-face2", (char *) "*rubik.faceColor2", XrmoptionSepArg, NULL},
	{(char *) "-face3", (char *) "*rubik.faceColor3", XrmoptionSepArg, NULL},
	{(char *) "-face4", (char *) "*rubik.faceColor4", XrmoptionSepArg, NULL},
	{(char *) "-face5", (char *) "*rubik.faceColor5", XrmoptionSepArg, NULL},
	{(char *) "-sizex", (char *) "*rubik.sizex", XrmoptionSepArg, NULL},
	{(char *) "-sizey", (char *) "*rubik.sizey", XrmoptionSepArg, NULL},
	{(char *) "-sizez", (char *) "*rubik.sizez", XrmoptionSepArg, NULL},
	{(char *) "-orient", (char *) "*rubik.orient", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-noorient", (char *) "*rubik.orient", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-practice", (char *) "*rubik.practice", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nopractice", (char *) "*rubik.practice", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-delay", (char *) "*rubik.delay", XrmoptionSepArg, NULL},
	{(char *) "-username", (char *) "*rubik.userName", XrmoptionSepArg, NULL}
};

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *message)
{
	Arg         wargs[1];
	XmString    xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(wargs[0], XmNlabelString, xmstr);
	XtSetValues(w, wargs, 1);
}
#else
static void
PrintState(Widget w, int dim, int sizeX, int sizeY, int sizeZ, int moves, char *message)
{
	(void) sprintf(titleDsp, "%s%dd: %dx%dx%d@ (%d/%s) - %s",
		progDsp, dim, sizeX, sizeY, sizeZ, moves, recordDsp, message);
	XtSetArg(arg[0], XtNtitle, titleDsp);
	XtSetValues(w, arg, 1);
}
#endif

static void
InitRecords(void)
{
	int         i, j, k, orient;

	for (orient = 0; orient < 2; orient++)
		for (i = 0; i < MAXCUBES - MINCUBES + 1; i++)
			for (j = i; j < MAXCUBES - MINCUBES + 1; j++)
				for (k = j; k < MAXCUBES - MINCUBES + 1; k++) {
					rubikRecord[orient][k][j][i].score = rubikRecord[orient][k][i][j].score =
						rubikRecord[orient][j][k][i].score = rubikRecord[orient][j][i][k].score =
						rubikRecord[orient][i][k][j].score = rubikRecord[orient][i][j][k].score = NEVER;
					(void) strcpy(rubikRecord[orient][k][j][i].name, NOACCESS);
					(void) strcpy(rubikRecord[orient][k][i][j].name, NOACCESS);
					(void) strcpy(rubikRecord[orient][j][k][i].name, NOACCESS);
					(void) strcpy(rubikRecord[orient][j][i][k].name, NOACCESS);
					(void) strcpy(rubikRecord[orient][i][k][j].name, NOACCESS);
					(void) strcpy(rubikRecord[orient][i][j][k].name, NOACCESS);
				}
}

static void
ReadRecords(void)
{
	FILE       *fp;
	int         i, j, k, n, orient;
	char        username[USERNAMELEN];

	if ((fp = fopen(SCOREFILE, "r")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif

	} else {
		for (orient = 0; orient < 2; orient++)
			for (i = 0; i < MAXCUBES - MINCUBES + 1; i++)
				for (j = i; j < MAXCUBES - MINCUBES + 1; j++)
					for (k = j; k < MAXCUBES - MINCUBES + 1; k++) {
						(void) fscanf(fp, "%d %s\n", &n, username);
						if (n <= rubikRecord[orient][i][j][k].score ||
						    rubikRecord[orient][i][j][k].score <= NEVER) {
							rubikRecord[orient][k][j][i].score = rubikRecord[orient][k][i][j].score =
								rubikRecord[orient][j][k][i].score = rubikRecord[orient][j][i][k].score =
								rubikRecord[orient][i][k][j].score = rubikRecord[orient][i][j][k].score = n;
							(void) strcpy(rubikRecord[orient][k][j][i].name, username);
							(void) strcpy(rubikRecord[orient][k][i][j].name, username);
							(void) strcpy(rubikRecord[orient][j][k][i].name, username);
							(void) strcpy(rubikRecord[orient][j][i][k].name, username);
							(void) strcpy(rubikRecord[orient][i][k][j].name, username);
							(void) strcpy(rubikRecord[orient][i][j][k].name, username);
						}
					}
		(void) fclose(fp);
	}
}

static void
WriteRecords(void)
{
	FILE       *fp;
	int         i, j, k, orient;

	if ((fp = fopen(SCOREFILE, "w")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
#if HAVE_FCNTL_H
		int         lfd;
		char        lockfile[FILENAMELEN];

		(void) strcpy(lockfile, SCOREFILE);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
		       errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr, "Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr, "Lock file exists... score not recorded - sorry.\n"
				);
			return;
#endif
		}
#endif
		for (orient = 0; orient < 2; orient++) {
			for (i = 0; i < MAXCUBES - MINCUBES + 1; i++) {
				for (j = i; j < MAXCUBES - MINCUBES + 1; j++) {
					for (k = j; k < MAXCUBES - MINCUBES + 1; k++)
						(void) fprintf(fp, "%d %s\n",
							       rubikRecord[orient][i][j][k].score, rubikRecord[orient][i][j][k].name);
					(void) fprintf(fp, "\n");
				}
				(void) fprintf(fp, "\n");
			}

			(void) fprintf(fp, "\n");
		}
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
}

static void
PrintRecord(int sizex, int sizey, int sizez, Boolean orient, Boolean practice)
{
	int         i = sizex - MINCUBES, j = sizey - MINCUBES, k = sizez - MINCUBES;
	int         l = (orient) ? 1 : 0;

	if (practice)
		(void) strcpy(recordDsp, "practice");
	else if (sizex > MAXCUBES || sizey > MAXCUBES || sizez > MAXCUBES)
		(void) strcpy(recordDsp, "NOT RECORDED");
	else if (rubikRecord[l][i][j][k].score <= NEVER) {
		(void) sprintf(recordDsp, "NEVER %s", NOACCESS);
	} else {
		(void) sprintf(recordDsp, "%d %s",
			rubikRecord[l][i][j][k].score,
			rubikRecord[l][i][j][k].name);
	}
#ifdef HAVE_MOTIF
	PrintState(record, recordDsp);
#endif
}

static int
HandleSolved(int counter, int sizex, int sizey, int sizez, Boolean orient)
{
	int         i = sizex - MINCUBES, j = sizey - MINCUBES, k = sizez - MINCUBES;
	int         l = (orient) ? 1 : 0;

	if (sizex <= MAXCUBES && sizey <= MAXCUBES && sizez <= MAXCUBES &&
	    (counter < rubikRecord[l][i][j][k].score ||
	    rubikRecord[l][i][j][k].score <= NEVER)) {
		ReadRecords();	/* Maybe its been updated by another */
		rubikRecord[l][i][j][k].score = counter;
		(void) strcpy(rubikRecord[l][i][j][k].name, usernameDsp);
		if ((sizex <= 2 && sizey <= 2) || (sizey <= 2 && sizez <= 2) ||
                    (sizez <= 2 && sizex <= 2) || (orient &&
			       (counter < rubikRecord[!l][i][j][k].score ||
			       rubikRecord[!l][i][j][k].score <= NEVER))) {
			rubikRecord[!l][i][j][k].score = counter;
			(void) strcpy(rubikRecord[!l][i][j][k].name, usernameDsp);
		}
		WriteRecords();
		PrintRecord(sizex, sizey, sizez, orient, False);
		return True;
	}
	return False;
}

/* There's probably a better way to assure that they are the same but I don't
   know it off hand. */
static void
MakeEquivalent(String * username, int *sizex, int *sizey, int *sizez, Boolean * orient, Boolean * practice, int *delay)
{
	Boolean     mono, reverse;
	Pixel       foreground, background, pieceBorder;
	String      faceColor[MAXFACES];

	XtVaGetValues(rubik2d,
		      XtNuserName, username,
		      XtNsizex, sizex,
		      XtNsizey, sizey,
		      XtNsizez, sizez,
		      XtNorient, orient,
		      XtNpractice, practice,
		      XtNmono, &mono,
		      XtNreverse, &reverse,
		      XtNdelay, delay,
		      XtNforeground, &foreground,
		      XtNbackground, &background,
		      XtNpieceBorder, &pieceBorder,
		      XtNfaceColor0, &(faceColor[0]),
		      XtNfaceColor1, &(faceColor[1]),
		      XtNfaceColor2, &(faceColor[2]),
		      XtNfaceColor3, &(faceColor[3]),
		      XtNfaceColor4, &(faceColor[4]),
		      XtNfaceColor5, &(faceColor[5]), NULL);
	XtVaSetValues(rubik2d,
		      XtNdirection, RUBIK_IGNORE,
		      XtNstart, False, NULL);
	XtVaSetValues(rubik3d,
		      XtNuserName, *username,
		      XtNsizex, *sizex,
		      XtNsizey, *sizey,
		      XtNsizez, *sizez,
		      XtNorient, *orient,
		      XtNmono, mono,
		      XtNreverse, reverse,
		      XtNdelay, *delay,
		      XtNdirection, RUBIK_IGNORE,
		      XtNpractice, *practice,
		      XtNstart, False,
		      XtNforeground, foreground,
		      XtNbackground, background,
		      XtNpieceBorder, pieceBorder,
		      XtNfaceColor0, faceColor[0],
		      XtNfaceColor1, faceColor[1],
		      XtNfaceColor2, faceColor[2],
		      XtNfaceColor3, faceColor[3],
		      XtNfaceColor4, faceColor[4],
		      XtNfaceColor5, faceColor[5], NULL);
}

static void
Initialize(void)
{
	int         sizex, sizey, sizez, delay;
	Boolean     orient, practice;
	String      username;

	MakeEquivalent(&username, &sizex, &sizey, &sizez, &orient, &practice, &delay);
#ifdef HAVE_MOTIF
	if (sizex > MAXCUBES)
		XtVaSetValues(cubesx, XmNmaximum, sizex, NULL);
	XmScaleSetValue(cubesx, sizex);
	if (sizey > MAXCUBES)
		XtVaSetValues(cubesy, XmNmaximum, sizey, NULL);
	XmScaleSetValue(cubesy, sizey);
	if (sizez > MAXCUBES)
		XtVaSetValues(cubesz, XmNmaximum, sizez, NULL);
	XmScaleSetValue(cubesz, sizez);
	XmToggleButtonSetState(orientSwitch, orient, True);
	XmToggleButtonSetState(practiceSwitch, practice, True);
	XmScaleSetValue(speed, MAXSPEED + MINSPEED - delay / MULTSPEED - 1);
#endif
	InitRecords();
	ReadRecords();
	(void) strcpy(usernameDsp, username);
	if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
	    !strcmp(usernameDsp, NOACCESS) || !strcmp(usernameDsp, NOBODY)) {
		/* The NOACCESS is not necessary, but it stops people from being cute. */
		(void) sprintf(usernameDsp, "%s", getlogin());
		if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
		    !strcmp(usernameDsp, NOACCESS))
			(void) sprintf(usernameDsp, "%s", NOBODY);	/* It really IS nobody */
	}
	PrintRecord(sizex, sizey, sizez, orient, practice);
#ifndef HAVE_MOTIF
	PrintState(XtParent(rubik2d), 2, sizex, sizey, sizez,
		   movesDsp, messageDsp);
	PrintState(XtParent(rubik3d), 3, sizex, sizey, sizez,
		   movesDsp, messageDsp);
#endif
}

static void
CallbackRubik(Widget w, caddr_t clientData, rubikCallbackStruct * callData)
{
	int         sizex, sizey, sizez;
	Boolean     orient, practice, start;
	Widget      otherw;
#ifndef HAVE_MOTIF
	int         dim, otherdim;
#endif

	if (w == rubik2d) {
		otherw = rubik3d;
#ifndef HAVE_MOTIF
		dim = 2, otherdim = 3;
#endif
	} else {	/* (w == rubik3d) */
		otherw = rubik2d;
#ifndef HAVE_MOTIF
		dim = 3, otherdim = 2;
#endif
	}
	XtVaGetValues(w,
		      XtNsizex, &sizex,
		      XtNsizey, &sizey,
		      XtNsizez, &sizez,
		      XtNorient, &orient,
		      XtNpractice, &practice,
		      XtNstart, &start, NULL);
	(void) strcpy(messageDsp, "");
	switch (callData->reason) {
		case RUBIK_RESTORE:
			XtSetArg(arg[0], XtNdirection, RUBIK_RESTORE);
			XtSetValues(otherw, arg, 1);
			XtSetValues(w, arg, 1);
			movesDsp = 0;
			break;
		case RUBIK_CLEAR:
			XtSetArg(arg[0], XtNdirection, RUBIK_CLEAR);
			XtSetValues(otherw, arg, 1);
			XtSetValues(w, arg, 1);
			movesDsp = 0;
			break;
		case RUBIK_RESET:
			movesDsp = 0;
			break;
		case RUBIK_ILLEGAL:
			if (practice || start)
				(void) strcpy(messageDsp, "Illegal move");
			else
				(void) strcpy(messageDsp, "Randomize to start");
			break;
		case RUBIK_MOVED:
			movesDsp++;
			XtSetArg(arg[0], XtNstart, True);
			XtSetArg(arg[1], XtNface, callData->face);
			XtSetArg(arg[2], XtNpos, callData->position);
			XtSetArg(arg[3], XtNdirection, callData->direction);
			XtSetValues(otherw, arg, 4);
			XtSetValues(w, arg, 1);
			break;
		case RUBIK_CONTROL:
			XtSetArg(arg[0], XtNface, callData->face);
			XtSetArg(arg[1], XtNpos, callData->position);
			XtSetArg(arg[2], XtNdirection, callData->direction);
			XtSetValues(otherw, arg, 3);
			return;
		case RUBIK_SOLVED:
			if (practice)
				movesDsp = 0;
			else {
				if (HandleSolved(movesDsp, sizex, sizey, sizez, orient))
					(void) sprintf(messageDsp, "Congratulations %s!!", usernameDsp);
				else
					(void) strcpy(messageDsp, "Solved!");
			}
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			break;
		case RUBIK_PRACTICE:
			movesDsp = 0;
			practice = !practice;
			if (!practice)
				(void) strcpy(messageDsp, "Randomize to start");
			XtSetArg(arg[0], XtNpractice, practice);
			XtSetArg(arg[1], XtNstart, False);
			XtSetValues(w, arg, 2);
			XtSetValues(otherw, arg, 2);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(practiceSwitch, practice, True);
#endif
			break;
		case RUBIK_RANDOMIZE:
			movesDsp = 0;
			XtSetArg(arg[0], XtNpractice, False);
			XtSetArg(arg[1], XtNstart, False);
			XtSetValues(w, arg, 2);
			XtSetValues(otherw, arg, 2);
			break;
		case RUBIK_ORIENT:
			movesDsp = 0;
			orient = !orient;
			XtSetArg(arg[0], XtNorient, orient);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			XmToggleButtonSetState(orientSwitch, orient, True);
#endif
			break;
		case RUBIK_DECX:
			movesDsp = 0;
			sizex--;
			XtSetArg(arg[0], XtNsizex, sizex);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			XmScaleSetValue(cubesx, sizex);
			if (sizex >= MAXCUBES)
				XtVaSetValues(cubesx, XmNmaximum, sizex, NULL);
#endif
			break;
		case RUBIK_INCX:
			movesDsp = 0;
			sizex++;
			XtSetArg(arg[0], XtNsizex, sizex);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			if (sizex > MAXCUBES)
				XtVaSetValues(cubesx, XmNmaximum, sizex, NULL);
			XmScaleSetValue(cubesx, sizex);
#endif
			break;
		case RUBIK_DECY:
			movesDsp = 0;
			sizey--;
			XtSetArg(arg[0], XtNsizey, sizey);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			XmScaleSetValue(cubesy, sizey);
			if (sizey >= MAXCUBES)
				XtVaSetValues(cubesy, XmNmaximum, sizey, NULL);
#endif
			break;
		case RUBIK_INCY:
			movesDsp = 0;
			sizey++;
			XtSetArg(arg[0], XtNsizey, sizey);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			if (sizey > MAXCUBES)
				XtVaSetValues(cubesy, XmNmaximum, sizey, NULL);
			XmScaleSetValue(cubesy, sizey);
#endif
			break;
		case RUBIK_DECZ:
			movesDsp = 0;
			sizez--;
			XtSetArg(arg[0], XtNsizez, sizez);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			XmScaleSetValue(cubesz, sizez);
			if (sizez >= MAXCUBES)
				XtVaSetValues(cubesz, XmNmaximum, sizez, NULL);
#endif
			break;
		case RUBIK_INCZ:
			movesDsp = 0;
			sizez++;
			XtSetArg(arg[0], XtNsizez, sizez);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			PrintRecord(sizex, sizey, sizez, orient, practice);
#ifdef HAVE_MOTIF
			if (sizez > MAXCUBES)
				XtVaSetValues(cubesz, XmNmaximum, sizez, NULL);
			XmScaleSetValue(cubesz, sizez);
#endif
			break;
		case RUBIK_COMPUTED:
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			XtSetValues(otherw, arg, 1);
			break;
		case RUBIK_UNDO:
			movesDsp--;
			XtSetArg(arg[0], XtNstart, True);
			XtSetArg(arg[1], XtNface, callData->face);
			XtSetArg(arg[2], XtNpos, callData->position);
			XtSetArg(arg[3], XtNdirection, callData->direction);
			XtSetValues(otherw, arg, 4);
			XtSetValues(w, arg, 1);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
#else
	PrintState(XtParent(w), dim, sizex, sizey, sizez,
		   movesDsp, messageDsp);
	PrintState(XtParent(otherw), otherdim, sizex, sizey, sizez,
		   movesDsp, messageDsp);
#endif
}

#ifdef HAVE_MOTIF
static void
CubesXSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         sizex = cbs->value, old, sizey, sizez;
	Boolean     orient, practice;

	XtVaGetValues(rubik2d,
		      XtNsizex, &old,
		      XtNsizey, &sizey,
		      XtNsizez, &sizez,
		      XtNorient, &orient,
		      XtNpractice, &practice, NULL);
	if (old != sizex) {
		XtVaSetValues(rubik2d,
			      XtNsizex, sizex, NULL);
		XtVaSetValues(rubik3d,
			      XtNsizex, sizex, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(sizex, sizey, sizez, orient, practice);
	}
}

static void
CubesYSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         sizey = cbs->value, old, sizex, sizez;
	Boolean     orient, practice;

	XtVaGetValues(rubik2d,
		      XtNsizex, &sizex,
		      XtNsizey, &old,
		      XtNsizez, &sizez,
		      XtNorient, &orient,
		      XtNpractice, &practice, NULL);
	if (old != sizey) {
		XtVaSetValues(rubik2d,
			      XtNsizey, sizey, NULL);
		XtVaSetValues(rubik3d,
			      XtNsizey, sizey, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(sizex, sizey, sizez, orient, practice);
	}
}

static void
CubesZSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         sizez = cbs->value, old, sizex, sizey;
	Boolean     orient, practice;

	XtVaGetValues(rubik2d,
		      XtNsizex, &sizex,
		      XtNsizey, &sizey,
		      XtNsizez, &old,
		      XtNorient, &orient,
		      XtNpractice, &practice, NULL);
	if (old != sizez) {
		XtVaSetValues(rubik2d,
			      XtNsizez, sizez, NULL);
		XtVaSetValues(rubik3d,
			      XtNsizez, sizez, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(sizex, sizey, sizez, orient, practice);
	}
}

static void
OrientToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         sizex, sizey, sizez;
	Boolean     orient = cbs->set, practice;

	XtVaGetValues(rubik2d,
		      XtNsizex, &sizex,
		      XtNsizey, &sizey,
		      XtNsizez, &sizez,
		      XtNpractice, &practice, NULL);
	XtVaSetValues(rubik2d,
		      XtNorient, orient, NULL);
	XtVaSetValues(rubik3d,
		      XtNorient, orient, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	PrintRecord(sizex, sizey, sizez, orient, practice);
}

static void
PracticeToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct * cbs)
{
	int         sizex, sizey, sizez;
	Boolean     orient, practice = cbs->set;

	XtVaSetValues(rubik2d,
		      XtNpractice, practice,
		      XtNstart, False, NULL);
	XtVaSetValues(rubik3d,
		      XtNpractice, practice,
		      XtNstart, False, NULL);
	XtVaGetValues(rubik2d,
		      XtNsizex, &sizex,
		      XtNsizey, &sizey,
		      XtNsizez, &sizez,
		      XtNpractice, &orient, NULL);
	movesDsp = 0;
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
	if (!practice)
		(void) strcpy(messageDsp, "Randomize to start");
	PrintRecord(sizex, sizey, sizez, orient, practice);
}

static void
SpeedSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int delay = MULTSPEED * (MAXSPEED + MINSPEED - cbs->value - 1), oldDelay;

	XtVaGetValues(rubik2d,
		      XtNdelay, &oldDelay, NULL);
	if (oldDelay != delay) {
		XtVaSetValues(rubik3d, XtNdelay, delay, NULL);
		XtVaSetValues(rubik2d, XtNdelay, delay, NULL);
	}
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2)
		exit(0);
	XtSetArg(arg[0], XtNmenu, val);
	XtSetValues(rubik2d, arg, 1);
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;
	XtSetArg(arg[0], XtNmenu, val + 3);  /* GWQ */
	XtSetValues(rubik2d, arg, 1);
}

static const char aboutHelp[] = {
	"Version 5.6.1 by David Bagley <bagleyd@tux.org> http://www.tux.org/~bagleyd"
};

static const char helpHelp[] = {
	"Consult man page.  :)"
};

static void
helpCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		(void) printf("%s: %s\n", progDsp, aboutHelp);
	else
		(void) printf("%s: %s\n", progDsp, helpHelp);
}
#endif

int
main(int argc, char **argv)
{
	Widget      topLevel;
#ifdef HAVE_MOTIF
	Widget      menubar, panel, panel2, panel3;
	Widget      rowcol, rowcol2, rowcol3, rowcol4;
	Pixmap      mouseLeftCursor, mouseRightCursor;
	Pixel       fg, bg;
	XmString    file, play, help, about;
	XmString    get, write, quit;
	XmString    clear, undo, randomize, solve, orient, practice;
#else
	Widget      shell;
#endif

	progDsp = argv[0];
	topLevel = XtInitialize(argv[0], "Rubik",
				options, XtNumber(options), &argc, argv);
	if (argc != 1)
		usage(argv[0]);

	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(topLevel),
				      RootWindowOfScreen(XtScreen(topLevel)),
			    (char *) rubik_bits, rubik_width, rubik_height));
#ifdef HAVE_MOTIF
	XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER);	/* not XmEXPLICIT */
	XtSetValues(topLevel, arg, 2);
	panel = XtVaCreateManagedWidget("panel",
				xmPanedWindowWidgetClass, topLevel,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	file = XmStringCreateSimple((char *) "File");
	play = XmStringCreateSimple((char *) "Play");
	help = XmStringCreateSimple((char *) "Help");
	menubar = XmVaCreateSimpleMenuBar(panel, (char *) "menubar",
		XmVaCASCADEBUTTON, file, 'F',
		XmVaCASCADEBUTTON, play, 'P',
		XmVaCASCADEBUTTON, help, 'H',
		NULL);
	XmStringFree(file);
	XmStringFree(play);
	get = XmStringCreateSimple((char *) "Get");
	write = XmStringCreateSimple((char *) "Write");
	quit = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "file_menu", 0, fileCB,
		XmVaPUSHBUTTON, get, 'G', NULL, NULL,
		XmVaPUSHBUTTON, write, 'W', NULL, NULL,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quit, 'Q', NULL, NULL,
		NULL);
	XmStringFree(get);
	XmStringFree(write);
	XmStringFree(quit);
	clear = XmStringCreateSimple((char *) "Clear");
	undo = XmStringCreateSimple((char *) "Undo");
	randomize = XmStringCreateSimple((char *) "Randomize");
	solve = XmStringCreateSimple((char *) "Solve");
	orient = XmStringCreateSimple((char *) "Orientize");
	practice = XmStringCreateSimple((char *) "Practice");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, clear, 'C', NULL, NULL,
		XmVaPUSHBUTTON, undo, 'U', NULL, NULL,
		XmVaPUSHBUTTON, randomize, 'R', NULL, NULL,
		XmVaPUSHBUTTON, solve, 'S', NULL, NULL,
		XmVaPUSHBUTTON, orient, 'O', NULL, NULL,
		XmVaPUSHBUTTON, practice, 'P', NULL, NULL,
		NULL);
	XmStringFree(clear);
	XmStringFree(undo);
	XmStringFree(randomize);
	XmStringFree(solve);
	XmStringFree(orient);
	XmStringFree(practice);
	about = XmStringCreateSimple((char *) "About");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "help_menu", 2, helpCB,
		XmVaPUSHBUTTON, about, 'A', NULL, NULL,
		XmVaPUSHBUTTON, help, 'H', NULL, NULL,
		NULL);
	XmStringFree(about);
	XmStringFree(help);
	XtManageChild(menubar);
	panel2 = XtCreateManagedWidget("panel2",
				       xmPanedWindowWidgetClass, panel,
				       NULL, 0);
	panel3 = XtVaCreateManagedWidget("panel3",
					 xmPanedWindowWidgetClass, panel2,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	rowcol = XtVaCreateManagedWidget("Rowcol",
					 xmRowColumnWidgetClass, panel3,
					 XmNnumColumns, 2,
					 XmNorientation, XmHORIZONTAL,
					 XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(rowcol,
		      XmNforeground, &fg,
		      XmNbackground, &bg, NULL);
	mouseLeftCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	      RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_left_bits,
				 mouse_left_width, mouse_left_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	mouseRightCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	     RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_right_bits,
			       mouse_right_width, mouse_right_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	(void) XtVaCreateManagedWidget("mouseLeftText",
				xmLabelGadgetClass, rowcol,
		   XtVaTypedArg, XmNlabelString, XmRString, "Move", 5, NULL);
	(void) XtVaCreateManagedWidget("mouseLeft",
				xmLabelGadgetClass, rowcol,
	      XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseLeftCursor, NULL);
	(void) XtVaCreateManagedWidget("mouseRightText", xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Randomize", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseRight",
				xmLabelGadgetClass, rowcol,
				XmNlabelType, XmPIXMAP,
				XmNlabelPixmap, mouseRightCursor, NULL);
	(void) XtVaCreateManagedWidget("movesText", xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Moves", 6, NULL);
	moves = XtVaCreateManagedWidget("0",
					xmLabelWidgetClass, rowcol, NULL);
	(void) XtVaCreateManagedWidget("recordText",
				xmLabelGadgetClass, rowcol,
		 XtVaTypedArg, XmNlabelString, XmRString, "Record", 7, NULL);
	record = XtVaCreateManagedWidget("0",
					 xmLabelWidgetClass, rowcol, NULL);

	rowcol2 = XtVaCreateManagedWidget("Rowcol2",
				          xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN,
					  NULL);
	cubesx = XtVaCreateManagedWidget("cubesx",
					 xmScaleWidgetClass, rowcol2,
			XtVaTypedArg, XmNtitleString, XmRString, "Size X", 8,
					 XmNminimum, MINCUBES,
					 XmNmaximum, MAXCUBES,
					 XmNvalue, MINCUBES,
					 XmNshowValue, True,
					 XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(cubesx,
		      XmNvalueChangedCallback, (XtCallbackProc) CubesXSlider, (XtPointer) NULL);
	cubesy = XtVaCreateManagedWidget("cubesy",
					 xmScaleWidgetClass, rowcol2,
			XtVaTypedArg, XmNtitleString, XmRString, "Size Y", 8,
					 XmNminimum, MINCUBES,
					 XmNmaximum, MAXCUBES,
					 XmNvalue, MINCUBES,
					 XmNshowValue, True,
					 XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(cubesy,
		      XmNvalueChangedCallback, (XtCallbackProc) CubesYSlider, (XtPointer) NULL);
	cubesz = XtVaCreateManagedWidget("cubesz",
					 xmScaleWidgetClass, rowcol2,
			XtVaTypedArg, XmNtitleString, XmRString, "Size Z", 8,
					 XmNminimum, MINCUBES,
					 XmNmaximum, MAXCUBES,
					 XmNvalue, MINCUBES,
					 XmNshowValue, True,
					 XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(cubesz,
		      XmNvalueChangedCallback, (XtCallbackProc) CubesZSlider, (XtPointer) NULL);
	rowcol3 = XtVaCreateManagedWidget("Rowcol3",
					  xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	orientSwitch = XtVaCreateManagedWidget("Oriented",
				   xmToggleButtonWidgetClass, rowcol3, NULL);
	XtAddCallback(orientSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) OrientToggle, (XtPointer) NULL);
	practiceSwitch = XtVaCreateManagedWidget("Practice",
				   xmToggleButtonWidgetClass, rowcol3, NULL);
	XtAddCallback(practiceSwitch,
		      XmNvalueChangedCallback, (XtCallbackProc) PracticeToggle, (XtPointer) NULL);
	speed = XtVaCreateManagedWidget("speed",
					 xmScaleWidgetClass, rowcol3,
		      XtVaTypedArg, XmNtitleString, XmRString, "Animation Speed", 8,
					 XmNminimum, MINSPEED,
					 XmNmaximum, MAXSPEED,
					 XmNvalue, MAXSPEED - 10,
					 XmNshowValue, True,
					 XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(speed,
		      XmNvalueChangedCallback, (XtCallbackProc) SpeedSlider, (XtPointer) NULL);
	rowcol4 = XtVaCreateManagedWidget("Rowcol4",
				       xmRowColumnWidgetClass, panel3, NULL);
	message = XtVaCreateManagedWidget("Play Rubik's Cube! (use mouse and keypad)",
					  xmLabelWidgetClass, rowcol4, NULL);

	rubik2d = XtCreateManagedWidget("rubik",
					rubik2dWidgetClass, panel2, NULL, 0);
	XtVaSetValues(rubik2d,
		      XtNheight, 200, NULL);
	XtAddCallback(rubik2d,
	  XtNselectCallback, (XtCallbackProc) CallbackRubik, (XtPointer) NULL);
	rubik3d = XtCreateManagedWidget("rubik",
					rubik3dWidgetClass, panel2, NULL, 0);
	XtVaSetValues(rubik3d,
		      XtNheight, 200, NULL);
#else
	XtSetArg(arg[1], XtNinput, True);
	XtSetValues(topLevel, arg, 2);
	shell = XtCreateApplicationShell(argv[0],
					 topLevelShellWidgetClass, NULL, 0);
	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(shell),
					 RootWindowOfScreen(XtScreen(shell)),
			    (char *) rubik_bits, rubik_width, rubik_height));
	XtSetValues(shell, arg, 2);
	rubik2d = XtCreateManagedWidget("rubik",
				      rubik2dWidgetClass, topLevel, NULL, 0);
	XtAddCallback(rubik2d,
	  XtNselectCallback, (XtCallbackProc) CallbackRubik, (XtPointer) NULL);
	rubik3d = XtCreateManagedWidget("rubik",
					rubik3dWidgetClass, shell, NULL, 0);
#endif
	XtAddCallback(rubik3d,
	  XtNselectCallback, (XtCallbackProc) CallbackRubik, (XtPointer) NULL);
	Initialize();
	XtRealizeWidget(topLevel);
#ifndef HAVE_MOTIF
	XtRealizeWidget(shell);
#endif
	XGrabButton(XtDisplay(rubik2d), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(rubik2d), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(rubik2d),
		    XCreateFontCursor(XtDisplay(rubik2d), XC_crosshair));
	XGrabButton(XtDisplay(rubik3d), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(rubik3d), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(rubik3d),
		    XCreateFontCursor(XtDisplay(rubik3d), XC_crosshair));
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
