/*
# X-BASED THREED
#
#  Threed.c
#
###
#
#  Copyright (c) 1994 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "useful",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Methods file for Threed */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef VMS
#define M_PI 3.14159265358979323846
#endif
#include <X11/IntrinsicP.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/CoreP.h>
#include "ThreedP.h"

#ifndef DATAFILE
#define DATAFILE "/usr/games/lib/threed.data"
#endif

static void InitializeThreeD (Widget request, Widget renew);
static void ExposeThreeD (Widget renew, XEvent * event, Region region);
static void ResizeThreeD (ThreeDWidget w);
static void DestroyThreeD (Widget old);
static Boolean SetValuesThreeD (Widget current, Widget request, Widget renew);

static void QuitThreeD (ThreeDWidget w, XEvent * event, char **args,
			int n_args);
static void SurfaceModeThreeD (ThreeDWidget w, XEvent * event, char **args,
			       int n_args);
static void ObjectModeThreeD (ThreeDWidget w, XEvent * event, char **args,
			      int n_args);
static void MoveOutThreeD (ThreeDWidget w, XEvent * event, char **args,
			   int n_args);
static void MoveLowerPhiThreeD (ThreeDWidget w, XEvent * event, char **args,
				int n_args);
static void MoveLowerThetaThreeD (ThreeDWidget w, XEvent * event, char **args,
				  int n_args);
static void MoveInThreeD (ThreeDWidget w, XEvent * event, char **args,
			  int n_args);
static void MoveRaiseThetaThreeD (ThreeDWidget w, XEvent * event, char **args,
				  int n_args);
static void MoveRaisePhiThreeD (ThreeDWidget w, XEvent * event, char **args,
				int n_args);
static void ReadThreeD (ThreeDWidget w);
static void CheckGraphicsThreeD (ThreeDWidget w);
static void RepaintGraphicsThreeD (ThreeDWidget w);

static char defaultTranslationsThreeD[] = "<KeyPress>q: quit()\n\
   Ctrl<KeyPress>C: quit()\n\
   <KeyPress>o: moveOut()\n\
   <KeyPress>KP_Divide: moveOut()\n\
   <KeyPress>R5: moveOut()\n\
   <KeyPress>Up: moveRaisePhi()\n\
   <KeyPress>osfUp: moveRaisePhi()\n\
   <KeyPress>KP_Up: moveRaisePhi()\n\
   <KeyPress>KP_8: moveRaisePhi()\n\
   <KeyPress>R8: moveRaisePhi()\n\
   <KeyPress>Left: moveRaiseTheta()\n\
   <KeyPress>osfLeft: moveRaiseTheta()\n\
   <KeyPress>KP_Left: moveRaiseTheta()\n\
   <KeyPress>KP_4: moveRaiseTheta()\n\
   <KeyPress>R10: moveRaiseTheta()\n\
   <KeyPress>i: moveIn()\n\
   <KeyPress>Begin: moveIn()\n\
   <KeyPress>KP_5: moveIn()\n\
   <KeyPress>R11: moveIn()\n\
   <KeyPress>Right: moveLowerTheta()\n\
   <KeyPress>osfRight: moveLowerTheta()\n\
   <KeyPress>KP_Right: moveLowerTheta()\n\
   <KeyPress>KP_6: moveLowerTheta()\n\
   <KeyPress>R12: moveLowerTheta()\n\
   <KeyPress>Down: moveLowerPhi()\n\
   <KeyPress>osfDown: moveLowerPhi()\n\
   <KeyPress>KP_Down: moveLowerPhi()\n\
   <KeyPress>KP_2: moveLowerPhi()\n\
   <KeyPress>s: surface()\n\
   <KeyPress>b: object()";

static XtActionsRec actionsListThreeD[] = {
  {(char *) "quit", (XtActionProc) QuitThreeD},
  {(char *) "moveOut", (XtActionProc) MoveOutThreeD},
  {(char *) "moveLowerPhi", (XtActionProc) MoveLowerPhiThreeD},
  {(char *) "moveLowerTheta", (XtActionProc) MoveLowerThetaThreeD},
  {(char *) "moveIn", (XtActionProc) MoveInThreeD},
  {(char *) "moveRaiseTheta", (XtActionProc) MoveRaiseThetaThreeD},
  {(char *) "moveRaisePhi", (XtActionProc) MoveRaisePhiThreeD},
  {(char *) "surface", (XtActionProc) SurfaceModeThreeD},
  {(char *) "object", (XtActionProc) ObjectModeThreeD},
};

static XtResource resourcesThreeD[] = {
  {XtNwidth, XtCWidth, XtRDimension, sizeof (Dimension),
   XtOffset (ThreeDWidget, core.width), XtRString, (caddr_t) "400"}
  ,
  {XtNheight, XtCHeight, XtRDimension, sizeof (Dimension),
   XtOffset (ThreeDWidget, core.height), XtRString, (caddr_t) "200"}
  ,
  {XtNdistance, XtCDistance, XtRInt, sizeof (int),
   XtOffset (ThreeDWidget, threed.distance), XtRString, (caddr_t) "1"},
  {XtNthetaDegrees, XtCThetaDegrees, XtRInt, sizeof (int),
   XtOffset (ThreeDWidget, threed.thetaDegrees), XtRString, (caddr_t) "0"},
  {XtNphiDegrees, XtCPhiDegrees, XtRInt, sizeof (int),
   XtOffset (ThreeDWidget, threed.phiDegrees), XtRString, (caddr_t) "0"},
  {XtNsurface, XtCSurface, XtRBoolean, sizeof (Boolean),
   XtOffset (ThreeDWidget, threed.surface), XtRString, (caddr_t) "TRUE"}
  ,
  {XtNobject, XtCObject, XtRInt, sizeof (int),
   XtOffset (ThreeDWidget, threed.object), XtRString, (caddr_t) "0"},
  {XtNobjectName, XtCObjectName, XtRString, sizeof (String),
   XtOffset (ThreeDWidget, threed.name), XtRString, (caddr_t) "None"}
  ,
  {XtNobjectNumber, XtCObjectNumber, XtRInt, sizeof (int),
   XtOffset (ThreeDWidget, threed.numObjects), XtRString, (caddr_t) "0"},
  {XtNobjectList, XtCObjectList, XtRString, sizeof (String),
   XtOffset (ThreeDWidget, threed.list), XtRString, (caddr_t) "None"}
  ,
  {XtNwhiteBrush, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.shadeColor[WHITE_BRUSH]), XtRString,
   (caddr_t) "Yellow"}
  ,
  {XtNltgrayBrush, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.shadeColor[LTGRAY_BRUSH]), XtRString,
   (caddr_t) "Green"}
  ,
  {XtNgrayBrush, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.shadeColor[GRAY_BRUSH]), XtRString,
   (caddr_t) "Red"}
  ,
  {XtNdkgrayBrush, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.shadeColor[DKGRAY_BRUSH]), XtRString,
   (caddr_t) "Blue"}
  ,
  {XtNblackBrush, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.shadeColor[BLACK_BRUSH]), XtRString,
   (caddr_t) "Black"}
  ,
  {XtNanotherBrush, XtCColor, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.shadeColor[ANOTHER_BRUSH]), XtRString,
   (caddr_t) "Brown"}
  ,
  {XtNforeground, XtCForeground, XtRPixel, sizeof (Pixel),
   XtOffset (ThreeDWidget, threed.foreground), XtRString, (caddr_t) "Black"}
  ,
  {XtNselectCallback, XtCCallback, XtRCallback, sizeof (caddr_t),
   XtOffset (ThreeDWidget, threed.select), XtRCallback, NULL}
};

ThreeDClassRec threedClassRec = {
  {
   (WidgetClass) & widgetClassRec,	/* superclass */
   (char *) "ThreeD",		/* class name */
   sizeof (ThreeDRec),		/* widget size */
   NULL,			/* class initialize */
   NULL,			/* class part initialize */
   FALSE,			/* class inited */
   (XtInitProc) InitializeThreeD,	/* initialize */
   NULL,			/* initialize hook */
   XtInheritRealize,		/* realize */
   actionsListThreeD,		/* actions */
   XtNumber (actionsListThreeD),	/* num actions */
   resourcesThreeD,		/* resources */
   XtNumber (resourcesThreeD),	/* num resources */
   NULLQUARK,			/* xrm class */
   TRUE,			/* compress motion */
   TRUE,			/* compress exposure */
   TRUE,			/* compress enterleave */
   TRUE,			/* visible interest */
   (XtWidgetProc) DestroyThreeD,	/* destroy */
   (XtWidgetProc) ResizeThreeD,	/* resize */
   (XtExposeProc) ExposeThreeD,	/* expose */
   (XtSetValuesFunc) SetValuesThreeD,	/* set values */
   NULL,			/* set values hook */
   XtInheritSetValuesAlmost,	/* set values almost */
   NULL,			/* get values hook */
   NULL,			/* accept focus */
   XtVersion,			/* version */
   NULL,			/* callback private */
   defaultTranslationsThreeD,	/* tm table */
   NULL,			/* query geometry */
   NULL,			/* display accelerator */
   NULL				/* extension */
   }
  ,
  {
   0				/* ignore */
   }
};

WidgetClass threedWidgetClass = (WidgetClass) & threedClassRec;
static ThreeDWidget *W;
static int O;

static void DrawObject (ThreeDWidget w, int o);
static int CompareObject (const void *elem1, const void *elem2);

static void
InitializeThreeD (Widget request, Widget renew)
{
  ThreeDWidget w = (ThreeDWidget) renew;
  XGCValues values;
  XtGCMask valueMask;
  int i, color;

  ReadThreeD (w);
  w->threed.object = 0;
  w->threed.name = w->threed.objects[w->threed.object].name;
  w->threed.list = (char **) calloc (w->threed.numObjects, sizeof (char **));
  for (i = 0; i < w->threed.numObjects; i++)
    w->threed.list[i] = w->threed.objects[i].name;
#ifdef DEBUG
  (void) printf ("%s\n", w->threed.name);
#endif
  CheckGraphicsThreeD (w);
  valueMask = GCForeground | GCBackground;
  values.foreground = w->threed.foreground;
  values.background = w->core.background_pixel;
  w->threed.graphicsGC = XtGetGC (renew, valueMask, &values);
  for (color = 0; color < COLORS; color++)
    {
      if (DefaultDepthOfScreen (XtScreen (w)) > 1)
	values.foreground = w->threed.shadeColor[color];
      w->threed.shadeGC[color] = XtGetGC (renew, valueMask, &values);
    }
  valueMask = GCForeground | GCBackground;
  values.foreground = w->core.background_pixel;
  values.background = w->threed.foreground;
  w->threed.inverseGC = XtGetGC (renew, valueMask, &values);
  ResizeThreeD (w);
}

static void
DestroyThreeD (Widget old)
{
  ThreeDWidget w = (ThreeDWidget) old;
  int color;

  for (color = 0; color < COLORS; color++)
    XtReleaseGC (old, w->threed.shadeGC[color]);
  XtReleaseGC (old, w->threed.graphicsGC);
  XtReleaseGC (old, w->threed.inverseGC);
  XtRemoveCallbacks (old, XtNselectCallback, w->threed.select);
}

static void
ResizeThreeD (ThreeDWidget w)
{
  w->threed.center.x = w->core.width / 2;
  w->threed.center.y = w->core.height / 2;
  w->threed.size = MIN (w->threed.center.x, w->threed.center.y);
}

static void
ExposeThreeD (Widget renew, XEvent * event, Region region)
		 /* Not used */
{
  ThreeDWidget w = (ThreeDWidget) renew;

  if (w->core.visible)
    RepaintGraphicsThreeD (w);
}

static Boolean
SetValuesThreeD (Widget current, Widget request, Widget renew)
{
  ThreeDWidget c = (ThreeDWidget) current, w = (ThreeDWidget) renew;
  XGCValues values;
  XtGCMask valueMask;
  Boolean redraw = FALSE;

  CheckGraphicsThreeD (w);
  if (w->threed.foreground != c->threed.foreground)
    {
      valueMask = GCForeground | GCBackground;
      values.foreground = w->threed.foreground;
      values.background = w->core.background_pixel;
      XtReleaseGC (renew, w->threed.graphicsGC);
      w->threed.graphicsGC = XtGetGC (renew, valueMask, &values);
      redraw = TRUE;
    }
  if (w->threed.distance != c->threed.distance ||
      w->threed.thetaDegrees != c->threed.thetaDegrees ||
      w->threed.phiDegrees != c->threed.phiDegrees ||
      w->threed.surface != c->threed.surface ||
      w->threed.object != c->threed.object)
    {
      w->threed.name = w->threed.objects[w->threed.object].name;
#ifdef DEBUG
      (void) printf ("%s\n", w->threed.name);
#endif
      ResizeThreeD (w);
      redraw = TRUE;
    }
  return (redraw);
}

static void
QuitThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  XtCloseDisplay (XtDisplay (w));
  exit (0);
}

static void
MoveOutThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  if (w->threed.distance < MAXDISTANCE)
    w->threed.distance += DELTADISTANCE;
  cb.reason = THREED_SET;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
  RepaintGraphicsThreeD (w);
}

static void
MoveLowerPhiThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  w->threed.phiDegrees -= DELTADISTANCE;
  if (w->threed.phiDegrees < MINDEGREES)
    w->threed.phiDegrees += NUMDEGREES;
  cb.reason = THREED_SET;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
  RepaintGraphicsThreeD (w);
}

static void
MoveLowerThetaThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  w->threed.thetaDegrees -= DELTADISTANCE;
  if (w->threed.thetaDegrees < MINDEGREES)
    w->threed.thetaDegrees += NUMDEGREES;
  cb.reason = THREED_SET;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
  RepaintGraphicsThreeD (w);
}

static void
MoveRaiseThetaThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  w->threed.thetaDegrees += DELTADEGREES;
  if (w->threed.thetaDegrees > MAXDEGREES)
    w->threed.thetaDegrees -= NUMDEGREES;
  cb.reason = THREED_SET;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
  RepaintGraphicsThreeD (w);
}

static void
MoveRaisePhiThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  w->threed.phiDegrees += DELTADEGREES;
  if (w->threed.phiDegrees > MAXDEGREES)
    w->threed.phiDegrees -= NUMDEGREES;
  cb.reason = THREED_SET;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
  RepaintGraphicsThreeD (w);
}

static void
MoveInThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  if (w->threed.distance > MINDISTANCE)
    w->threed.distance -= DELTADEGREES;
  cb.reason = THREED_SET;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
  RepaintGraphicsThreeD (w);
}

static void
ReadThreeD (ThreeDWidget w)
{
  FILE *fp;
  char c;
  int i, j, k, l;
  double ox, oy, oz;		/* DUMMY variables */

  if ((fp = fopen (DATAFILE, "r")) == NULL)
    (void) printf ("Can not read %s to enter.\n", DATAFILE);
  else
    {
      while ((c = getc (fp)) != EOF && c != SYMBOL);
      (void) fscanf (fp, "%d", &(w->threed.numObjects));
      if (!(w->threed.objects = (OBJECT *)
	    malloc (w->threed.numObjects * sizeof (OBJECT))))
	XtError ("Not enough memory, exiting.");
      while ((c = getc (fp)) != EOF && c != SYMBOL);
      (void) fscanf (fp, "%d %d %d", &(w->threed.distance),
		     &(w->threed.thetaDegrees), &(w->threed.phiDegrees));
      while ((c = getc (fp)) != EOF && c != SYMBOL);
      (void) fscanf (fp, "%lg %lg %lg", &ox, &oy, &oz);
#ifdef DEBUG
      (void) printf ("OBJECTS%c %d\n", SYMBOL, w->threed.numObjects);
      (void) printf ("Distance Theta Edges%c\n %d %d %d\n", SYMBOL,
		     w->threed.distance, w->threed.thetaDegrees,
		     w->threed.phiDegrees);
      (void) printf ("Origin%c  %g %g %g\n", SYMBOL, ox, oy, oz);
#endif
      for (i = 0; i < w->threed.numObjects; i++)
	{
	  while ((c = getc (fp)) != EOF && c != SYMBOL);
	  (void) fscanf (fp, "\t%s", w->threed.objects[i].name);
	  while ((c = getc (fp)) != EOF && c != SYMBOL);
	  (void) fscanf (fp, "%d %d %d", &(w->threed.objects[i].numVertices),
			 &(w->threed.objects[i].numSurfaces),
			 &(w->threed.objects[i].numSurfaceVertices));
	  if (!(w->threed.objects[i].world = (POINT3D *)
		malloc (w->threed.objects[i].numVertices * sizeof (POINT3D))))
	    XtError ("Not enough memory, exiting.");
	  if (!(w->threed.objects[i].vertex = (VERTEX *)
		malloc (w->threed.objects[i].numVertices * sizeof (VERTEX))))
	    XtError ("Not enough memory, exiting.");
	  if (!(w->threed.objects[i].info = (SURFACE *)
		malloc (w->threed.objects[i].numSurfaces * sizeof (SURFACE))))
	    XtError ("Not enough memory, exiting.");
	  if (!(w->threed.objects[i].map = (int *)
		malloc (w->threed.objects[i].numSurfaceVertices *
			sizeof (int))))
	    XtError ("Not enough memory, exiting.");
#ifdef DEBUG
	  (void) printf ("\nNAME%c %s\n", SYMBOL, w->threed.objects[i].name);
	  (void) printf ("Vertices Surfaces SurfaceVertices%c\n%d\t%d\t%d\n",
			 SYMBOL, w->threed.objects[i].numVertices,
			 w->threed.objects[i].numSurfaces,
			 w->threed.objects[i].numSurfaceVertices);
#endif

	  while ((c = getc (fp)) != EOF && c != SYMBOL);
	  for (j = 0; j < w->threed.objects[i].numVertices; j++)
	    (void) fscanf (fp, "%lg %lg %lg",
			   &(w->threed.objects[i].world[j].x),
			   &(w->threed.objects[i].world[j].y),
			   &(w->threed.objects[i].world[j].z));
#ifdef DEBUG
	  (void) printf ("\nVERTICES%c\n", SYMBOL);
	  for (j = 0; j < w->threed.objects[i].numVertices; j++)
	    (void) printf ("%g\t%g\t%g\n",
			   w->threed.objects[i].world[j].x,
			   w->threed.objects[i].world[j].y,
			   w->threed.objects[i].world[j].z);
#endif
	  while ((c = getc (fp)) != EOF && c != SYMBOL);
	  l = 0;
	  for (j = 0; j < w->threed.objects[i].numSurfaces; j++)
	    {
	      (void) fscanf (fp, "%d",
			     &(w->threed.objects[i].info[j].numVertices));
	      for (k = 0; k < w->threed.objects[i].info[j].numVertices; k++)
		{
		  (void) fscanf (fp, "%d", &(w->threed.objects[i].map[l]));
		  w->threed.objects[i].map[l++]--;
		}
	    }
#ifdef DEBUG
	  (void) printf ("\nSURFACESVERTICES%c\n", SYMBOL);
	  l = 0;
	  for (j = 0; j < w->threed.objects[i].numSurfaces; j++)
	    {
	      (void) printf ("%d\t",
			     w->threed.objects[i].info[j].numVertices);
	      for (k = 0; k < w->threed.objects[i].info[j].numVertices; k++)
		(void) printf ("%d\t", w->threed.objects[i].map[l++] + 1);
	      (void) printf ("\n");
	    }
	  (void) printf ("\n");
#endif
	  while ((c = getc (fp)) != EOF && c != SYMBOL);
	  for (j = 0; j < w->threed.objects[i].numSurfaces; j++)
	    (void) fscanf (fp, "%d %d %d",
			   &(w->threed.objects[i].info[j].mapIndex),
			   &(w->threed.objects[i].info[j].depthIndex),
			   &(w->threed.objects[i].info[j].brushColor));
#ifdef DEBUG
	  (void) printf ("\nSURFACES%c\n", SYMBOL);
	  for (j = 0; j < w->threed.objects[i].numSurfaces; j++)
	    (void) printf ("%d\t%d\t%d\n",
			   w->threed.objects[i].info[j].mapIndex,
			   w->threed.objects[i].info[j].depthIndex,
			   w->threed.objects[i].info[j].brushColor);
#endif
	}
    }
}

static void
SurfaceModeThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  cb.reason = THREED_SURFACE;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
}

static void
ObjectModeThreeD (ThreeDWidget w, XEvent * event, char **args, int n_args)
{
  threedCallbackStruct cb;

  cb.reason = THREED_OBJECT;
  XtCallCallbacks ((Widget) w, XtNselectCallback, &cb);
}

static void
CheckGraphicsThreeD (ThreeDWidget w)
{
  char buf[121];

  if (w->threed.distance < MINDISTANCE || w->threed.distance > MAXDISTANCE)
    {
      (void) sprintf (buf, "Distance of 3D %d out of bounds, use %d..%d",
		      w->threed.distance, MINDISTANCE, MAXDISTANCE);
      XtWarning (buf);
      w->threed.distance = (MAXDISTANCE + MINDISTANCE) / 2;
    }
  if (w->threed.thetaDegrees < MINDEGREES ||
      w->threed.thetaDegrees > MAXDEGREES)
    w->threed.thetaDegrees = w->threed.thetaDegrees % NUMDEGREES;
  if (w->threed.phiDegrees < MINDEGREES || w->threed.phiDegrees > MAXDEGREES)
    w->threed.phiDegrees = w->threed.phiDegrees % NUMDEGREES;
  if (w->threed.object < 0 || w->threed.object >= w->threed.numObjects)
    {
      (void) sprintf (buf, "Object types of 3D %d, out of bounds, use %d..%d",
		      w->threed.object, 0, w->threed.numObjects - 1);
      XtWarning (buf);
      w->threed.object = 0;
    }
}

static void
RepaintGraphicsThreeD (ThreeDWidget w)
{
  XClearWindow (XtDisplay (w), XtWindow (w));
  DrawObject (w, w->threed.object);
}

static void
DrawObject (ThreeDWidget w, int o)
{
  double sinTheta, cosTheta, sinPhi, cosPhi;
  double s1 = 0.0, s2 = 0.0, s3 = 0.0;
  POINT3D *v1, *v2, *v3;
  POINT points[10];
  int surface, vertex, mapIndex, vertexIndex, loop;

  if (!w->threed.objects)
    return;
  cosTheta = cos (RADIANS ((double) w->threed.thetaDegrees));
  sinTheta = sin (RADIANS ((double) w->threed.thetaDegrees));
  cosPhi = cos (RADIANS ((double) w->threed.phiDegrees));
  sinPhi = sin (RADIANS ((double) w->threed.phiDegrees));
  for (loop = 0; loop < w->threed.objects[o].numVertices; loop++)
    {
      w->threed.objects[o].vertex[loop].eye.x = -
	(w->threed.objects[o].world[loop].x * sinTheta) +
	(w->threed.objects[o].world[loop].y * cosTheta);
      w->threed.objects[o].vertex[loop].eye.y = -
	(w->threed.objects[o].world[loop].x * cosTheta * cosPhi) -
	(w->threed.objects[o].world[loop].y * sinTheta * cosPhi) +
	(w->threed.objects[o].world[loop].z * sinPhi);
      w->threed.objects[o].vertex[loop].eye.z = -
	(w->threed.objects[o].world[loop].x * cosTheta * sinPhi) -
	(w->threed.objects[o].world[loop].y * sinTheta * sinPhi) -
	(w->threed.objects[o].world[loop].z * cosPhi) +
	(double) w->threed.distance;
      if (w->threed.objects[o].vertex[loop].eye.z < 0.000001 &&
	  w->threed.objects[o].vertex[loop].eye.z >= 0.0)
	w->threed.objects[o].vertex[loop].eye.z = 0.000001;
      if (w->threed.objects[o].vertex[loop].eye.z < 0.0 &&
	  w->threed.objects[o].vertex[loop].eye.z > -0.000001)
	w->threed.objects[o].vertex[loop].eye.z = -0.000001;
      w->threed.objects[o].vertex[loop].screen.x = (int) (VIEWRATIO *
							  (w->threed.
							   objects[o].
							   vertex[loop].eye.
							   x /
							   w->threed.
							   objects[o].
							   vertex[loop].eye.
							   z) *
							  w->threed.size +
							  w->threed.center.x);
      w->threed.objects[o].vertex[loop].screen.y =
	(int) (-VIEWRATIO *
	       (w->threed.objects[o].vertex[loop].eye.y /
		w->threed.objects[o].vertex[loop].eye.z) * w->threed.size +
	       w->threed.center.y);
    }
  /* Draw graphical object */
  if (w->threed.surface)
    {
      W = &w;
      O = o;
      (void) qsort ((void *) w->threed.objects[o].info,
		    w->threed.objects[o].numSurfaces,
		    sizeof (SURFACE), CompareObject);
    }
  for (surface = 0; surface < w->threed.objects[o].numSurfaces; surface++)
    {
      mapIndex = w->threed.objects[o].info[surface].mapIndex;
      if (w->threed.surface)
	{
	  v1 = &w->threed.objects[o].vertex
	    [w->threed.objects[o].map[mapIndex]].eye;
	  v2 = &w->threed.objects[o].vertex
	    [w->threed.objects[o].map[mapIndex + 1]].eye;
	  v3 = &w->threed.objects[o].vertex
	    [w->threed.objects[o].map[mapIndex + 2]].eye;
	  s1 = v1->x * (v2->y * v3->z - v3->y * v2->z);
	  s2 = v2->x * (v3->y * v1->z - v1->y * v3->z);
	  s3 = v3->x * (v1->y * v2->z - v2->y * v1->z);
	}
      if (!w->threed.surface || -s1 - s2 - s3 <= 0.0)
	{
	  for (vertex = 0;
	       vertex < w->threed.objects[o].info[surface].numVertices;
	       vertex++, mapIndex++)
	    {
	      vertexIndex = w->threed.objects[o].map[mapIndex];
	      points[vertex].x =
		w->threed.objects[o].vertex[vertexIndex].screen.x;
	      points[vertex].y =
		w->threed.objects[o].vertex[vertexIndex].screen.y;
	    }
	  if (!w->threed.surface || s1 + s2 + s3 == 0.0)
	    {
	      points[w->threed.objects[o].info[surface].numVertices].x =
		points[0].x;
	      points[w->threed.objects[o].info[surface].numVertices].y =
		points[0].y;
	      XDrawLines (XtDisplay (w), XtWindow (w),
			  w->threed.shadeGC[w->threed.objects[o].
					    info[surface].brushColor],
			  &points[0], vertex + 1, CoordModeOrigin);
	    }
	  else
	    XFillPolygon (XtDisplay (w), XtWindow (w),
			  w->threed.shadeGC[w->threed.objects[o].
					    info[surface].brushColor],
			  &points[0], vertex, Complex, CoordModeOrigin);
	}
    }
}

static int
CompareObject (const void *elem1, const void *elem2)
{
  if ((*W)->threed.objects[O].vertex[((SURFACE *) elem1)->depthIndex].eye.z >
      (*W)->threed.objects[O].vertex[((SURFACE *) elem2)->depthIndex].eye.z)
    return -1;
  if ((*W)->threed.objects[O].vertex[((SURFACE *) elem1)->depthIndex].eye.z <
      (*W)->threed.objects[O].vertex[((SURFACE *) elem2)->depthIndex].eye.z)
    return 1;
  return 0;
}
