/*
# X-BASED THREED
#
#  ThreedP.h
#
###
#
#  Copyright (c) 1994 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "useful",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/* Private header file for Threed */

#ifndef _ThreedP_h
#define _ThreedP_h

#include "Threed.h"

#define SYMBOL '='

#define NUMDEGREES 360
#define RADIANS(a) (a*(2.0*M_PI/((double)NUMDEGREES)))
#define DEGREES(a) (a*((double)NUMDEGREES/(2.0*M_PI)))
#define ABS(a) (((a)<0)?(-a):(a))
#define SIGN(a) (((a)<0)?-1:1)
#define MIN(a,b) (((int)(a)<(int)(b))?(int)(a):(int)(b))
#define MAX(a,b) (((int)(a)>(int)(b))?(int)(a):(int)(b))

#define VIEWRATIO 4.0
#define MINDISTANCE 0
#define MAXDISTANCE 200
#define DELTADISTANCE 5
#define DELTADEGREES 5
#define MINDEGREES 0
#define MAXDEGREES (NUMDEGREES-1)
#define COLORS 6
#define WHITE_BRUSH 0
#define LTGRAY_BRUSH 1
#define GRAY_BRUSH 2
#define DKGRAY_BRUSH 3
#define BLACK_BRUSH 4
#define ANOTHER_BRUSH 5
#define HIDDENSURFACEREMOVAL True
#define WIREFRAME False

typedef XPoint POINT;

typedef struct tagPoint3D {
  double x, y, z;
} POINT3D;

typedef struct tagVertex {
  POINT3D eye;
  POINT screen;
} VERTEX;

typedef struct tagSurface {
  int mapIndex, numVertices, depthIndex, brushColor;
} SURFACE;

typedef struct tagBox
{
  char name[80];
  int numVertices, numSurfaces, numSurfaceVertices;
  POINT3D *world; /* Vertex Array */
  SURFACE *info; /* Points to Surface Map */
  int *map; /* Points to Vertex Array */
  VERTEX *vertex; /* Vertex Array */
} OBJECT;

typedef struct _ThreeDPart {
  Pixel foreground;
  Pixel shadeColor[COLORS];
  int distance, thetaDegrees, phiDegrees;
  Boolean surface;
  int object, numObjects;
  char *name;
  char **list;
  OBJECT *objects;
  Position size;
  XPoint center;
  GC graphicsGC;
  GC shadeGC[COLORS];
  GC inverseGC;
  XtCallbackList select;
} ThreeDPart;

typedef struct _ThreeDRec {
  CorePart core;
  ThreeDPart threed;
} ThreeDRec;

/* This gets around C's inability to do inheritance */
typedef struct _ThreeDClassPart {
  int ignore;
} ThreeDClassPart;

typedef struct _ThreeDClassRec {
  CoreClassPart core_class;
  ThreeDClassPart threed_class;
} ThreeDClassRec;

extern ThreeDClassRec threedClassRec;

#endif /* _ThreedP_h */
