/*
# MOTIF/X-BASED THREED
#
#  xthreed.c
#
###
#
#  Copyright (c) 1994 - 2003	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "useful",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*
  Version 5: 95/10/02 Xt/Motif
  Version 4: 94/11/17 Xt
  Version 3: 93/05/17 Motif
  Version 2: 93/05/14 XView
  Version 1: 93/03/01 MS Windows by Thomas W. Olsen "The C Users Journal"
*/

#include <stdlib.h>
#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>
#include <Xm/List.h>
#endif
#include "Threed.h"
#include "threed.xbm"

#define NUMDEGREES 360
#define MINDISTANCE 0
#define MAXDISTANCE 200
#define MINDEGREES 0
#define MAXDEGREES (NUMDEGREES-1)
#define HIDDENSURFACEREMOVAL True
#define WIREFRAME False

#define MAXPROGNAME 80
#define MAXNAME (2+MAXPROGNAME)

#ifdef HAVE_MOTIF
static Widget dist, thetaAng, phiAng, surfaceSwitch, objectList;
#else
static char titleDsp[MAXNAME+1];
#endif
static Widget threed;
static Arg arg[2];

static void
Usage(char * programName)
{
  (void) fprintf(stderr, "usage: %s\n", programName);
  (void) fprintf(stderr,
    "\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
  (void) fprintf(stderr,
    "\t[-display [{host}]:[{vs}]][-fg {color}] [-bg {color}]\n");
  (void) fprintf(stderr,
    "\t[-[no]surface] [-object {int}] [-distance {int}] [-theta {int}]\n");
  (void) fprintf(stderr,
    "\t[-phi {int}] [-white {color}] [-[lt|dk]gray {color}]\n");
  (void) fprintf(stderr,
    "\t[-black {color}] [-another {color}]\n");
  exit(1);
}

static XrmOptionDescRec options[] = {
  {(char *) "-fg", (char *) "*cubes.Foreground", XrmoptionSepArg, NULL},
  {(char *) "-bd", (char *) "*cubes.Foreground", XrmoptionSepArg, NULL},
  {(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
  {(char *) "-foreground", (char *) "*cubes.Foreground", XrmoptionSepArg, NULL},
  {(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
  {(char *) "-surface", (char *) "*threed.surface", XrmoptionNoArg, (char *) "TRUE"},
  {(char *) "-nosurface", (char *) "*threed.surface", XrmoptionNoArg, (char *) "FALSE"},
  {(char *) "-object", (char *) "*threed.object", XrmoptionSepArg, (char *) "0"},
  {(char *) "-distance", (char *) "*threed.distance", XrmoptionSepArg, (char *) "1"},
  {(char *) "-theta", (char *) "*threed.thetaDegrees", XrmoptionSepArg, (char *) "0"},
  {(char *) "-phi", (char *) "*threed.phiDegrees", XrmoptionSepArg, (char *) "0"},
  {(char *) "-white", (char *) "*threed.whiteBrush", XrmoptionSepArg, NULL},
  {(char *) "-ltgray", (char *) "*threed.ltgrayBrush", XrmoptionSepArg, NULL},
  {(char *) "-gray", (char *) "*threed.grayBrush", XrmoptionSepArg, NULL},
  {(char *) "-dkgray", (char *) "*threed.dkgrayBrush", XrmoptionSepArg, NULL},
  {(char *) "-black", (char *) "*threed.blackBrush", XrmoptionSepArg, NULL},
  {(char *) "-another", (char *) "*threed.anotherBrush", XrmoptionSepArg, NULL},
};

#ifndef HAVE_MOTIF
static void PrintState(Widget w, String name, Boolean surface, int object, int distance, int theta, int phi)
{
  (void) sprintf(titleDsp, "%s %d %s:dist %d:theta %d:phi %d", name, 
    object, (surface) ? "surface" : "wire",
    distance, theta, phi);
  XtSetArg(arg[0], XtNtitle, titleDsp);
  XtSetValues(w, arg, 1);
}
#endif

static void Initialize(Widget w)
{
  int object, distance, theta, phi;
  Boolean surface;
  String name;
#ifndef HAVE_MOTIF
  String *list;
  int number, i;
#endif

  XtVaGetValues(w,
    XtNdistance, &distance,
    XtNthetaDegrees, &theta,
    XtNphiDegrees, &phi,
    XtNobject, &object,
    XtNsurface, &surface,
    XtNobjectName, &name,
#ifndef HAVE_MOTIF
    XtNobjectNumber, &number,
    XtNobjectList, &list,
#endif
    NULL);
#ifdef HAVE_MOTIF
  XmScaleSetValue(dist, distance);
  XmScaleSetValue(thetaAng, theta);
  XmScaleSetValue(phiAng, phi);
  XmToggleButtonSetState(surfaceSwitch, surface, True);
#else
  PrintState(XtParent(w), name, surface, object, distance, theta, phi);
  (void) printf("Objects:\n");
  for (i = 0; i < number; i++)
    (void) printf("\t%s\n", list[i]);
#endif
}

#ifdef HAVE_MOTIF
static void makePosVisible(Widget list_w, int item_no)
{
  int top, visible;

  XtVaGetValues(list_w,
    XmNtopItemPosition, &top,
    XmNvisibleItemCount, &visible,
    NULL);
  if (item_no < top)
    XmListSetPos(list_w, item_no);
  else if (item_no >= top + visible)
    XmListSetBottomPos(list_w, item_no);
}
#endif

static void CallbackThreeD(Widget w, caddr_t clientData, threedCallbackStruct *callData)
{
  Boolean surface;
  int object, distance, theta, phi, numObjects;
  String name;

  XtVaGetValues(w,
    XtNobject, &object,
    XtNsurface, &surface,
    XtNdistance, &distance,
    XtNthetaDegrees, &theta,
    XtNphiDegrees, &phi,
    XtNobjectNumber, &numObjects,
    XtNobjectName, &name,
    NULL);
  switch (callData->reason) {
    case THREED_SURFACE:
      surface = !surface;
      XtSetArg(arg[0], XtNsurface, surface);
      XtSetValues(w, arg, 1);
#ifdef HAVE_MOTIF
      XmToggleButtonSetState(surfaceSwitch, surface, True);
#endif
      break;
    case THREED_OBJECT:
      object = (object + 1) % numObjects;
      XtSetArg(arg[0], XtNobject, object);
      XtSetValues(w, arg, 1);
      XtVaGetValues(w,
        XtNobjectName, &name,
        NULL);
#ifdef HAVE_MOTIF
      makePosVisible(objectList, object + 1);
      XmListSelectPos(objectList, object + 1, True);
#endif
      break;
  }
#ifdef HAVE_MOTIF
  XmScaleSetValue(dist, distance);
  XmScaleSetValue(thetaAng, theta);
  XmScaleSetValue(phiAng, phi);
#else
  PrintState(XtParent(w), name, surface, object, distance, theta, phi);
#endif
}

#ifdef HAVE_MOTIF
static void
DistSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct *cbs)
{
  int distance = cbs->value, old;

  XtVaGetValues(threed,
    XtNdistance, &old,
    NULL);
  if (old != distance)
    XtVaSetValues(threed,
      XtNdistance, distance,
      NULL);
}

static void
ThetaSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct *cbs)
{
  int theta = cbs->value, old;

  XtVaGetValues(threed,
    XtNthetaDegrees, &old,
    NULL);
  if (old != theta)
    XtVaSetValues(threed,
      XtNthetaDegrees, theta,
      NULL);
}

static void
PhiSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct *cbs)
{
  int phi = cbs->value, old;

  XtVaGetValues(threed,
    XtNphiDegrees, &old,
    NULL);
  if (old != phi)
    XtVaSetValues(threed,
      XtNphiDegrees, phi,
      NULL);
}


static void
SurfaceToggle(Widget w, XtPointer clientData, XmToggleButtonCallbackStruct *cbs)
{
  XtVaSetValues(threed,
    XtNsurface, cbs->set,
    NULL);
}

static void
ObjectResponse(Widget listItem, XtPointer clientData, XmListCallbackStruct *cbs)
{
  String name;
  int object = cbs->item_position - 1, old;

  XtVaGetValues(threed,
    XtNobject, &old,
    NULL);
  if (old != object)
    XtVaSetValues(threed,
      XtNobject, object,
      NULL);
  XtVaGetValues(threed,
    XtNobjectName, &name,
    NULL);
}

#if 0
static void
createMenu(Widget *menu, Widget rowcol, int init)
{
  XmString choice, tetra, hexa, octa, dodeca, icosa, cubes, f16;

  choice = XmStringCreateSimple("Object:");
  tetra = XmStringCreateSimple("Tetrahedron");
  hexa = XmStringCreateSimple("Hexahedron");
  octa = XmStringCreateSimple("Octahedron");
  dodeca = XmStringCreateSimple("Dodecahedron");
  icosa = XmStringCreateSimple("Icosahedron");
  cubes = XmStringCreateSimple("Cubes");
  f16 = XmStringCreateSimple("F16");
  *menu = XmVaCreateSimpleOptionMenu(rowcol,
    "mode_menu", choice, 'M', init, object_response,
    XmVaPUSHBUTTON, tetra, 'T', NULL, NULL,
    XmVaPUSHBUTTON, hexa, 'H', NULL, NULL,
    XmVaPUSHBUTTON, octa, 'O', NULL, NULL,
    XmVaPUSHBUTTON, dodeca, 'D', NULL, NULL,
    XmVaPUSHBUTTON, icosa, 'I', NULL, NULL,
    XmVaPUSHBUTTON, cubes, 'C', NULL, NULL,
    XmVaPUSHBUTTON, f16, 'F', NULL, NULL,
    NULL);
  XmStringFree(tetra);
  XmStringFree(hexa);
  XmStringFree(octa);
  XmStringFree(dodeca);
  XmStringFree(icosa);
  XmStringFree(cubes);
  XmStringFree(f16);
  XtManageChild(*menu);
}
#endif

static void
createList(Widget *rowcol, String *list, int init, int num)
{
  XmString label;
  XmStringTable table;
  int i;

  label = XmStringCreateSimple((char *) "Object list:");
  XtVaCreateManagedWidget("listLabel", xmLabelWidgetClass, *rowcol,
    XmNlabelString, label,
    NULL);
  XmStringFree(label);

  table = (XmStringTable) XtMalloc (num * sizeof (XmString *));
  for (i = 0; i < num; i++)
    table[i] = XmStringCreateSimple(list[i]);

  objectList = XmCreateScrolledList(*rowcol, (char *) "scrolledList", NULL, 0);
  XtVaSetValues(objectList,
    XmNvisibleItemCount, 5,
    XmNitemCount, num,
    XmNitems, table,
    NULL);
  XtManageChild(objectList);
  XtAddCallback(objectList, XmNbrowseSelectionCallback,
   (XtCallbackProc) ObjectResponse, (XtPointer) NULL);
  XmListSelectPos(objectList, init + 1, True);

  for (i = 0; i < num; i++)
    XmStringFree(table[i]);
  XtFree((char *) table);
}

#endif

int main(int argc, char **argv)
{
  Widget topLevel;
#ifdef HAVE_MOTIF
  Widget panel, rowcol;
  Pixel fg, bg;
  String *list;
  int object, number;
#endif

  topLevel = XtInitialize(argv[0], "Threed",
     options, XtNumber(options), &argc, argv);
  if (argc != 1)
    Usage(argv[0]);

  XtSetArg(arg[0], XtNiconPixmap,
    XCreateBitmapFromData(XtDisplay(topLevel), XtScreen(topLevel)->root,
      (char *) threed_bits, threed_width, threed_height));
#ifdef HAVE_MOTIF
  XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER); /* not XmEXPLICIT */
  XtSetValues(topLevel, arg, 2);
  panel = XtCreateManagedWidget("panel", xmPanedWindowWidgetClass, topLevel,
    NULL, 0);
  rowcol = XtVaCreateManagedWidget("Rowcol2", xmRowColumnWidgetClass, panel,
    NULL);
  XtVaGetValues(rowcol, XmNforeground, &fg, XmNbackground, &bg, NULL);
  dist = XtVaCreateManagedWidget("dist", xmScaleWidgetClass, rowcol,
    XtVaTypedArg, XmNtitleString, XmRString, "Distance", 9,
    XmNminimum, MINDISTANCE,
    XmNmaximum, MAXDISTANCE,
    XmNvalue, MAXDISTANCE,
    XmNshowValue, True,
    XmNorientation, XmHORIZONTAL,
    NULL);
  XtAddCallback(dist, XmNvalueChangedCallback, (XtCallbackProc) DistSlider,
    (XtPointer) NULL);
  thetaAng = XtVaCreateManagedWidget("thetaAng", xmScaleWidgetClass, rowcol,
    XtVaTypedArg, XmNtitleString, XmRString, "Theta", 6,
    XmNminimum, MINDEGREES,
    XmNmaximum, MAXDEGREES,
    XmNvalue, MINDEGREES,
    XmNshowValue, True,
    XmNorientation, XmHORIZONTAL,
    NULL);
  XtAddCallback(thetaAng, XmNvalueChangedCallback, (XtCallbackProc) ThetaSlider,
    (XtPointer) NULL);
  phiAng = XtVaCreateManagedWidget("phiAng", xmScaleWidgetClass, rowcol,
    XtVaTypedArg, XmNtitleString, XmRString, "Phi", 4,
    XmNminimum, MINDEGREES,
    XmNmaximum, MAXDEGREES,
    XmNvalue, MINDEGREES,
    XmNshowValue, True,
    XmNorientation, XmHORIZONTAL,
    NULL);
  XtAddCallback(phiAng, XmNvalueChangedCallback, (XtCallbackProc) PhiSlider,
    (XtPointer) NULL);
  surfaceSwitch = XtVaCreateManagedWidget("Surface",
    xmToggleButtonWidgetClass, rowcol,
    NULL);
  XtAddCallback(surfaceSwitch, XmNvalueChangedCallback,
    (XtCallbackProc) SurfaceToggle, (XtPointer) NULL);

  threed = XtCreateManagedWidget("threed", threedWidgetClass, panel,
    NULL, 0);
  XtAddCallback(threed, XtNselectCallback, (XtCallbackProc) CallbackThreeD,
    (XtPointer) NULL);
  XtVaGetValues(threed,
    XtNobject, &object,
    XtNobjectNumber, &number,
    XtNobjectList, &list,
    NULL);
  /*createMenu(&objectMenu, rowcol, object);*/
  createList(&rowcol, list, object, number);
#else
  XtSetArg(arg[1], XtNinput, True);
  XtSetValues(topLevel, arg, 2);
  threed = XtCreateManagedWidget("threed", threedWidgetClass, topLevel,
    NULL, 0);
#endif
  XtAddCallback(threed,
    XtNselectCallback, (XtCallbackProc) CallbackThreeD, (XtPointer) NULL);
  Initialize(threed);
  XtRealizeWidget(topLevel);
  XGrabButton(XtDisplay(threed), (unsigned int) AnyButton, AnyModifier,
    XtWindow(threed), TRUE,
    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
    GrabModeAsync, GrabModeAsync, XtWindow(threed),
    XCreateFontCursor(XtDisplay(threed), XC_crosshair));
  XtMainLoop();

#ifdef VMS
  return 1;
#else
  return 0;
#endif
}
