/*-
# MOTIF/X-BASED TRIANGLES
#
#  xtriangles.c
#
###
#
#  Copyright (c) 1993 - 2002	David Albert Bagley, bagleyd@tux.org
#
#                   All Rights Reserved
#
#  Permission to use, copy, modify, and distribute this software and
#  its documentation for any purpose and without fee is hereby granted,
#  provided that the above copyright notice appear in all copies and
#  that both that copyright notice and this permission notice appear in
#  supporting documentation, and that the name of the author not be
#  used in advertising or publicity pertaining to distribution of the
#  software without specific, written prior permission.
#
#  This program is distributed in the hope that it will be "playable",
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
*/

/*-
  Version 5: 95/10/01 Xt/Motif
  Version 4: 94/06/07 Xt
  Version 3: 93/04/01 Motif
  Version 2: 92/01/06 XView
  Version 1: 91/09/05 SunView
*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#ifdef VMS
#include <unixlib.h>
#define getlogin cuserid
#else
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#endif
#if HAVE_FCNTL_H
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#ifdef HAVE_MOTIF
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/LabelG.h>
#include <Xm/Scale.h>
#include "mouse-l.xbm"
#include "mouse-r.xbm"
#endif
#include "Triangles.h"
#include "triangles.xbm"

#ifndef SCOREFILE
#define SCOREFILE "/usr/games/lib/triangles.scores"
#endif

/* The following is in TrianglesP.h also */
#define MINTRIANGLES 1

#define MAXTRIANGLES 16
#define NEVER -1
#define FILENAMELEN 1024
#define USERNAMELEN 128
#define NOACCESS "noaccess"
#define NOBODY "nobody"

typedef struct {
	int         score;
	char        name[USERNAMELEN];
} GameRecord;

#ifdef HAVE_MOTIF
static Widget moves, record, message, tri;
static char buff[16];
#else
static char titleDsp[256] = "";
#endif

static Widget triangles;
static Arg  arg[2];
static GameRecord trianglesRecord[MAXTRIANGLES - MINTRIANGLES + 1];
static int  movesDsp = 0;
static char *progDsp;
static char messageDsp[128] = "Welcome";
static char recordDsp[16] = "NOT RECORDED";
static char usernameDsp[USERNAMELEN] = "";

static void
Usage(char * programName)
{
	(void) fprintf(stderr, "usage: %s\n", programName);
	(void) fprintf(stderr,
		"\t[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]]\n");
	(void) fprintf(stderr,
		"\t[-display [{host}]:[{vs}]][-fg {color}] [-[no]mono] [-[no]{reverse|rv}]\n");
	(void) fprintf(stderr,
		"\t[-{foreground|fg} {color}] [-{background|bg} {color}]\n");
	(void) fprintf(stderr,
		"\t[-{border|bd} {color}] [-tile {color}] [-size {int}]\n");
	(void) fprintf(stderr,
		"\t[-base {int}] [-username {string}]\n");
	exit(1);
}

static XrmOptionDescRec options[] =
{
	{(char *) "-mono", (char *) "*triangles.mono", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-nomono", (char *) "*triangles.mono", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-rv", (char *) "*triangles.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-reverse", (char *) "*triangles.reverse", XrmoptionNoArg, (char *) "TRUE"},
	{(char *) "-norv", (char *) "*triangles.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-noreverse", (char *) "*triangles.reverse", XrmoptionNoArg, (char *) "FALSE"},
	{(char *) "-fg", (char *) "*triangles.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-foreground", (char *) "*triangles.Foreground", XrmoptionSepArg, NULL},
	{(char *) "-bg", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-background", (char *) "*Background", XrmoptionSepArg, NULL},
	{(char *) "-bd", (char *) "*triangles.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-border", (char *) "*triangles.tileBorder", XrmoptionSepArg, NULL},
	{(char *) "-tile", (char *) "*triangles.tileColor", XrmoptionSepArg, NULL},
	{(char *) "-size", (char *) "*triangles.size", XrmoptionSepArg, NULL},
	{(char *) "-base", (char *) "*triangles.base", XrmoptionSepArg, NULL},
	{(char *) "-username", (char *) "*triangles.userName", XrmoptionSepArg, NULL}
};

#ifdef HAVE_MOTIF
static void
PrintState(Widget w, char *message)
{
	Arg         wargs[1];
	XmString    xmstr;

	if (!XtIsSubclass(w, xmLabelWidgetClass))
		XtError("PrintState() requires a Label Widget");
	xmstr = XmStringCreateLtoR(message, XmSTRING_DEFAULT_CHARSET);
	XtSetArg(wargs[0], XmNlabelString, xmstr);
	XtSetValues(w, wargs, 1);
}
#else
static void
PrintState(Widget w, int size, int moves, char *message)
{
	(void) sprintf(titleDsp, "%s: %d@ (%d/%s) - %s",
		progDsp, size, moves, recordDsp, message);
	XtSetArg(arg[0], XtNtitle, titleDsp);
	XtSetValues(w, arg, 1);
}
#endif

static void
InitRecords(void)
{
	int         i;

	for (i = 0; i < MAXTRIANGLES - MINTRIANGLES + 1; i++) {
		trianglesRecord[i].score = NEVER;
		(void) strcpy(trianglesRecord[i].name, NOACCESS);
	}
}

static void
ReadRecords(void)
{
	FILE       *fp;
	int         i, n;
	char        username[USERNAMELEN];

	if ((fp = fopen(SCOREFILE, "r")) == NULL) {
		(void) sprintf(messageDsp, "Can not open %s, taking defaults.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
		for (i = 0; i < MAXTRIANGLES - MINTRIANGLES + 1; i++) {
			(void) fscanf(fp, "%d %s\n", &n, username);
			if (n <= trianglesRecord[i].score ||
			    trianglesRecord[i].score <= NEVER) {
				trianglesRecord[i].score = n;
				(void) strcpy(trianglesRecord[i].name, username);
			}
		}
		(void) fclose(fp);
	}
}

static void
WriteRecords(void)
{
	FILE       *fp;
	int         i;

	if ((fp = fopen(SCOREFILE, "w")) == NULL) {
		(void) sprintf(messageDsp, "Can not write to %s.", SCOREFILE);
#ifdef HAVE_MOTIF
		PrintState(message, messageDsp);
#endif
	} else {
#if HAVE_FCNTL_H
		int         lfd;
		char        lockfile[FILENAMELEN];

		(void) strcpy(lockfile, SCOREFILE);
		(void) strcat(lockfile, ".lock");
		while (((lfd = open(lockfile, O_CREAT | O_EXCL, 0644)) < 0) &&
		       errno == EEXIST)
			(void) sleep(1);
		if (lfd < 0) {
#if 1
			(void) fprintf(stderr, "Lock file exists... guessing its an old one.\n");
#else
			(void) fprintf(stderr, "Lock file exists... score not recorded - sorry.\n"
				);
			return;
#endif
		}
#endif
		for (i = 0; i < MAXTRIANGLES - MINTRIANGLES + 1; i++)
			(void) fprintf(fp, "%d %s\n",
			  trianglesRecord[i].score, trianglesRecord[i].name);
#if HAVE_FCNTL_H
		(void) close(lfd);
		(void) unlink(lockfile);
#endif
		(void) fclose(fp);
	}
}

static void
PrintRecord(int size)
{
	int         i = size - MINTRIANGLES;

	if (size > MAXTRIANGLES)
		(void) strcpy(recordDsp, "NOT RECORDED");
	else if (trianglesRecord[i].score <= NEVER) {
		(void) sprintf(recordDsp, "NEVER %s", NOACCESS);
	} else {
		(void) sprintf(recordDsp, "%d %s",
			  trianglesRecord[i].score, trianglesRecord[i].name);
	}
#ifdef HAVE_MOTIF
	PrintState(record, recordDsp);
#endif
}

static      Boolean
HandleSolved(int counter, int size)
{
	int         i = size - MINTRIANGLES;

	if (size <= MAXTRIANGLES && (counter < trianglesRecord[i].score ||
	    trianglesRecord[i].score <= NEVER)) {
		ReadRecords();	/* Maybe its been updated by another */
		trianglesRecord[i].score = counter;
		(void) strcpy(trianglesRecord[i].name, usernameDsp);
		WriteRecords();
		PrintRecord(size);
		return True;
	}
	return False;
}

static void
Initialize(Widget w)
{
	int         size;
	String      username;

	XtVaSetValues(w,
		      XtNstart, False, NULL);
	XtVaGetValues(w,
		      XtNuserName, &username,
		      XtNsize, &size, NULL);
#ifdef HAVE_MOTIF
	if (size > MAXTRIANGLES)
		XtVaSetValues(tri, XmNmaximum, size, NULL);
	XmScaleSetValue(tri, size);
#endif
	InitRecords();
	ReadRecords();
	(void) strcpy(usernameDsp, username);
	if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
	    !strcmp(usernameDsp, NOACCESS) || !strcmp(usernameDsp, NOBODY)) {
		/* The NOACCESS is not necessary, but it stops people from being cute. */
		(void) sprintf(usernameDsp, "%s", getlogin());
		if (!strcmp(usernameDsp, "") || !strcmp(usernameDsp, "(null)") ||
		    !strcmp(usernameDsp, NOACCESS))
			(void) sprintf(usernameDsp, "%s", NOBODY);	/* It really IS nobody */
	}
	PrintRecord(size);
#ifndef HAVE_MOTIF
	PrintState(XtParent(w), size, movesDsp, messageDsp);
#endif
}

static void
CallbackTriangles(Widget w, caddr_t clientData,
		  trianglesCallbackStruct * callData)
{
	int         size;

	XtVaGetValues(w,
		      XtNsize, &size, NULL);
	(void) strcpy(messageDsp, "");
	switch (callData->reason) {
		case TRIANGLES_RESTORE:
		case TRIANGLES_RESET:
			movesDsp = 0;
			break;
		case TRIANGLES_BLOCKED:
			(void) strcpy(messageDsp, "Blocked");
			break;
		case TRIANGLES_SPACE:
#if 0
			/* Too annoying */
			(void) strcpy(messageDsp, "Spaces can not slide");
#endif
			break;
		case TRIANGLES_IGNORE:
			(void) strcpy(messageDsp, "Randomize to start");
			break;
		case TRIANGLES_MOVED:
			movesDsp++;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
		case TRIANGLES_SOLVED:
			if (HandleSolved(movesDsp, size))
				(void) sprintf(messageDsp, "Congratulations %s!!", usernameDsp);
			else
				(void) strcpy(messageDsp, "Solved!");
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case TRIANGLES_RANDOMIZE:
			movesDsp = 0;
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case TRIANGLES_DEC:
			movesDsp = 0;
			size--;
			XtSetArg(arg[0], XtNsize, size);
			XtSetValues(w, arg, 1);
			PrintRecord(size);
#ifdef HAVE_MOTIF
			XmScaleSetValue(tri, size);
			if (size >= MAXTRIANGLES)
				XtVaSetValues(tri, XmNmaximum, size, NULL);
#endif
			break;
		case TRIANGLES_INC:
			movesDsp = 0;
			size++;
			XtSetArg(arg[0], XtNsize, size);
			XtSetValues(w, arg, 1);
			PrintRecord(size);
#ifdef HAVE_MOTIF
			if (size > MAXTRIANGLES)
				XtVaSetValues(tri, XmNmaximum, size, NULL);
			XmScaleSetValue(tri, size);
#endif
			break;
		case TRIANGLES_COMPUTED:
			XtSetArg(arg[0], XtNstart, False);
			XtSetValues(w, arg, 1);
			break;
		case TRIANGLES_UNDO:
			movesDsp--;
			XtSetArg(arg[0], XtNstart, True);
			XtSetValues(w, arg, 1);
			break;
	}
#ifdef HAVE_MOTIF
	PrintState(message, messageDsp);
	(void) sprintf(buff, "%d", movesDsp);
	PrintState(moves, buff);
#else
	PrintState(XtParent(w), size, movesDsp, messageDsp);
#endif
}

#ifdef HAVE_MOTIF
static void
triSlider(Widget w, XtPointer clientData, XmScaleCallbackStruct * cbs)
{
	int         size = cbs->value, old;

	XtVaGetValues(triangles,
		      XtNsize, &old, NULL);
	if (old != size) {
		XtVaSetValues(triangles,
			      XtNsize, size, NULL);
		movesDsp = 0;
		(void) sprintf(buff, "%d", movesDsp);
		PrintState(moves, buff);
		PrintRecord(size);
	}
}

static void
fileCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 2)
		exit(0);
	XtSetArg(arg[0], XtNmenu, val);
	XtSetValues(triangles, arg, 1);
}

static void
playCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;
	XtSetArg(arg[0], XtNmenu, val + 3);  /* GWQ */
	XtSetValues(triangles, arg, 1);
}

static const char aboutHelp[] = {
	"Version 5.6.1 by David Bagley <bagleyd@tux.org> http://www.tux.org/~bagleyd"
};

static const char helpHelp[] = {
	"Consult man page.  :)"
};

static void
helpCB(Widget w, void *value, void *clientData)
{
	int val = (int) value;

	if (val == 0)
		(void) printf("%s: %s\n", progDsp, aboutHelp);
	else
		(void) printf("%s: %s\n", progDsp, helpHelp);
}
#endif

int
main(int argc, char **argv)
{
	Widget      topLevel;
#ifdef HAVE_MOTIF
	Widget      menubar, panel, panel2, panel3, rowcol, rowcol2, rowcol3;
	Pixmap      mouseLeftCursor, mouseRightCursor;
	Pixel       fg, bg;
	XmString    file, play, help, about;
	XmString    get, write, quit;
	XmString    clear, undo, randomize, solve, increment, decrement;
#endif

	progDsp = argv[0];
	topLevel = XtInitialize(argv[0], "Triangles",
				options, XtNumber(options), &argc, argv);
	if (argc != 1)
		Usage(argv[0]);

	XtSetArg(arg[0],
		 XtNiconPixmap, XCreateBitmapFromData(XtDisplay(topLevel),
				      RootWindowOfScreen(XtScreen(topLevel)),
		(char *) triangles_bits, triangles_width, triangles_height));
#ifdef HAVE_MOTIF
	XtSetArg(arg[1], XmNkeyboardFocusPolicy, XmPOINTER);	/* not XmEXPLICIT */
	XtSetValues(topLevel, arg, 2);
	panel = XtVaCreateManagedWidget("panel",
				xmPanedWindowWidgetClass, topLevel,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	file = XmStringCreateSimple((char *) "File");
	play = XmStringCreateSimple((char *) "Play");
	help = XmStringCreateSimple((char *) "Help");
	menubar = XmVaCreateSimpleMenuBar(panel, (char *) "menubar",
		XmVaCASCADEBUTTON, file, 'F',
		XmVaCASCADEBUTTON, play, 'P',
		XmVaCASCADEBUTTON, help, 'H',
		NULL);
	XmStringFree(file);
	XmStringFree(play);
	get = XmStringCreateSimple((char *) "Get");
	write = XmStringCreateSimple((char *) "Write");
	quit = XmStringCreateSimple((char *) "Quit");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "file_menu", 0, fileCB,
		XmVaPUSHBUTTON, get, 'G', NULL, NULL,
		XmVaPUSHBUTTON, write, 'W', NULL, NULL,
		XmVaSEPARATOR,
		XmVaPUSHBUTTON, quit, 'Q', NULL, NULL,
		NULL);
	XmStringFree(get);
	XmStringFree(write);
	XmStringFree(quit);
	clear = XmStringCreateSimple((char *) "Clear");
	undo = XmStringCreateSimple((char *) "Undo");
	randomize = XmStringCreateSimple((char *) "Randomize");
	solve = XmStringCreateSimple((char *) "(Solve)");
	increment = XmStringCreateSimple((char *) "Increment");
	decrement = XmStringCreateSimple((char *) "Decrement");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "play_menu", 1, playCB,
		XmVaPUSHBUTTON, clear, 'C', NULL, NULL,
		XmVaPUSHBUTTON, undo, 'U', NULL, NULL,
		XmVaPUSHBUTTON, randomize, 'R', NULL, NULL,
		XmVaPUSHBUTTON, solve, 'S', NULL, NULL,
		XmVaPUSHBUTTON, increment, 'I', NULL, NULL,
		XmVaPUSHBUTTON, decrement, 'D', NULL, NULL,
		NULL);
	XmStringFree(clear);
	XmStringFree(undo);
	XmStringFree(randomize);
	XmStringFree(solve);
	XmStringFree(increment);
	XmStringFree(decrement);
	about = XmStringCreateSimple((char *) "About");
	XmVaCreateSimplePulldownMenu(menubar, (char *) "help_menu", 2, helpCB,
		XmVaPUSHBUTTON, about, 'A', NULL, NULL,
		XmVaPUSHBUTTON, help, 'H', NULL, NULL,
		NULL);
	XmStringFree(about);
	XmStringFree(help);
	XtManageChild(menubar);
	panel2 = XtCreateManagedWidget("panel2",
				       xmPanedWindowWidgetClass, panel,
				       NULL, 0);
	panel3 = XtVaCreateManagedWidget("panel3",
					 xmPanedWindowWidgetClass, panel2,
					 XmNseparatorOn, False,
					 XmNsashWidth, 1,
					 XmNsashHeight, 1, NULL);
	rowcol = XtVaCreateManagedWidget("Rowcol",
					 xmRowColumnWidgetClass, panel3,
					 XmNnumColumns, 2,
					 XmNorientation, XmHORIZONTAL,
					 XmNpacking, XmPACK_COLUMN, NULL);
	XtVaGetValues(rowcol,
		      XmNforeground, &fg,
		      XmNbackground, &bg, NULL);
	mouseLeftCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	      RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_left_bits,
				 mouse_left_width, mouse_left_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	mouseRightCursor = XCreatePixmapFromBitmapData(XtDisplay(rowcol),
	     RootWindowOfScreen(XtScreen(rowcol)), (char *) mouse_right_bits,
			       mouse_right_width, mouse_right_height, fg, bg,
				     DefaultDepthOfScreen(XtScreen(rowcol)));
	(void) XtVaCreateManagedWidget("mouseLeftText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Move tile", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseLeft",
				xmLabelGadgetClass, rowcol,
	      XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseLeftCursor, NULL);
	(void) XtVaCreateManagedWidget("mouseRightText",
				xmLabelGadgetClass, rowcol,
	     XtVaTypedArg, XmNlabelString, XmRString, "Randomize", 10, NULL);
	(void) XtVaCreateManagedWidget("mouseRight",
				xmLabelGadgetClass, rowcol,
	     XmNlabelType, XmPIXMAP, XmNlabelPixmap, mouseRightCursor, NULL);
	(void) XtVaCreateManagedWidget("movesText",
				xmLabelGadgetClass, rowcol,
		  XtVaTypedArg, XmNlabelString, XmRString, "Moves", 6, NULL);
	moves = XtVaCreateManagedWidget("0",
					xmLabelWidgetClass, rowcol, NULL);
	(void) XtVaCreateManagedWidget("recordText",
				xmLabelGadgetClass, rowcol,
		 XtVaTypedArg, XmNlabelString, XmRString, "Record", 7, NULL);
	record = XtVaCreateManagedWidget("0",
					 xmLabelWidgetClass, rowcol, NULL);

	rowcol2 = XtVaCreateManagedWidget("Rowcol2",
				       xmRowColumnWidgetClass, panel3,
					  XmNnumColumns, 1,
					  XmNorientation, XmHORIZONTAL,
					  XmNpacking, XmPACK_COLUMN, NULL);
	tri = XtVaCreateManagedWidget("tris",
				      xmScaleWidgetClass, rowcol2,
			  XtVaTypedArg, XmNtitleString, XmRString, "Size", 5,
				      XmNminimum, MINTRIANGLES,
				      XmNmaximum, MAXTRIANGLES,
				      XmNvalue, MINTRIANGLES,
				      XmNshowValue, True,
				      XmNorientation, XmHORIZONTAL, NULL);
	XtAddCallback(tri,
		      XmNvalueChangedCallback, (XtCallbackProc) triSlider, (XtPointer) NULL);
	rowcol3 = XtVaCreateManagedWidget("Rowcol3",
				       xmRowColumnWidgetClass, panel3, NULL);
	message = XtVaCreateManagedWidget("Play Triangles! (use mouse or keypad)",
					  xmLabelWidgetClass, rowcol3, NULL);

	triangles = XtCreateManagedWidget("triangles",
				       trianglesWidgetClass, panel2, NULL, 0);
#else
	XtSetArg(arg[1], XtNinput, True);
	XtSetValues(topLevel, arg, 2);
	triangles = XtCreateManagedWidget("triangles",
				      trianglesWidgetClass, topLevel, NULL, 0);
#endif
	XtAddCallback(triangles,
		      XtNselectCallback, (XtCallbackProc) CallbackTriangles, (XtPointer) NULL);
	Initialize(triangles);
	XtRealizeWidget(topLevel);
	XGrabButton(XtDisplay(triangles), (unsigned int) AnyButton, AnyModifier,
		    XtWindow(triangles), TRUE,
		    (unsigned int) (ButtonPressMask | ButtonMotionMask | ButtonReleaseMask),
		    GrabModeAsync, GrabModeAsync, XtWindow(triangles),
		    XCreateFontCursor(XtDisplay(triangles), XC_crosshair));
	XtMainLoop();

#ifdef VMS
	return 1;
#else
	return 0;
#endif
}
